# Generated from jac.g4 by ANTLR 4.9.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3m")
        buf.write("\u04e9\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
        buf.write("\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4\30\t\30\4\31")
        buf.write("\t\31\4\32\t\32\4\33\t\33\4\34\t\34\4\35\t\35\4\36\t\36")
        buf.write('\4\37\t\37\4 \t \4!\t!\4"\t"\4#\t#\4$\t$\4%\t%\4&\t')
        buf.write("&\4'\t'\4(\t(\4)\t)\4*\t*\4+\t+\4,\t,\4-\t-\4.\t.\4")
        buf.write("/\t/\4\60\t\60\4\61\t\61\4\62\t\62\4\63\t\63\4\64\t\64")
        buf.write("\4\65\t\65\4\66\t\66\4\67\t\67\48\t8\49\t9\4:\t:\4;\t")
        buf.write(";\4<\t<\4=\t=\4>\t>\4?\t?\4@\t@\4A\tA\4B\tB\4C\tC\4D\t")
        buf.write("D\4E\tE\4F\tF\4G\tG\4H\tH\4I\tI\4J\tJ\4K\tK\4L\tL\4M\t")
        buf.write("M\4N\tN\4O\tO\4P\tP\4Q\tQ\4R\tR\4S\tS\4T\tT\4U\tU\4V\t")
        buf.write("V\4W\tW\4X\tX\4Y\tY\4Z\tZ\4[\t[\4\\\t\\\4]\t]\4^\t^\4")
        buf.write("_\t_\4`\t`\4a\ta\4b\tb\4c\tc\4d\td\4e\te\4f\tf\4g\tg\4")
        buf.write("h\th\4i\ti\4j\tj\4k\tk\4l\tl\4m\tm\4n\tn\4o\to\4p\tp\4")
        buf.write("q\tq\3\2\5\2\u00e4\n\2\3\2\7\2\u00e7\n\2\f\2\16\2\u00ea")
        buf.write("\13\2\3\2\7\2\u00ed\n\2\f\2\16\2\u00f0\13\2\3\2\3\2\3")
        buf.write("\3\3\3\3\3\3\3\5\3\u00f8\n\3\3\3\3\3\3\3\3\3\3\3\3\4\3")
        buf.write("\4\3\4\3\4\5\4\u0103\n\4\3\5\3\5\3\5\5\5\u0108\n\5\3\5")
        buf.write("\3\5\5\5\u010c\n\5\3\5\3\5\3\5\5\5\u0111\n\5\3\5\3\5\5")
        buf.write("\5\u0115\n\5\3\5\3\5\3\5\5\5\u011a\n\5\3\5\3\5\5\5\u011e")
        buf.write("\n\5\3\5\3\5\3\5\5\5\u0123\n\5\3\5\3\5\5\5\u0127\n\5\3")
        buf.write("\5\3\5\3\5\5\5\u012c\n\5\3\5\3\5\5\5\u0130\n\5\3\5\3\5")
        buf.write("\3\5\5\5\u0135\n\5\3\5\3\5\5\5\u0139\n\5\5\5\u013b\n\5")
        buf.write("\3\6\3\6\3\6\3\6\3\6\3\6\3\6\5\6\u0144\n\6\3\7\3\7\3\7")
        buf.write("\5\7\u0149\n\7\3\b\3\b\3\b\3\b\3\b\3\b\3\b\3\b\7\b\u0153")
        buf.write("\n\b\f\b\16\b\u0156\13\b\3\b\3\b\3\t\3\t\3\t\3\t\7\t\u015e")
        buf.write("\n\t\f\t\16\t\u0161\13\t\3\t\3\t\5\t\u0165\n\t\3\t\3\t")
        buf.write("\3\t\3\t\3\t\7\t\u016c\n\t\f\t\16\t\u016f\13\t\3\t\3\t")
        buf.write("\3\t\3\t\3\t\3\t\3\t\3\t\5\t\u0179\n\t\3\t\3\t\3\t\5\t")
        buf.write("\u017e\n\t\3\t\5\t\u0181\n\t\3\n\3\n\7\n\u0185\n\n\f\n")
        buf.write("\16\n\u0188\13\n\3\n\5\n\u018b\n\n\3\n\3\n\7\n\u018f\n")
        buf.write("\n\f\n\16\n\u0192\13\n\3\n\5\n\u0195\n\n\3\n\3\n\3\13")
        buf.write("\3\13\3\13\3\13\3\13\3\13\5\13\u019f\n\13\3\13\3\13\3")
        buf.write("\13\5\13\u01a4\n\13\3\13\3\13\5\13\u01a8\n\13\3\13\3\13")
        buf.write("\5\13\u01ac\n\13\3\f\3\f\3\f\3\r\3\r\3\r\3\r\3\16\3\16")
        buf.write("\3\16\3\16\3\17\3\17\3\17\3\17\3\20\3\20\7\20\u01bf\n")
        buf.write("\20\f\20\16\20\u01c2\13\20\3\20\3\20\3\20\3\20\5\20\u01c8")
        buf.write("\n\20\3\21\3\21\5\21\u01cc\n\21\3\22\3\22\7\22\u01d0\n")
        buf.write("\22\f\22\16\22\u01d3\13\22\3\22\3\22\3\22\3\22\5\22\u01d9")
        buf.write("\n\22\3\23\7\23\u01dc\n\23\f\23\16\23\u01df\13\23\3\24")
        buf.write("\3\24\5\24\u01e3\n\24\3\25\3\25\3\25\3\25\3\25\3\25\3")
        buf.write("\25\3\25\3\25\3\25\3\25\3\25\3\25\3\25\5\25\u01f3\n\25")
        buf.write("\3\26\3\26\3\26\3\26\3\26\3\26\3\26\3\26\3\26\3\26\5\26")
        buf.write("\u01ff\n\26\3\27\3\27\3\27\3\27\3\27\3\30\3\30\3\30\3")
        buf.write("\30\7\30\u020a\n\30\f\30\16\30\u020d\13\30\3\30\3\30\3")
        buf.write("\31\5\31\u0212\n\31\3\31\5\31\u0215\n\31\3\31\3\31\3\31")
        buf.write("\3\31\5\31\u021b\n\31\3\32\3\32\3\32\3\32\3\32\5\32\u0222")
        buf.write("\n\32\3\32\3\32\3\32\3\32\3\32\5\32\u0229\n\32\7\32\u022b")
        buf.write("\n\32\f\32\16\32\u022e\13\32\3\32\3\32\3\32\3\32\3\32")
        buf.write("\5\32\u0235\n\32\3\32\5\32\u0238\n\32\3\33\3\33\5\33\u023c")
        buf.write("\n\33\3\33\3\33\3\34\3\34\5\34\u0242\n\34\3\34\3\34\3")
        buf.write("\34\5\34\u0247\n\34\3\35\3\35\3\35\3\35\3\36\3\36\3\36")
        buf.write("\7\36\u0250\n\36\f\36\16\36\u0253\13\36\3\37\3\37\3\37")
        buf.write("\7\37\u0258\n\37\f\37\16\37\u025b\13\37\3 \3 \7 \u025f")
        buf.write('\n \f \16 \u0262\13 \3 \3 \3 \5 \u0267\n \3!\3!\3!\3"')
        buf.write('\3"\3"\3"\3"\3"\3"\3"\3"\3"\3"\3"\3"\3"\3')
        buf.write('"\3"\3"\3"\5"\u027e\n"\3#\3#\3#\3#\7#\u0284\n#\f')
        buf.write("#\16#\u0287\13#\3#\5#\u028a\n#\3$\3$\3$\5$\u028f\n$\3")
        buf.write("%\3%\3%\3%\5%\u0295\n%\3%\3%\3%\3%\5%\u029b\n%\3%\5%\u029e")
        buf.write("\n%\3&\3&\3&\3&\3'\3'\3'\3(\3(\3(\3(\3(\3(\3(\3(\3")
        buf.write("(\3(\3(\3(\5(\u02b3\n(\3(\3(\3(\3(\5(\u02b9\n(\3)\3)\3")
        buf.write(")\3)\3*\3*\3+\3+\3+\3,\3,\3,\3,\3-\3-\3-\3-\3-\3-\3-\3")
        buf.write("-\3-\3-\3-\5-\u02d3\n-\3.\3.\3.\3.\5.\u02d9\n.\3/\3/\3")
        buf.write("/\3/\3\60\3\60\3\60\5\60\u02e2\n\60\3\60\3\60\3\60\5\60")
        buf.write("\u02e7\n\60\3\61\3\61\3\61\5\61\u02ec\n\61\3\62\3\62\3")
        buf.write("\62\3\62\3\62\5\62\u02f3\n\62\3\63\3\63\3\63\3\63\5\63")
        buf.write("\u02f9\n\63\3\64\3\64\3\64\3\65\3\65\3\65\3\66\3\66\3")
        buf.write("\66\3\67\3\67\5\67\u0306\n\67\3\67\3\67\3\67\5\67\u030b")
        buf.write("\n\67\38\38\38\78\u0310\n8\f8\168\u0313\138\39\39\39\3")
        buf.write("9\39\39\79\u031b\n9\f9\169\u031e\139\59\u0320\n9\3:\3")
        buf.write(":\3:\3:\3:\3:\3:\3:\5:\u032a\n:\3;\3;\3;\3<\3<\3<\7<\u0332")
        buf.write("\n<\f<\16<\u0335\13<\3=\3=\3=\7=\u033a\n=\f=\16=\u033d")
        buf.write("\13=\3>\3>\3>\5>\u0342\n>\3?\3?\3?\7?\u0347\n?\f?\16?")
        buf.write("\u034a\13?\3@\3@\3@\3@\5@\u0350\n@\3A\3A\3A\3A\3A\3A\3")
        buf.write("A\3A\3A\3A\3A\5A\u035d\nA\3A\3A\3A\3A\3A\3A\3A\3A\3A\5")
        buf.write("A\u0368\nA\3A\3A\3A\3A\3A\3A\5A\u0370\nA\3A\3A\6A\u0374")
        buf.write("\nA\rA\16A\u0375\7A\u0378\nA\fA\16A\u037b\13A\3B\3B\3")
        buf.write("B\3B\3B\3B\3B\5B\u0384\nB\3B\3B\3B\3B\5B\u038a\nB\5B\u038c")
        buf.write("\nB\3C\3C\3C\3C\5C\u0392\nC\3D\3D\3D\3E\3E\3E\3F\3F\3")
        buf.write("F\3F\3F\3F\5F\u03a0\nF\3G\3G\3H\3H\3I\3I\3I\3I\3I\3I\3")
        buf.write("I\3I\5I\u03ae\nI\3I\3I\3I\3I\3I\5I\u03b5\nI\5I\u03b7\n")
        buf.write("I\3J\3J\3J\3J\5J\u03bd\nJ\3J\3J\3J\3J\3J\5J\u03c4\nJ\5")
        buf.write("J\u03c6\nJ\3K\3K\3K\5K\u03cb\nK\3K\3K\3K\3K\3K\5K\u03d2")
        buf.write("\nK\3L\3L\5L\u03d6\nL\3L\3L\3L\5L\u03db\nL\5L\u03dd\n")
        buf.write("L\5L\u03df\nL\3M\3M\3M\3M\3N\3N\3N\3N\3O\3O\3O\3O\3P\3")
        buf.write("P\3P\3P\3Q\3Q\3Q\5Q\u03f4\nQ\3R\3R\3R\3R\3R\3R\5R\u03fc")
        buf.write("\nR\3R\5R\u03ff\nR\3R\5R\u0402\nR\3S\3S\3S\3S\3S\3S\5")
        buf.write("S\u040a\nS\3S\5S\u040d\nS\3S\5S\u0410\nS\3T\3T\3T\3T\3")
        buf.write("T\3T\5T\u0418\nT\3T\5T\u041b\nT\3T\5T\u041e\nT\3U\3U\5")
        buf.write("U\u0422\nU\3U\3U\3V\3V\3V\3V\3V\3V\3V\3V\3V\3V\5V\u0430")
        buf.write("\nV\3W\3W\3W\3W\7W\u0436\nW\fW\16W\u0439\13W\5W\u043b")
        buf.write("\nW\3W\3W\3X\3X\3X\3X\3Y\3Y\3Y\3Z\3Z\3Z\5Z\u0449\nZ\3")
        buf.write("[\3[\3[\3\\\5\\\u044f\n\\\3\\\3\\\5\\\u0453\n\\\3]\5]")
        buf.write("\u0456\n]\3]\3]\3^\3^\5^\u045c\n^\3^\3^\5^\u0460\n^\3")
        buf.write("_\3_\3_\3_\7_\u0466\n_\f_\16_\u0469\13_\5_\u046b\n_\3")
        buf.write("_\3_\3`\3`\3`\3`\7`\u0473\n`\f`\16`\u0476\13`\5`\u0478")
        buf.write("\n`\3`\3`\3a\3a\3a\3a\3b\3b\3b\3b\3c\3c\3d\5d\u0487\n")
        buf.write("d\3d\3d\5d\u048b\nd\3d\3d\3d\3d\3e\3e\5e\u0493\ne\7e\u0495")
        buf.write("\ne\fe\16e\u0498\13e\3f\3f\3f\3f\3f\3f\3f\3f\5f\u04a2")
        buf.write("\nf\3g\3g\3g\3h\3h\5h\u04a9\nh\3h\6h\u04ac\nh\rh\16h\u04ad")
        buf.write("\3i\3i\3i\5i\u04b3\ni\3i\5i\u04b6\ni\6i\u04b8\ni\ri\16")
        buf.write("i\u04b9\3j\3j\5j\u04be\nj\3j\3j\5j\u04c2\nj\3k\3k\3k\5")
        buf.write("k\u04c7\nk\6k\u04c9\nk\rk\16k\u04ca\3l\3l\3m\3m\5m\u04d1")
        buf.write("\nm\3n\3n\5n\u04d5\nn\3o\3o\3o\3o\5o\u04db\no\3p\3p\5")
        buf.write("p\u04df\np\5p\u04e1\np\3p\3p\3p\3p\3q\3q\3q\2\3\u0080")
        buf.write('r\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 "$&(*,.\60\62')
        buf.write("\64\668:<>@BDFHJLNPRTVXZ\\^`bdfhjlnprtvxz|~\u0080\u0082")
        buf.write("\u0084\u0086\u0088\u008a\u008c\u008e\u0090\u0092\u0094")
        buf.write("\u0096\u0098\u009a\u009c\u009e\u00a0\u00a2\u00a4\u00a6")
        buf.write("\u00a8\u00aa\u00ac\u00ae\u00b0\u00b2\u00b4\u00b6\u00b8")
        buf.write("\u00ba\u00bc\u00be\u00c0\u00c2\u00c4\u00c6\u00c8\u00ca")
        buf.write("\u00cc\u00ce\u00d0\u00d2\u00d4\u00d6\u00d8\u00da\u00dc")
        buf.write('\u00de\u00e0\2\17\4\2\33\34""\4\2BCFF\3\2\64\67\3\2')
        buf.write("9:\3\2YZ\3\2[]\3\2\37!\4\2\36\36hh\5\2\13\21\26\26,,\4")
        buf.write("\2\22\22\24\24\5\2\22\22\26\26,,\4\2\6\6\n\n\b\2\22\22")
        buf.write("\26\26,,cdffhh\2\u0551\2\u00e3\3\2\2\2\4\u00f3\3\2\2\2")
        buf.write("\6\u00fe\3\2\2\2\b\u013a\3\2\2\2\n\u0143\3\2\2\2\f\u0148")
        buf.write("\3\2\2\2\16\u014a\3\2\2\2\20\u0180\3\2\2\2\22\u0182\3")
        buf.write("\2\2\2\24\u0198\3\2\2\2\26\u01ad\3\2\2\2\30\u01b0\3\2")
        buf.write("\2\2\32\u01b4\3\2\2\2\34\u01b8\3\2\2\2\36\u01c7\3\2\2")
        buf.write('\2 \u01cb\3\2\2\2"\u01d8\3\2\2\2$\u01dd\3\2\2\2&\u01e2')
        buf.write("\3\2\2\2(\u01f2\3\2\2\2*\u01fe\3\2\2\2,\u0200\3\2\2\2")
        buf.write(".\u0205\3\2\2\2\60\u0211\3\2\2\2\62\u0237\3\2\2\2\64\u0239")
        buf.write("\3\2\2\2\66\u023f\3\2\2\28\u0248\3\2\2\2:\u024c\3\2\2")
        buf.write("\2<\u0254\3\2\2\2>\u0266\3\2\2\2@\u0268\3\2\2\2B\u027d")
        buf.write("\3\2\2\2D\u027f\3\2\2\2F\u028b\3\2\2\2H\u029d\3\2\2\2")
        buf.write("J\u029f\3\2\2\2L\u02a3\3\2\2\2N\u02b8\3\2\2\2P\u02ba\3")
        buf.write("\2\2\2R\u02be\3\2\2\2T\u02c0\3\2\2\2V\u02c3\3\2\2\2X\u02d2")
        buf.write("\3\2\2\2Z\u02d8\3\2\2\2\\\u02da\3\2\2\2^\u02de\3\2\2\2")
        buf.write("`\u02e8\3\2\2\2b\u02ed\3\2\2\2d\u02f4\3\2\2\2f\u02fa\3")
        buf.write("\2\2\2h\u02fd\3\2\2\2j\u0300\3\2\2\2l\u0303\3\2\2\2n\u030c")
        buf.write("\3\2\2\2p\u031f\3\2\2\2r\u0329\3\2\2\2t\u032b\3\2\2\2")
        buf.write("v\u032e\3\2\2\2x\u0336\3\2\2\2z\u0341\3\2\2\2|\u0343\3")
        buf.write("\2\2\2~\u034b\3\2\2\2\u0080\u036f\3\2\2\2\u0082\u038b")
        buf.write("\3\2\2\2\u0084\u0391\3\2\2\2\u0086\u0393\3\2\2\2\u0088")
        buf.write("\u0396\3\2\2\2\u008a\u039f\3\2\2\2\u008c\u03a1\3\2\2\2")
        buf.write("\u008e\u03a3\3\2\2\2\u0090\u03b6\3\2\2\2\u0092\u03c5\3")
        buf.write("\2\2\2\u0094\u03ca\3\2\2\2\u0096\u03de\3\2\2\2\u0098\u03e0")
        buf.write("\3\2\2\2\u009a\u03e4\3\2\2\2\u009c\u03e8\3\2\2\2\u009e")
        buf.write("\u03ec\3\2\2\2\u00a0\u03f3\3\2\2\2\u00a2\u0401\3\2\2\2")
        buf.write("\u00a4\u040f\3\2\2\2\u00a6\u041d\3\2\2\2\u00a8\u041f\3")
        buf.write("\2\2\2\u00aa\u042f\3\2\2\2\u00ac\u0431\3\2\2\2\u00ae\u043e")
        buf.write("\3\2\2\2\u00b0\u0442\3\2\2\2\u00b2\u0448\3\2\2\2\u00b4")
        buf.write("\u044a\3\2\2\2\u00b6\u044e\3\2\2\2\u00b8\u0455\3\2\2\2")
        buf.write("\u00ba\u0459\3\2\2\2\u00bc\u0461\3\2\2\2\u00be\u046e\3")
        buf.write("\2\2\2\u00c0\u047b\3\2\2\2\u00c2\u047f\3\2\2\2\u00c4\u0483")
        buf.write("\3\2\2\2\u00c6\u0486\3\2\2\2\u00c8\u0496\3\2\2\2\u00ca")
        buf.write("\u04a1\3\2\2\2\u00cc\u04a3\3\2\2\2\u00ce\u04ab\3\2\2\2")
        buf.write("\u00d0\u04b7\3\2\2\2\u00d2\u04bd\3\2\2\2\u00d4\u04c8\3")
        buf.write("\2\2\2\u00d6\u04cc\3\2\2\2\u00d8\u04ce\3\2\2\2\u00da\u04d2")
        buf.write("\3\2\2\2\u00dc\u04d6\3\2\2\2\u00de\u04e0\3\2\2\2\u00e0")
        buf.write("\u04e6\3\2\2\2\u00e2\u00e4\5\6\4\2\u00e3\u00e2\3\2\2\2")
        buf.write("\u00e3\u00e4\3\2\2\2\u00e4\u00e8\3\2\2\2\u00e5\u00e7\5")
        buf.write("\4\3\2\u00e6\u00e5\3\2\2\2\u00e7\u00ea\3\2\2\2\u00e8\u00e6")
        buf.write("\3\2\2\2\u00e8\u00e9\3\2\2\2\u00e9\u00ee\3\2\2\2\u00ea")
        buf.write("\u00e8\3\2\2\2\u00eb\u00ed\5\f\7\2\u00ec\u00eb\3\2\2\2")
        buf.write("\u00ed\u00f0\3\2\2\2\u00ee\u00ec\3\2\2\2\u00ee\u00ef\3")
        buf.write("\2\2\2\u00ef\u00f1\3\2\2\2\u00f0\u00ee\3\2\2\2\u00f1\u00f2")
        buf.write("\7\2\2\3\u00f2\3\3\2\2\2\u00f3\u00f4\7#\2\2\u00f4\u00f7")
        buf.write("\7*\2\2\u00f5\u00f8\5\b\5\2\u00f6\u00f8\7[\2\2\u00f7\u00f5")
        buf.write("\3\2\2\2\u00f7\u00f6\3\2\2\2\u00f8\u00f9\3\2\2\2\u00f9")
        buf.write("\u00fa\7+\2\2\u00fa\u00fb\7\32\2\2\u00fb\u00fc\7d\2\2")
        buf.write("\u00fc\u00fd\7\62\2\2\u00fd\5\3\2\2\2\u00fe\u00ff\7\3")
        buf.write("\2\2\u00ff\u0100\7$\2\2\u0100\u0102\7d\2\2\u0101\u0103")
        buf.write("\7\62\2\2\u0102\u0101\3\2\2\2\u0102\u0103\3\2\2\2\u0103")
        buf.write("\7\3\2\2\2\u0104\u0107\7-\2\2\u0105\u0108\7[\2\2\u0106")
        buf.write("\u0108\5\n\6\2\u0107\u0105\3\2\2\2\u0107\u0106\3\2\2\2")
        buf.write("\u0108\u010b\3\2\2\2\u0109\u010a\7W\2\2\u010a\u010c\5")
        buf.write("\b\5\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c\u013b")
        buf.write("\3\2\2\2\u010d\u0110\7\26\2\2\u010e\u0111\7[\2\2\u010f")
        buf.write("\u0111\5\n\6\2\u0110\u010e\3\2\2\2\u0110\u010f\3\2\2\2")
        buf.write("\u0111\u0114\3\2\2\2\u0112\u0113\7W\2\2\u0113\u0115\5")
        buf.write("\b\5\2\u0114\u0112\3\2\2\2\u0114\u0115\3\2\2\2\u0115\u013b")
        buf.write("\3\2\2\2\u0116\u0119\7,\2\2\u0117\u011a\7[\2\2\u0118\u011a")
        buf.write("\5\n\6\2\u0119\u0117\3\2\2\2\u0119\u0118\3\2\2\2\u011a")
        buf.write("\u011d\3\2\2\2\u011b\u011c\7W\2\2\u011c\u011e\5\b\5\2")
        buf.write("\u011d\u011b\3\2\2\2\u011d\u011e\3\2\2\2\u011e\u013b\3")
        buf.write("\2\2\2\u011f\u0122\7\22\2\2\u0120\u0123\7[\2\2\u0121\u0123")
        buf.write("\5\n\6\2\u0122\u0120\3\2\2\2\u0122\u0121\3\2\2\2\u0123")
        buf.write("\u0126\3\2\2\2\u0124\u0125\7W\2\2\u0125\u0127\5\b\5\2")
        buf.write("\u0126\u0124\3\2\2\2\u0126\u0127\3\2\2\2\u0127\u013b\3")
        buf.write("\2\2\2\u0128\u012b\7U\2\2\u0129\u012c\7[\2\2\u012a\u012c")
        buf.write("\5\n\6\2\u012b\u0129\3\2\2\2\u012b\u012a\3\2\2\2\u012c")
        buf.write("\u012f\3\2\2\2\u012d\u012e\7W\2\2\u012e\u0130\5\b\5\2")
        buf.write("\u012f\u012d\3\2\2\2\u012f\u0130\3\2\2\2\u0130\u013b\3")
        buf.write("\2\2\2\u0131\u0134\7\21\2\2\u0132\u0135\7[\2\2\u0133\u0135")
        buf.write("\5\n\6\2\u0134\u0132\3\2\2\2\u0134\u0133\3\2\2\2\u0135")
        buf.write("\u0138\3\2\2\2\u0136\u0137\7W\2\2\u0137\u0139\5\b\5\2")
        buf.write("\u0138\u0136\3\2\2\2\u0138\u0139\3\2\2\2\u0139\u013b\3")
        buf.write("\2\2\2\u013a\u0104\3\2\2\2\u013a\u010d\3\2\2\2\u013a\u0116")
        buf.write("\3\2\2\2\u013a\u011f\3\2\2\2\u013a\u0128\3\2\2\2\u013a")
        buf.write("\u0131\3\2\2\2\u013b\t\3\2\2\2\u013c\u013d\7%\2\2\u013d")
        buf.write("\u0144\7h\2\2\u013e\u013f\7%\2\2\u013f\u0140\7*\2\2\u0140")
        buf.write("\u0141\5:\36\2\u0141\u0142\7+\2\2\u0142\u0144\3\2\2\2")
        buf.write("\u0143\u013c\3\2\2\2\u0143\u013e\3\2\2\2\u0144\13\3\2")
        buf.write("\2\2\u0145\u0149\5\16\b\2\u0146\u0149\5\20\t\2\u0147\u0149")
        buf.write("\5\24\13\2\u0148\u0145\3\2\2\2\u0148\u0146\3\2\2\2\u0148")
        buf.write("\u0147\3\2\2\2\u0149\r\3\2\2\2\u014a\u014b\7U\2\2\u014b")
        buf.write("\u014c\7h\2\2\u014c\u014d\7\63\2\2\u014d\u0154\5d\63\2")
        buf.write("\u014e\u014f\7W\2\2\u014f\u0150\7h\2\2\u0150\u0151\7\63")
        buf.write("\2\2\u0151\u0153\5d\63\2\u0152\u014e\3\2\2\2\u0153\u0156")
        buf.write("\3\2\2\2\u0154\u0152\3\2\2\2\u0154\u0155\3\2\2\2\u0155")
        buf.write("\u0157\3\2\2\2\u0156\u0154\3\2\2\2\u0157\u0158\7\62\2")
        buf.write("\2\u0158\17\3\2\2\2\u0159\u015a\7\26\2\2\u015a\u015f\7")
        buf.write("h\2\2\u015b\u015c\7$\2\2\u015c\u015e\7h\2\2\u015d\u015b")
        buf.write("\3\2\2\2\u015e\u0161\3\2\2\2\u015f\u015d\3\2\2\2\u015f")
        buf.write("\u0160\3\2\2\2\u0160\u0164\3\2\2\2\u0161\u015f\3\2\2\2")
        buf.write("\u0162\u0163\7$\2\2\u0163\u0165\7f\2\2\u0164\u0162\3\2")
        buf.write("\2\2\u0164\u0165\3\2\2\2\u0165\u0166\3\2\2\2\u0166\u0181")
        buf.write("\5\36\20\2\u0167\u0168\7,\2\2\u0168\u016d\7h\2\2\u0169")
        buf.write("\u016a\7$\2\2\u016a\u016c\7h\2\2\u016b\u0169\3\2\2\2\u016c")
        buf.write("\u016f\3\2\2\2\u016d\u016b\3\2\2\2\u016d\u016e\3\2\2\2")
        buf.write("\u016e\u0170\3\2\2\2\u016f\u016d\3\2\2\2\u0170\u0181\5")
        buf.write("\36\20\2\u0171\u0172\7\21\2\2\u0172\u0173\7h\2\2\u0173")
        buf.write('\u0181\5"\22\2\u0174\u0175\7\22\2\2\u0175\u0176\7h\2')
        buf.write("\2\u0176\u0181\5&\24\2\u0177\u0179\7.\2\2\u0178\u0177")
        buf.write("\3\2\2\2\u0178\u0179\3\2\2\2\u0179\u017a\3\2\2\2\u017a")
        buf.write("\u017b\7-\2\2\u017b\u017d\7h\2\2\u017c\u017e\5\26\f\2")
        buf.write("\u017d\u017c\3\2\2\2\u017d\u017e\3\2\2\2\u017e\u017f\3")
        buf.write("\2\2\2\u017f\u0181\5\22\n\2\u0180\u0159\3\2\2\2\u0180")
        buf.write("\u0167\3\2\2\2\u0180\u0171\3\2\2\2\u0180\u0174\3\2\2\2")
        buf.write("\u0180\u0178\3\2\2\2\u0181\21\3\2\2\2\u0182\u0186\7*\2")
        buf.write("\2\u0183\u0185\5 \21\2\u0184\u0183\3\2\2\2\u0185\u0188")
        buf.write("\3\2\2\2\u0186\u0184\3\2\2\2\u0186\u0187\3\2\2\2\u0187")
        buf.write("\u018a\3\2\2\2\u0188\u0186\3\2\2\2\u0189\u018b\5\30\r")
        buf.write("\2\u018a\u0189\3\2\2\2\u018a\u018b\3\2\2\2\u018b\u0190")
        buf.write('\3\2\2\2\u018c\u018f\5B"\2\u018d\u018f\5\34\17\2\u018e')
        buf.write("\u018c\3\2\2\2\u018e\u018d\3\2\2\2\u018f\u0192\3\2\2\2")
        buf.write("\u0190\u018e\3\2\2\2\u0190\u0191\3\2\2\2\u0191\u0194\3")
        buf.write("\2\2\2\u0192\u0190\3\2\2\2\u0193\u0195\5\32\16\2\u0194")
        buf.write("\u0193\3\2\2\2\u0194\u0195\3\2\2\2\u0195\u0196\3\2\2\2")
        buf.write("\u0196\u0197\7+\2\2\u0197\23\3\2\2\2\u0198\u0199\7\60")
        buf.write("\2\2\u0199\u019a\7d\2\2\u019a\u019e\7\32\2\2\u019b\u019f")
        buf.write("\5\u009cO\2\u019c\u019d\7\22\2\2\u019d\u019f\5&\24\2\u019e")
        buf.write("\u019b\3\2\2\2\u019e\u019c\3\2\2\2\u019f\u01a0\3\2\2\2")
        buf.write("\u01a0\u01ab\7@\2\2\u01a1\u01a3\5\u009aN\2\u01a2\u01a4")
        buf.write("\5\u00bc_\2\u01a3\u01a2\3\2\2\2\u01a3\u01a4\3\2\2\2\u01a4")
        buf.write("\u01a7\3\2\2\2\u01a5\u01a8\5> \2\u01a6\u01a8\7\62\2\2")
        buf.write("\u01a7\u01a5\3\2\2\2\u01a7\u01a6\3\2\2\2\u01a8\u01ac\3")
        buf.write("\2\2\2\u01a9\u01aa\7-\2\2\u01aa\u01ac\5\22\n\2\u01ab\u01a1")
        buf.write("\3\2\2\2\u01ab\u01a9\3\2\2\2\u01ac\25\3\2\2\2\u01ad\u01ae")
        buf.write("\7$\2\2\u01ae\u01af\5:\36\2\u01af\27\3\2\2\2\u01b0\u01b1")
        buf.write("\7\32\2\2\u01b1\u01b2\7\33\2\2\u01b2\u01b3\5> \2\u01b3")
        buf.write("\31\3\2\2\2\u01b4\u01b5\7\32\2\2\u01b5\u01b6\7\34\2\2")
        buf.write("\u01b6\u01b7\5> \2\u01b7\33\3\2\2\2\u01b8\u01b9\7\32\2")
        buf.write('\2\u01b9\u01ba\7"\2\2\u01ba\u01bb\5> \2\u01bb\35\3\2')
        buf.write("\2\2\u01bc\u01c0\7*\2\2\u01bd\u01bf\5 \21\2\u01be\u01bd")
        buf.write("\3\2\2\2\u01bf\u01c2\3\2\2\2\u01c0\u01be\3\2\2\2\u01c0")
        buf.write("\u01c1\3\2\2\2\u01c1\u01c3\3\2\2\2\u01c2\u01c0\3\2\2\2")
        buf.write("\u01c3\u01c8\7+\2\2\u01c4\u01c5\7$\2\2\u01c5\u01c8\5 ")
        buf.write("\21\2\u01c6\u01c8\7\62\2\2\u01c7\u01bc\3\2\2\2\u01c7\u01c4")
        buf.write("\3\2\2\2\u01c7\u01c6\3\2\2\2\u01c8\37\3\2\2\2\u01c9\u01cc")
        buf.write("\5.\30\2\u01ca\u01cc\5\62\32\2\u01cb\u01c9\3\2\2\2\u01cb")
        buf.write("\u01ca\3\2\2\2\u01cc!\3\2\2\2\u01cd\u01d1\7*\2\2\u01ce")
        buf.write("\u01d0\5.\30\2\u01cf\u01ce\3\2\2\2\u01d0\u01d3\3\2\2\2")
        buf.write("\u01d1\u01cf\3\2\2\2\u01d1\u01d2\3\2\2\2\u01d2\u01d4\3")
        buf.write("\2\2\2\u01d3\u01d1\3\2\2\2\u01d4\u01d9\7+\2\2\u01d5\u01d6")
        buf.write("\7$\2\2\u01d6\u01d9\5.\30\2\u01d7\u01d9\7\62\2\2\u01d8")
        buf.write("\u01cd\3\2\2\2\u01d8\u01d5\3\2\2\2\u01d8\u01d7\3\2\2\2")
        buf.write("\u01d9#\3\2\2\2\u01da\u01dc\5\62\32\2\u01db\u01da\3\2")
        buf.write("\2\2\u01dc\u01df\3\2\2\2\u01dd\u01db\3\2\2\2\u01dd\u01de")
        buf.write("\3\2\2\2\u01de%\3\2\2\2\u01df\u01dd\3\2\2\2\u01e0\u01e3")
        buf.write("\5(\25\2\u01e1\u01e3\5*\26\2\u01e2\u01e0\3\2\2\2\u01e2")
        buf.write("\u01e1\3\2\2\2\u01e3'\3\2\2\2\u01e4\u01e5\7*\2\2\u01e5")
        buf.write("\u01e6\5,\27\2\u01e6\u01e7\5$\23\2\u01e7\u01e8\7\31\2")
        buf.write("\2\u01e8\u01e9\5> \2\u01e9\u01ea\7+\2\2\u01ea\u01f3\3")
        buf.write("\2\2\2\u01eb\u01ec\7$\2\2\u01ec\u01ed\5,\27\2\u01ed\u01ee")
        buf.write("\5$\23\2\u01ee\u01ef\7\31\2\2\u01ef\u01f0\5> \2\u01f0")
        buf.write("\u01f1\7\62\2\2\u01f1\u01f3\3\2\2\2\u01f2\u01e4\3\2\2")
        buf.write("\2\u01f2\u01eb\3\2\2\2\u01f3)\3\2\2\2\u01f4\u01f5\7*\2")
        buf.write("\2\u01f5\u01f6\5,\27\2\u01f6\u01f7\5\u00c6d\2\u01f7\u01f8")
        buf.write("\7+\2\2\u01f8\u01ff\3\2\2\2\u01f9\u01fa\7$\2\2\u01fa\u01fb")
        buf.write("\5,\27\2\u01fb\u01fc\5\u00c6d\2\u01fc\u01fd\7\62\2\2\u01fd")
        buf.write("\u01ff\3\2\2\2\u01fe\u01f4\3\2\2\2\u01fe\u01f9\3\2\2\2")
        buf.write("\u01ff+\3\2\2\2\u0200\u0201\7T\2\2\u0201\u0202\7S\2\2")
        buf.write("\u0202\u0203\7h\2\2\u0203\u0204\7\62\2\2\u0204-\3\2\2")
        buf.write("\2\u0205\u0206\7T\2\2\u0206\u020b\5\60\31\2\u0207\u0208")
        buf.write("\7W\2\2\u0208\u020a\5\60\31\2\u0209\u0207\3\2\2\2\u020a")
        buf.write("\u020d\3\2\2\2\u020b\u0209\3\2\2\2\u020b\u020c\3\2\2\2")
        buf.write("\u020c\u020e\3\2\2\2\u020d\u020b\3\2\2\2\u020e\u020f\7")
        buf.write("\62\2\2\u020f/\3\2\2\2\u0210\u0212\7V\2\2\u0211\u0210")
        buf.write("\3\2\2\2\u0211\u0212\3\2\2\2\u0212\u0214\3\2\2\2\u0213")
        buf.write("\u0215\7S\2\2\u0214\u0213\3\2\2\2\u0214\u0215\3\2\2\2")
        buf.write("\u0215\u021a\3\2\2\2\u0216\u021b\7h\2\2\u0217\u0218\7")
        buf.write("h\2\2\u0218\u0219\7\63\2\2\u0219\u021b\5d\63\2\u021a\u0216")
        buf.write("\3\2\2\2\u021a\u0217\3\2\2\2\u021b\61\3\2\2\2\u021c\u021d")
        buf.write("\7X\2\2\u021d\u0221\58\35\2\u021e\u021f\5\66\34\2\u021f")
        buf.write("\u0220\5\64\33\2\u0220\u0222\3\2\2\2\u0221\u021e\3\2\2")
        buf.write("\2\u0221\u0222\3\2\2\2\u0222\u022c\3\2\2\2\u0223\u0224")
        buf.write("\7W\2\2\u0224\u0228\58\35\2\u0225\u0226\5\66\34\2\u0226")
        buf.write("\u0227\5\64\33\2\u0227\u0229\3\2\2\2\u0228\u0225\3\2\2")
        buf.write("\2\u0228\u0229\3\2\2\2\u0229\u022b\3\2\2\2\u022a\u0223")
        buf.write("\3\2\2\2\u022b\u022e\3\2\2\2\u022c\u022a\3\2\2\2\u022c")
        buf.write("\u022d\3\2\2\2\u022d\u022f\3\2\2\2\u022e\u022c\3\2\2\2")
        buf.write("\u022f\u0230\7\62\2\2\u0230\u0238\3\2\2\2\u0231\u0232")
        buf.write("\7X\2\2\u0232\u0234\7h\2\2\u0233\u0235\5\64\33\2\u0234")
        buf.write("\u0233\3\2\2\2\u0234\u0235\3\2\2\2\u0235\u0236\3\2\2\2")
        buf.write("\u0236\u0238\5> \2\u0237\u021c\3\2\2\2\u0237\u0231\3\2")
        buf.write("\2\2\u0238\63\3\2\2\2\u0239\u023b\7\32\2\2\u023a\u023c")
        buf.write("\5:\36\2\u023b\u023a\3\2\2\2\u023b\u023c\3\2\2\2\u023c")
        buf.write("\u023d\3\2\2\2\u023d\u023e\t\2\2\2\u023e\65\3\2\2\2\u023f")
        buf.write("\u0241\7%\2\2\u0240\u0242\5<\37\2\u0241\u0240\3\2\2\2")
        buf.write("\u0241\u0242\3\2\2\2\u0242\u0246\3\2\2\2\u0243\u0247\7")
        buf.write("%\2\2\u0244\u0245\7)\2\2\u0245\u0247\5d\63\2\u0246\u0243")
        buf.write("\3\2\2\2\u0246\u0244\3\2\2\2\u0247\67\3\2\2\2\u0248\u0249")
        buf.write("\7h\2\2\u0249\u024a\7J\2\2\u024a\u024b\7h\2\2\u024b9\3")
        buf.write("\2\2\2\u024c\u0251\7h\2\2\u024d\u024e\7W\2\2\u024e\u0250")
        buf.write("\7h\2\2\u024f\u024d\3\2\2\2\u0250\u0253\3\2\2\2\u0251")
        buf.write("\u024f\3\2\2\2\u0251\u0252\3\2\2\2\u0252;\3\2\2\2\u0253")
        buf.write("\u0251\3\2\2\2\u0254\u0259\5d\63\2\u0255\u0256\7W\2\2")
        buf.write("\u0256\u0258\5d\63\2\u0257\u0255\3\2\2\2\u0258\u025b\3")
        buf.write("\2\2\2\u0259\u0257\3\2\2\2\u0259\u025a\3\2\2\2\u025a=")
        buf.write("\3\2\2\2\u025b\u0259\3\2\2\2\u025c\u0260\7*\2\2\u025d")
        buf.write('\u025f\5B"\2\u025e\u025d\3\2\2\2\u025f\u0262\3\2\2\2')
        buf.write("\u0260\u025e\3\2\2\2\u0260\u0261\3\2\2\2\u0261\u0263\3")
        buf.write("\2\2\2\u0262\u0260\3\2\2\2\u0263\u0267\7+\2\2\u0264\u0265")
        buf.write('\7$\2\2\u0265\u0267\5B"\2\u0266\u025c\3\2\2\2\u0266\u0264')
        buf.write("\3\2\2\2\u0267?\3\2\2\2\u0268\u0269\5:\36\2\u0269\u026a")
        buf.write("\5> \2\u026aA\3\2\2\2\u026b\u027e\5> \2\u026c\u027e\5")
        buf.write("@!\2\u026d\u026e\5d\63\2\u026e\u026f\7\62\2\2\u026f\u027e")
        buf.write("\3\2\2\2\u0270\u027e\5D#\2\u0271\u027e\5F$\2\u0272\u027e")
        buf.write("\5N(\2\u0273\u027e\5P)\2\u0274\u0275\5T+\2\u0275\u0276")
        buf.write("\7\62\2\2\u0276\u027e\3\2\2\2\u0277\u0278\5R*\2\u0278")
        buf.write("\u0279\7\62\2\2\u0279\u027e\3\2\2\2\u027a\u027e\5V,\2")
        buf.write("\u027b\u027e\5X-\2\u027c\u027e\5Z.\2\u027d\u026b\3\2\2")
        buf.write("\2\u027d\u026c\3\2\2\2\u027d\u026d\3\2\2\2\u027d\u0270")
        buf.write("\3\2\2\2\u027d\u0271\3\2\2\2\u027d\u0272\3\2\2\2\u027d")
        buf.write("\u0273\3\2\2\2\u027d\u0274\3\2\2\2\u027d\u0277\3\2\2\2")
        buf.write("\u027d\u027a\3\2\2\2\u027d\u027b\3\2\2\2\u027d\u027c\3")
        buf.write("\2\2\2\u027eC\3\2\2\2\u027f\u0280\7;\2\2\u0280\u0281\5")
        buf.write("d\63\2\u0281\u0285\5> \2\u0282\u0284\5J&\2\u0283\u0282")
        buf.write("\3\2\2\2\u0284\u0287\3\2\2\2\u0285\u0283\3\2\2\2\u0285")
        buf.write("\u0286\3\2\2\2\u0286\u0289\3\2\2\2\u0287\u0285\3\2\2\2")
        buf.write("\u0288\u028a\5L'\2\u0289\u0288\3\2\2\2\u0289\u028a\3")
        buf.write("\2\2\2\u028aE\3\2\2\2\u028b\u028c\7I\2\2\u028c\u028e\5")
        buf.write("> \2\u028d\u028f\5H%\2\u028e\u028d\3\2\2\2\u028e\u028f")
        buf.write("\3\2\2\2\u028fG\3\2\2\2\u0290\u0294\7=\2\2\u0291\u0292")
        buf.write("\7_\2\2\u0292\u0293\7h\2\2\u0293\u0295\7`\2\2\u0294\u0291")
        buf.write("\3\2\2\2\u0294\u0295\3\2\2\2\u0295\u0296\3\2\2\2\u0296")
        buf.write("\u029e\5> \2\u0297\u029a\7=\2\2\u0298\u0299\7\32\2\2\u0299")
        buf.write("\u029b\7h\2\2\u029a\u0298\3\2\2\2\u029a\u029b\3\2\2\2")
        buf.write("\u029b\u029c\3\2\2\2\u029c\u029e\5> \2\u029d\u0290\3\2")
        buf.write("\2\2\u029d\u0297\3\2\2\2\u029eI\3\2\2\2\u029f\u02a0\7")
        buf.write("<\2\2\u02a0\u02a1\5d\63\2\u02a1\u02a2\5> \2\u02a2K\3\2")
        buf.write("\2\2\u02a3\u02a4\7=\2\2\u02a4\u02a5\5> \2\u02a5M\3\2\2")
        buf.write("\2\u02a6\u02a7\7>\2\2\u02a7\u02a8\5d\63\2\u02a8\u02a9")
        buf.write("\7?\2\2\u02a9\u02aa\5d\63\2\u02aa\u02ab\7@\2\2\u02ab\u02ac")
        buf.write("\5d\63\2\u02ac\u02ad\5> \2\u02ad\u02b9\3\2\2\2\u02ae\u02af")
        buf.write("\7>\2\2\u02af\u02b2\7h\2\2\u02b0\u02b1\7W\2\2\u02b1\u02b3")
        buf.write("\7h\2\2\u02b2\u02b0\3\2\2\2\u02b2\u02b3\3\2\2\2\u02b3")
        buf.write("\u02b4\3\2\2\2\u02b4\u02b5\7R\2\2\u02b5\u02b6\5d\63\2")
        buf.write("\u02b6\u02b7\5> \2\u02b7\u02b9\3\2\2\2\u02b8\u02a6\3\2")
        buf.write("\2\2\u02b8\u02ae\3\2\2\2\u02b9O\3\2\2\2\u02ba\u02bb\7")
        buf.write("A\2\2\u02bb\u02bc\5d\63\2\u02bc\u02bd\5> \2\u02bdQ\3\2")
        buf.write("\2\2\u02be\u02bf\t\3\2\2\u02bfS\3\2\2\2\u02c0\u02c1\7")
        buf.write("\61\2\2\u02c1\u02c2\5d\63\2\u02c2U\3\2\2\2\u02c3\u02c4")
        buf.write("\7H\2\2\u02c4\u02c5\5d\63\2\u02c5\u02c6\7\62\2\2\u02c6")
        buf.write("W\3\2\2\2\u02c7\u02c8\7G\2\2\u02c8\u02c9\5d\63\2\u02c9")
        buf.write("\u02ca\7\62\2\2\u02ca\u02d3\3\2\2\2\u02cb\u02cc\7G\2\2")
        buf.write("\u02cc\u02cd\7$\2\2\u02cd\u02ce\7h\2\2\u02ce\u02cf\7\63")
        buf.write("\2\2\u02cf\u02d0\5d\63\2\u02d0\u02d1\7\62\2\2\u02d1\u02d3")
        buf.write("\3\2\2\2\u02d2\u02c7\3\2\2\2\u02d2\u02cb\3\2\2\2\u02d3")
        buf.write("Y\3\2\2\2\u02d4\u02d9\5\\/\2\u02d5\u02d9\5^\60\2\u02d6")
        buf.write("\u02d9\5`\61\2\u02d7\u02d9\5b\62\2\u02d8\u02d4\3\2\2\2")
        buf.write("\u02d8\u02d5\3\2\2\2\u02d8\u02d6\3\2\2\2\u02d8\u02d7\3")
        buf.write("\2\2\2\u02d9[\3\2\2\2\u02da\u02db\7\27\2\2\u02db\u02dc")
        buf.write("\5d\63\2\u02dc\u02dd\7\62\2\2\u02dd]\3\2\2\2\u02de\u02e1")
        buf.write("\7\30\2\2\u02df\u02e0\7$\2\2\u02e0\u02e2\7h\2\2\u02e1")
        buf.write("\u02df\3\2\2\2\u02e1\u02e2\3\2\2\2\u02e2\u02e3\3\2\2\2")
        buf.write("\u02e3\u02e6\5d\63\2\u02e4\u02e7\7\62\2\2\u02e5\u02e7")
        buf.write("\5L'\2\u02e6\u02e4\3\2\2\2\u02e6\u02e5\3\2\2\2\u02e7")
        buf.write("_\3\2\2\2\u02e8\u02eb\7D\2\2\u02e9\u02ec\5X-\2\u02ea\u02ec")
        buf.write("\7\62\2\2\u02eb\u02e9\3\2\2\2\u02eb\u02ea\3\2\2\2\u02ec")
        buf.write("a\3\2\2\2\u02ed\u02f2\7E\2\2\u02ee\u02f3\5X-\2\u02ef\u02f3")
        buf.write("\5`\61\2\u02f0\u02f3\5^\60\2\u02f1\u02f3\7\62\2\2\u02f2")
        buf.write("\u02ee\3\2\2\2\u02f2\u02ef\3\2\2\2\u02f2\u02f0\3\2\2\2")
        buf.write("\u02f2\u02f1\3\2\2\2\u02f3c\3\2\2\2\u02f4\u02f8\5l\67")
        buf.write("\2\u02f5\u02f9\5f\64\2\u02f6\u02f9\5h\65\2\u02f7\u02f9")
        buf.write("\5j\66\2\u02f8\u02f5\3\2\2\2\u02f8\u02f6\3\2\2\2\u02f8")
        buf.write("\u02f7\3\2\2\2\u02f8\u02f9\3\2\2\2\u02f9e\3\2\2\2\u02fa")
        buf.write("\u02fb\7\63\2\2\u02fb\u02fc\5d\63\2\u02fcg\3\2\2\2\u02fd")
        buf.write("\u02fe\78\2\2\u02fe\u02ff\5d\63\2\u02ffi\3\2\2\2\u0300")
        buf.write("\u0301\t\4\2\2\u0301\u0302\5d\63\2\u0302k\3\2\2\2\u0303")
        buf.write("\u030a\5n8\2\u0304\u0306\7K\2\2\u0305\u0304\3\2\2\2\u0305")
        buf.write("\u0306\3\2\2\2\u0306\u0307\3\2\2\2\u0307\u0308\5\u00a0")
        buf.write("Q\2\u0308\u0309\5d\63\2\u0309\u030b\3\2\2\2\u030a\u0305")
        buf.write("\3\2\2\2\u030a\u030b\3\2\2\2\u030bm\3\2\2\2\u030c\u0311")
        buf.write("\5p9\2\u030d\u030e\t\5\2\2\u030e\u0310\5p9\2\u030f\u030d")
        buf.write("\3\2\2\2\u0310\u0313\3\2\2\2\u0311\u030f\3\2\2\2\u0311")
        buf.write("\u0312\3\2\2\2\u0312o\3\2\2\2\u0313\u0311\3\2\2\2\u0314")
        buf.write("\u0315\7K\2\2\u0315\u0320\5p9\2\u0316\u031c\5v<\2\u0317")
        buf.write("\u0318\5r:\2\u0318\u0319\5v<\2\u0319\u031b\3\2\2\2\u031a")
        buf.write("\u0317\3\2\2\2\u031b\u031e\3\2\2\2\u031c\u031a\3\2\2\2")
        buf.write("\u031c\u031d\3\2\2\2\u031d\u0320\3\2\2\2\u031e\u031c\3")
        buf.write("\2\2\2\u031f\u0314\3\2\2\2\u031f\u0316\3\2\2\2\u0320q")
        buf.write("\3\2\2\2\u0321\u032a\7L\2\2\u0322\u032a\7M\2\2\u0323\u032a")
        buf.write("\7N\2\2\u0324\u032a\7O\2\2\u0325\u032a\7P\2\2\u0326\u032a")
        buf.write("\7Q\2\2\u0327\u032a\7R\2\2\u0328\u032a\5t;\2\u0329\u0321")
        buf.write("\3\2\2\2\u0329\u0322\3\2\2\2\u0329\u0323\3\2\2\2\u0329")
        buf.write("\u0324\3\2\2\2\u0329\u0325\3\2\2\2\u0329\u0326\3\2\2\2")
        buf.write("\u0329\u0327\3\2\2\2\u0329\u0328\3\2\2\2\u032as\3\2\2")
        buf.write("\2\u032b\u032c\7K\2\2\u032c\u032d\7R\2\2\u032du\3\2\2")
        buf.write("\2\u032e\u0333\5x=\2\u032f\u0330\t\6\2\2\u0330\u0332\5")
        buf.write("x=\2\u0331\u032f\3\2\2\2\u0332\u0335\3\2\2\2\u0333\u0331")
        buf.write("\3\2\2\2\u0333\u0334\3\2\2\2\u0334w\3\2\2\2\u0335\u0333")
        buf.write("\3\2\2\2\u0336\u033b\5z>\2\u0337\u0338\t\7\2\2\u0338\u033a")
        buf.write("\5z>\2\u0339\u0337\3\2\2\2\u033a\u033d\3\2\2\2\u033b\u0339")
        buf.write("\3\2\2\2\u033b\u033c\3\2\2\2\u033cy\3\2\2\2\u033d\u033b")
        buf.write("\3\2\2\2\u033e\u033f\t\6\2\2\u033f\u0342\5z>\2\u0340\u0342")
        buf.write("\5|?\2\u0341\u033e\3\2\2\2\u0341\u0340\3\2\2\2\u0342{")
        buf.write("\3\2\2\2\u0343\u0348\5\u0080A\2\u0344\u0345\7^\2\2\u0345")
        buf.write("\u0347\5z>\2\u0346\u0344\3\2\2\2\u0347\u034a\3\2\2\2\u0348")
        buf.write("\u0346\3\2\2\2\u0348\u0349\3\2\2\2\u0349}\3\2\2\2\u034a")
        buf.write("\u0348\3\2\2\2\u034b\u034c\7U\2\2\u034c\u034f\7J\2\2\u034d")
        buf.write("\u0350\5\u008eH\2\u034e\u0350\7h\2\2\u034f\u034d\3\2\2")
        buf.write("\2\u034f\u034e\3\2\2\2\u0350\177\3\2\2\2\u0351\u0352\b")
        buf.write("A\1\2\u0352\u0370\7f\2\2\u0353\u0370\7c\2\2\u0354\u0370")
        buf.write("\7d\2\2\u0355\u0370\7e\2\2\u0356\u0370\7g\2\2\u0357\u0370")
        buf.write("\7h\2\2\u0358\u0370\5~@\2\u0359\u0370\5\u0096L\2\u035a")
        buf.write("\u035c\5\u009eP\2\u035b\u035d\5\u00bc_\2\u035c\u035b\3")
        buf.write("\2\2\2\u035c\u035d\3\2\2\2\u035d\u0370\3\2\2\2\u035e\u0370")
        buf.write("\5\u00a8U\2\u035f\u0370\5\u00acW\2\u0360\u0361\7_\2\2")
        buf.write("\u0361\u0362\5d\63\2\u0362\u0363\7`\2\2\u0363\u0370\3")
        buf.write("\2\2\2\u0364\u0365\5\u0084C\2\u0365\u0367\7h\2\2\u0366")
        buf.write("\u0368\5\u00bc_\2\u0367\u0366\3\2\2\2\u0367\u0368\3\2")
        buf.write("\2\2\u0368\u0370\3\2\2\2\u0369\u036a\7/\2\2\u036a\u0370")
        buf.write("\5\u0080A\7\u036b\u0370\5\u00b0Y\2\u036c\u0370\5\u0086")
        buf.write("D\2\u036d\u0370\5\u0088E\2\u036e\u0370\5\u00c4c\2\u036f")
        buf.write("\u0351\3\2\2\2\u036f\u0353\3\2\2\2\u036f\u0354\3\2\2\2")
        buf.write("\u036f\u0355\3\2\2\2\u036f\u0356\3\2\2\2\u036f\u0357\3")
        buf.write("\2\2\2\u036f\u0358\3\2\2\2\u036f\u0359\3\2\2\2\u036f\u035a")
        buf.write("\3\2\2\2\u036f\u035e\3\2\2\2\u036f\u035f\3\2\2\2\u036f")
        buf.write("\u0360\3\2\2\2\u036f\u0364\3\2\2\2\u036f\u0369\3\2\2\2")
        buf.write("\u036f\u036b\3\2\2\2\u036f\u036c\3\2\2\2\u036f\u036d\3")
        buf.write("\2\2\2\u036f\u036e\3\2\2\2\u0370\u0379\3\2\2\2\u0371\u0373")
        buf.write("\f\b\2\2\u0372\u0374\5\u0082B\2\u0373\u0372\3\2\2\2\u0374")
        buf.write("\u0375\3\2\2\2\u0375\u0373\3\2\2\2\u0375\u0376\3\2\2\2")
        buf.write("\u0376\u0378\3\2\2\2\u0377\u0371\3\2\2\2\u0378\u037b\3")
        buf.write("\2\2\2\u0379\u0377\3\2\2\2\u0379\u037a\3\2\2\2\u037a\u0081")
        buf.write("\3\2\2\2\u037b\u0379\3\2\2\2\u037c\u037d\7J\2\2\u037d")
        buf.write("\u038c\5\u008aF\2\u037e\u037f\7J\2\2\u037f\u038c\7h\2")
        buf.write("\2\u0380\u038c\5\u00aaV\2\u0381\u0383\7_\2\2\u0382\u0384")
        buf.write("\5<\37\2\u0383\u0382\3\2\2\2\u0383\u0384\3\2\2\2\u0384")
        buf.write("\u0385\3\2\2\2\u0385\u038c\7`\2\2\u0386\u0387\5\u0084")
        buf.write("C\2\u0387\u0389\7h\2\2\u0388\u038a\5\u00bc_\2\u0389\u0388")
        buf.write("\3\2\2\2\u0389\u038a\3\2\2\2\u038a\u038c\3\2\2\2\u038b")
        buf.write("\u037c\3\2\2\2\u038b\u037e\3\2\2\2\u038b\u0380\3\2\2\2")
        buf.write("\u038b\u0381\3\2\2\2\u038b\u0386\3\2\2\2\u038c\u0083\3")
        buf.write("\2\2\2\u038d\u0392\7%\2\2\u038e\u038f\7%\2\2\u038f\u0390")
        buf.write("\7h\2\2\u0390\u0392\7$\2\2\u0391\u038d\3\2\2\2\u0391\u038e")
        buf.write("\3\2\2\2\u0392\u0085\3\2\2\2\u0393\u0394\7\4\2\2\u0394")
        buf.write("\u0395\5\u0080A\2\u0395\u0087\3\2\2\2\u0396\u0397\7[\2")
        buf.write("\2\u0397\u0398\5\u0080A\2\u0398\u0089\3\2\2\2\u0399\u03a0")
        buf.write("\3\2\2\2\u039a\u03a0\5\u0094K\2\u039b\u03a0\5\u0090I\2")
        buf.write("\u039c\u03a0\5\u0092J\2\u039d\u03a0\5\u008eH\2\u039e\u03a0")
        buf.write("\5\u008cG\2\u039f\u0399\3\2\2\2\u039f\u039a\3\2\2\2\u039f")
        buf.write("\u039b\3\2\2\2\u039f\u039c\3\2\2\2\u039f\u039d\3\2\2\2")
        buf.write("\u039f\u039e\3\2\2\2\u03a0\u008b\3\2\2\2\u03a1\u03a2\5")
        buf.write("\u00c4c\2\u03a2\u008d\3\2\2\2\u03a3\u03a4\t\b\2\2\u03a4")
        buf.write("\u008f\3\2\2\2\u03a5\u03b7\7\36\2\2\u03a6\u03a7\7*\2\2")
        buf.write("\u03a7\u03a8\5:\36\2\u03a8\u03a9\7+\2\2\u03a9\u03b7\3")
        buf.write("\2\2\2\u03aa\u03ab\7\17\2\2\u03ab\u03ae\7%\2\2\u03ac\u03ae")
        buf.write("\7(\2\2\u03ad\u03aa\3\2\2\2\u03ad\u03ac\3\2\2\2\u03ae")
        buf.write("\u03af\3\2\2\2\u03af\u03b4\t\t\2\2\u03b0\u03b1\7_\2\2")
        buf.write("\u03b1\u03b2\5<\37\2\u03b2\u03b3\7`\2\2\u03b3\u03b5\3")
        buf.write("\2\2\2\u03b4\u03b0\3\2\2\2\u03b4\u03b5\3\2\2\2\u03b5\u03b7")
        buf.write("\3\2\2\2\u03b6\u03a5\3\2\2\2\u03b6\u03a6\3\2\2\2\u03b6")
        buf.write("\u03ad\3\2\2\2\u03b7\u0091\3\2\2\2\u03b8\u03c6\7\35\2")
        buf.write("\2\u03b9\u03ba\7\16\2\2\u03ba\u03bd\7%\2\2\u03bb\u03bd")
        buf.write("\7'\2\2\u03bc\u03b9\3\2\2\2\u03bc\u03bb\3\2\2\2\u03bd")
        buf.write("\u03be\3\2\2\2\u03be\u03c3\7h\2\2\u03bf\u03c0\7_\2\2\u03c0")
        buf.write("\u03c1\5<\37\2\u03c1\u03c2\7`\2\2\u03c2\u03c4\3\2\2\2")
        buf.write("\u03c3\u03bf\3\2\2\2\u03c3\u03c4\3\2\2\2\u03c4\u03c6\3")
        buf.write("\2\2\2\u03c5\u03b8\3\2\2\2\u03c5\u03bc\3\2\2\2\u03c6\u0093")
        buf.write("\3\2\2\2\u03c7\u03c8\7\13\2\2\u03c8\u03cb\7%\2\2\u03c9")
        buf.write("\u03cb\7&\2\2\u03ca\u03c7\3\2\2\2\u03ca\u03c9\3\2\2\2")
        buf.write("\u03cb\u03cc\3\2\2\2\u03cc\u03d1\7h\2\2\u03cd\u03ce\7")
        buf.write("_\2\2\u03ce\u03cf\5<\37\2\u03cf\u03d0\7`\2\2\u03d0\u03d2")
        buf.write("\3\2\2\2\u03d1\u03cd\3\2\2\2\u03d1\u03d2\3\2\2\2\u03d2")
        buf.write("\u0095\3\2\2\2\u03d3\u03d5\5\u0098M\2\u03d4\u03d6\5\u00be")
        buf.write("`\2\u03d5\u03d4\3\2\2\2\u03d5\u03d6\3\2\2\2\u03d6\u03df")
        buf.write("\3\2\2\2\u03d7\u03dc\5\u00a0Q\2\u03d8\u03da\5\u0098M\2")
        buf.write("\u03d9\u03db\5\u00be`\2\u03da\u03d9\3\2\2\2\u03da\u03db")
        buf.write("\3\2\2\2\u03db\u03dd\3\2\2\2\u03dc\u03d8\3\2\2\2\u03dc")
        buf.write("\u03dd\3\2\2\2\u03dd\u03df\3\2\2\2\u03de\u03d3\3\2\2\2")
        buf.write("\u03de\u03d7\3\2\2\2\u03df\u0097\3\2\2\2\u03e0\u03e1\7")
        buf.write("\26\2\2\u03e1\u03e2\7%\2\2\u03e2\u03e3\7h\2\2\u03e3\u0099")
        buf.write("\3\2\2\2\u03e4\u03e5\7-\2\2\u03e5\u03e6\7%\2\2\u03e6\u03e7")
        buf.write("\7h\2\2\u03e7\u009b\3\2\2\2\u03e8\u03e9\7\22\2\2\u03e9")
        buf.write("\u03ea\7%\2\2\u03ea\u03eb\7h\2\2\u03eb\u009d\3\2\2\2\u03ec")
        buf.write("\u03ed\7\21\2\2\u03ed\u03ee\7%\2\2\u03ee\u03ef\7h\2\2")
        buf.write("\u03ef\u009f\3\2\2\2\u03f0\u03f4\5\u00a2R\2\u03f1\u03f4")
        buf.write("\5\u00a4S\2\u03f2\u03f4\5\u00a6T\2\u03f3\u03f0\3\2\2\2")
        buf.write("\u03f3\u03f1\3\2\2\2\u03f3\u03f2\3\2\2\2\u03f4\u00a1\3")
        buf.write("\2\2\2\u03f5\u0402\7\5\2\2\u03f6\u03fe\7Z\2\2\u03f7\u03f8")
        buf.write("\7a\2\2\u03f8\u03fb\7h\2\2\u03f9\u03fc\5\u00bc_\2\u03fa")
        buf.write("\u03fc\5\u00be`\2\u03fb\u03f9\3\2\2\2\u03fb\u03fa\3\2")
        buf.write("\2\2\u03fb\u03fc\3\2\2\2\u03fc\u03fd\3\2\2\2\u03fd\u03ff")
        buf.write("\7b\2\2\u03fe\u03f7\3\2\2\2\u03fe\u03ff\3\2\2\2\u03ff")
        buf.write("\u0400\3\2\2\2\u0400\u0402\7\6\2\2\u0401\u03f5\3\2\2\2")
        buf.write("\u0401\u03f6\3\2\2\2\u0402\u00a3\3\2\2\2\u0403\u0410\7")
        buf.write("\7\2\2\u0404\u040c\7\b\2\2\u0405\u0406\7a\2\2\u0406\u0409")
        buf.write("\7h\2\2\u0407\u040a\5\u00bc_\2\u0408\u040a\5\u00be`\2")
        buf.write("\u0409\u0407\3\2\2\2\u0409\u0408\3\2\2\2\u0409\u040a\3")
        buf.write("\2\2\2\u040a\u040b\3\2\2\2\u040b\u040d\7b\2\2\u040c\u0405")
        buf.write("\3\2\2\2\u040c\u040d\3\2\2\2\u040d\u040e\3\2\2\2\u040e")
        buf.write("\u0410\7Z\2\2\u040f\u0403\3\2\2\2\u040f\u0404\3\2\2\2")
        buf.write("\u0410\u00a5\3\2\2\2\u0411\u041e\7\t\2\2\u0412\u041a\7")
        buf.write("\b\2\2\u0413\u0414\7a\2\2\u0414\u0417\7h\2\2\u0415\u0418")
        buf.write("\5\u00bc_\2\u0416\u0418\5\u00be`\2\u0417\u0415\3\2\2\2")
        buf.write("\u0417\u0416\3\2\2\2\u0417\u0418\3\2\2\2\u0418\u0419\3")
        buf.write("\2\2\2\u0419\u041b\7b\2\2\u041a\u0413\3\2\2\2\u041a\u041b")
        buf.write("\3\2\2\2\u041b\u041c\3\2\2\2\u041c\u041e\7\6\2\2\u041d")
        buf.write("\u0411\3\2\2\2\u041d\u0412\3\2\2\2\u041e\u00a7\3\2\2\2")
        buf.write("\u041f\u0421\7a\2\2\u0420\u0422\5<\37\2\u0421\u0420\3")
        buf.write("\2\2\2\u0421\u0422\3\2\2\2\u0422\u0423\3\2\2\2\u0423\u0424")
        buf.write("\7b\2\2\u0424\u00a9\3\2\2\2\u0425\u0426\7a\2\2\u0426\u0427")
        buf.write("\5d\63\2\u0427\u0428\7b\2\2\u0428\u0430\3\2\2\2\u0429")
        buf.write("\u042a\7a\2\2\u042a\u042b\5d\63\2\u042b\u042c\7$\2\2\u042c")
        buf.write("\u042d\5d\63\2\u042d\u042e\7b\2\2\u042e\u0430\3\2\2\2")
        buf.write("\u042f\u0425\3\2\2\2\u042f\u0429\3\2\2\2\u0430\u00ab\3")
        buf.write("\2\2\2\u0431\u043a\7*\2\2\u0432\u0437\5\u00aeX\2\u0433")
        buf.write("\u0434\7W\2\2\u0434\u0436\5\u00aeX\2\u0435\u0433\3\2\2")
        buf.write("\2\u0436\u0439\3\2\2\2\u0437\u0435\3\2\2\2\u0437\u0438")
        buf.write("\3\2\2\2\u0438\u043b\3\2\2\2\u0439\u0437\3\2\2\2\u043a")
        buf.write("\u0432\3\2\2\2\u043a\u043b\3\2\2\2\u043b\u043c\3\2\2\2")
        buf.write("\u043c\u043d\7+\2\2\u043d\u00ad\3\2\2\2\u043e\u043f\5")
        buf.write("d\63\2\u043f\u0440\7$\2\2\u0440\u0441\5d\63\2\u0441\u00af")
        buf.write("\3\2\2\2\u0442\u0443\7\31\2\2\u0443\u0444\5\u00b2Z\2\u0444")
        buf.write("\u00b1\3\2\2\2\u0445\u0449\5\u00b6\\\2\u0446\u0449\5\u00ba")
        buf.write("^\2\u0447\u0449\5\u00b8]\2\u0448\u0445\3\2\2\2\u0448\u0446")
        buf.write("\3\2\2\2\u0448\u0447\3\2\2\2\u0449\u00b3\3\2\2\2\u044a")
        buf.write("\u044b\5d\63\2\u044b\u044c\5\u00a0Q\2\u044c\u00b5\3\2")
        buf.write("\2\2\u044d\u044f\5\u00b4[\2\u044e\u044d\3\2\2\2\u044e")
        buf.write("\u044f\3\2\2\2\u044f\u0450\3\2\2\2\u0450\u0452\5\u0098")
        buf.write("M\2\u0451\u0453\5\u00bc_\2\u0452\u0451\3\2\2\2\u0452\u0453")
        buf.write("\3\2\2\2\u0453\u00b7\3\2\2\2\u0454\u0456\5\u00b4[\2\u0455")
        buf.write("\u0454\3\2\2\2\u0455\u0456\3\2\2\2\u0456\u0457\3\2\2\2")
        buf.write("\u0457\u0458\5\u009cO\2\u0458\u00b9\3\2\2\2\u0459\u045b")
        buf.write("\5d\63\2\u045a\u045c\7/\2\2\u045b\u045a\3\2\2\2\u045b")
        buf.write("\u045c\3\2\2\2\u045c\u045d\3\2\2\2\u045d\u045f\5\u009a")
        buf.write("N\2\u045e\u0460\5\u00bc_\2\u045f\u045e\3\2\2\2\u045f\u0460")
        buf.write("\3\2\2\2\u0460\u00bb\3\2\2\2\u0461\u046a\7_\2\2\u0462")
        buf.write("\u0467\5\u00c0a\2\u0463\u0464\7W\2\2\u0464\u0466\5\u00c0")
        buf.write("a\2\u0465\u0463\3\2\2\2\u0466\u0469\3\2\2\2\u0467\u0465")
        buf.write("\3\2\2\2\u0467\u0468\3\2\2\2\u0468\u046b\3\2\2\2\u0469")
        buf.write("\u0467\3\2\2\2\u046a\u0462\3\2\2\2\u046a\u046b\3\2\2\2")
        buf.write("\u046b\u046c\3\2\2\2\u046c\u046d\7`\2\2\u046d\u00bd\3")
        buf.write("\2\2\2\u046e\u0477\7_\2\2\u046f\u0474\5\u00c2b\2\u0470")
        buf.write("\u0471\7W\2\2\u0471\u0473\5\u00c2b\2\u0472\u0470\3\2\2")
        buf.write("\2\u0473\u0476\3\2\2\2\u0474\u0472\3\2\2\2\u0474\u0475")
        buf.write("\3\2\2\2\u0475\u0478\3\2\2\2\u0476\u0474\3\2\2\2\u0477")
        buf.write("\u046f\3\2\2\2\u0477\u0478\3\2\2\2\u0478\u0479\3\2\2\2")
        buf.write("\u0479\u047a\7`\2\2\u047a\u00bf\3\2\2\2\u047b\u047c\7")
        buf.write("h\2\2\u047c\u047d\7\63\2\2\u047d\u047e\5d\63\2\u047e\u00c1")
        buf.write("\3\2\2\2\u047f\u0480\7h\2\2\u0480\u0481\5r:\2\u0481\u0482")
        buf.write("\5d\63\2\u0482\u00c3\3\2\2\2\u0483\u0484\t\n\2\2\u0484")
        buf.write("\u00c5\3\2\2\2\u0485\u0487\7\23\2\2\u0486\u0485\3\2\2")
        buf.write("\2\u0486\u0487\3\2\2\2\u0487\u0488\3\2\2\2\u0488\u048a")
        buf.write("\t\13\2\2\u0489\u048b\5\u00e0q\2\u048a\u0489\3\2\2\2\u048a")
        buf.write("\u048b\3\2\2\2\u048b\u048c\3\2\2\2\u048c\u048d\7*\2\2")
        buf.write("\u048d\u048e\5\u00c8e\2\u048e\u048f\7+\2\2\u048f\u00c7")
        buf.write("\3\2\2\2\u0490\u0492\5\u00caf\2\u0491\u0493\7\62\2\2\u0492")
        buf.write("\u0491\3\2\2\2\u0492\u0493\3\2\2\2\u0493\u0495\3\2\2\2")
        buf.write("\u0494\u0490\3\2\2\2\u0495\u0498\3\2\2\2\u0496\u0494\3")
        buf.write("\2\2\2\u0496\u0497\3\2\2\2\u0497\u00c9\3\2\2\2\u0498\u0496")
        buf.write("\3\2\2\2\u0499\u04a2\5\u00d8m\2\u049a\u04a2\5\u00d2j\2")
        buf.write("\u049b\u04a2\5\u00ccg\2\u049c\u049d\5\u00e0q\2\u049d\u049e")
        buf.write("\7\63\2\2\u049e\u049f\5\u00e0q\2\u049f\u04a2\3\2\2\2\u04a0")
        buf.write("\u04a2\5\u00dep\2\u04a1\u0499\3\2\2\2\u04a1\u049a\3\2")
        buf.write("\2\2\u04a1\u049b\3\2\2\2\u04a1\u049c\3\2\2\2\u04a1\u04a0")
        buf.write("\3\2\2\2\u04a2\u00cb\3\2\2\2\u04a3\u04a4\t\f\2\2\u04a4")
        buf.write("\u04a5\5\u00ceh\2\u04a5\u00cd\3\2\2\2\u04a6\u04a8\7a\2")
        buf.write("\2\u04a7\u04a9\5\u00d0i\2\u04a8\u04a7\3\2\2\2\u04a8\u04a9")
        buf.write("\3\2\2\2\u04a9\u04aa\3\2\2\2\u04aa\u04ac\7b\2\2\u04ab")
        buf.write("\u04a6\3\2\2\2\u04ac\u04ad\3\2\2\2\u04ad\u04ab\3\2\2\2")
        buf.write("\u04ad\u04ae\3\2\2\2\u04ae\u00cf\3\2\2\2\u04af\u04b2\5")
        buf.write("\u00e0q\2\u04b0\u04b1\7\63\2\2\u04b1\u04b3\5\u00e0q\2")
        buf.write("\u04b2\u04b0\3\2\2\2\u04b2\u04b3\3\2\2\2\u04b3\u04b5\3")
        buf.write("\2\2\2\u04b4\u04b6\7W\2\2\u04b5\u04b4\3\2\2\2\u04b5\u04b6")
        buf.write("\3\2\2\2\u04b6\u04b8\3\2\2\2\u04b7\u04af\3\2\2\2\u04b8")
        buf.write("\u04b9\3\2\2\2\u04b9\u04b7\3\2\2\2\u04b9\u04ba\3\2\2\2")
        buf.write("\u04ba\u00d1\3\2\2\2\u04bb\u04be\5\u00dan\2\u04bc\u04be")
        buf.write("\5\u00dep\2\u04bd\u04bb\3\2\2\2\u04bd\u04bc\3\2\2\2\u04be")
        buf.write("\u04bf\3\2\2\2\u04bf\u04c1\5\u00d4k\2\u04c0\u04c2\5\u00ce")
        buf.write("h\2\u04c1\u04c0\3\2\2\2\u04c1\u04c2\3\2\2\2\u04c2\u00d3")
        buf.write("\3\2\2\2\u04c3\u04c6\5\u00d6l\2\u04c4\u04c7\5\u00dan\2")
        buf.write("\u04c5\u04c7\5\u00dep\2\u04c6\u04c4\3\2\2\2\u04c6\u04c5")
        buf.write("\3\2\2\2\u04c7\u04c9\3\2\2\2\u04c8\u04c3\3\2\2\2\u04c9")
        buf.write("\u04ca\3\2\2\2\u04ca\u04c8\3\2\2\2\u04ca\u04cb\3\2\2\2")
        buf.write("\u04cb\u00d5\3\2\2\2\u04cc\u04cd\t\r\2\2\u04cd\u00d7\3")
        buf.write("\2\2\2\u04ce\u04d0\5\u00dan\2\u04cf\u04d1\5\u00ceh\2\u04d0")
        buf.write("\u04cf\3\2\2\2\u04d0\u04d1\3\2\2\2\u04d1\u00d9\3\2\2\2")
        buf.write("\u04d2\u04d4\5\u00e0q\2\u04d3\u04d5\5\u00dco\2\u04d4\u04d3")
        buf.write("\3\2\2\2\u04d4\u04d5\3\2\2\2\u04d5\u00db\3\2\2\2\u04d6")
        buf.write("\u04d7\7$\2\2\u04d7\u04da\5\u00e0q\2\u04d8\u04d9\7$\2")
        buf.write("\2\u04d9\u04db\5\u00e0q\2\u04da\u04d8\3\2\2\2\u04da\u04db")
        buf.write("\3\2\2\2\u04db\u00dd\3\2\2\2\u04dc\u04de\7\25\2\2\u04dd")
        buf.write("\u04df\5\u00e0q\2\u04de\u04dd\3\2\2\2\u04de\u04df\3\2")
        buf.write("\2\2\u04df\u04e1\3\2\2\2\u04e0\u04dc\3\2\2\2\u04e0\u04e1")
        buf.write("\3\2\2\2\u04e1\u04e2\3\2\2\2\u04e2\u04e3\7*\2\2\u04e3")
        buf.write("\u04e4\5\u00c8e\2\u04e4\u04e5\7+\2\2\u04e5\u00df\3\2\2")
        buf.write("\2\u04e6\u04e7\t\16\2\2\u04e7\u00e1\3\2\2\2\u009b\u00e3")
        buf.write("\u00e8\u00ee\u00f7\u0102\u0107\u010b\u0110\u0114\u0119")
        buf.write("\u011d\u0122\u0126\u012b\u012f\u0134\u0138\u013a\u0143")
        buf.write("\u0148\u0154\u015f\u0164\u016d\u0178\u017d\u0180\u0186")
        buf.write("\u018a\u018e\u0190\u0194\u019e\u01a3\u01a7\u01ab\u01c0")
        buf.write("\u01c7\u01cb\u01d1\u01d8\u01dd\u01e2\u01f2\u01fe\u020b")
        buf.write("\u0211\u0214\u021a\u0221\u0228\u022c\u0234\u0237\u023b")
        buf.write("\u0241\u0246\u0251\u0259\u0260\u0266\u027d\u0285\u0289")
        buf.write("\u028e\u0294\u029a\u029d\u02b2\u02b8\u02d2\u02d8\u02e1")
        buf.write("\u02e6\u02eb\u02f2\u02f8\u0305\u030a\u0311\u031c\u031f")
        buf.write("\u0329\u0333\u033b\u0341\u0348\u034f\u035c\u0367\u036f")
        buf.write("\u0375\u0379\u0383\u0389\u038b\u0391\u039f\u03ad\u03b4")
        buf.write("\u03b6\u03bc\u03c3\u03c5\u03ca\u03d1\u03d5\u03da\u03dc")
        buf.write("\u03de\u03f3\u03fb\u03fe\u0401\u0409\u040c\u040f\u0417")
        buf.write("\u041a\u041d\u0421\u042f\u0437\u043a\u0448\u044e\u0452")
        buf.write("\u0455\u045b\u045f\u0467\u046a\u0474\u0477\u0486\u048a")
        buf.write("\u0492\u0496\u04a1\u04a8\u04ad\u04b2\u04b5\u04b9\u04bd")
        buf.write("\u04c1\u04c6\u04ca\u04d0\u04d4\u04da\u04de\u04e0")
        return buf.getvalue()


class jacParser(Parser):

    grammarFileName = "jac.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "'version'",
        "'&'",
        "'-->'",
        "'->'",
        "'<--'",
        "'<-'",
        "'<-->'",
        "'--'",
        "'str'",
        "'int'",
        "'float'",
        "'list'",
        "'dict'",
        "'bool'",
        "'type'",
        "'graph'",
        "'strict'",
        "'digraph'",
        "'subgraph'",
        "'node'",
        "'ignore'",
        "'take'",
        "'spawn'",
        "'with'",
        "'entry'",
        "'exit'",
        "'length'",
        "'keys'",
        "'context'",
        "'info'",
        "'details'",
        "'activity'",
        "'import'",
        "':'",
        "'::'",
        "'s::'",
        "'l::'",
        "'d::'",
        "'::>'",
        "'{'",
        "'}'",
        "'edge'",
        "'walker'",
        "'async'",
        "'sync'",
        "'test'",
        "'assert'",
        "';'",
        "'='",
        "'+='",
        "'-='",
        "'*='",
        "'/='",
        "':='",
        "<INVALID>",
        "<INVALID>",
        "'if'",
        "'elif'",
        "'else'",
        "'for'",
        "'to'",
        "'by'",
        "'while'",
        "'continue'",
        "'break'",
        "'disengage'",
        "'yield'",
        "'skip'",
        "'report'",
        "'destroy'",
        "'try'",
        "'.'",
        "<INVALID>",
        "'=='",
        "'<'",
        "'>'",
        "'<='",
        "'>='",
        "'!='",
        "'in'",
        "'anchor'",
        "'has'",
        "'global'",
        "'private'",
        "','",
        "'can'",
        "'+'",
        "'-'",
        "'*'",
        "'/'",
        "'%'",
        "'^'",
        "'('",
        "')'",
        "'['",
        "']'",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "'null'",
    ]

    symbolicNames = [
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "TYP_STRING",
        "TYP_INT",
        "TYP_FLOAT",
        "TYP_LIST",
        "TYP_DICT",
        "TYP_BOOL",
        "KW_TYPE",
        "KW_GRAPH",
        "KW_STRICT",
        "KW_DIGRAPH",
        "KW_SUBGRAPH",
        "KW_NODE",
        "KW_IGNORE",
        "KW_TAKE",
        "KW_SPAWN",
        "KW_WITH",
        "KW_ENTRY",
        "KW_EXIT",
        "KW_LENGTH",
        "KW_KEYS",
        "KW_CONTEXT",
        "KW_INFO",
        "KW_DETAILS",
        "KW_ACTIVITY",
        "KW_IMPORT",
        "COLON",
        "DBL_COLON",
        "STR_DBL_COLON",
        "LIST_DBL_COLON",
        "DICT_DBL_COLON",
        "COLON_OUT",
        "LBRACE",
        "RBRACE",
        "KW_EDGE",
        "KW_WALKER",
        "KW_ASYNC",
        "KW_SYNC",
        "KW_TEST",
        "KW_ASSERT",
        "SEMI",
        "EQ",
        "PEQ",
        "MEQ",
        "TEQ",
        "DEQ",
        "CPY_EQ",
        "KW_AND",
        "KW_OR",
        "KW_IF",
        "KW_ELIF",
        "KW_ELSE",
        "KW_FOR",
        "KW_TO",
        "KW_BY",
        "KW_WHILE",
        "KW_CONTINUE",
        "KW_BREAK",
        "KW_DISENGAGE",
        "KW_YIELD",
        "KW_SKIP",
        "KW_REPORT",
        "KW_DESTROY",
        "KW_TRY",
        "DOT",
        "NOT",
        "EE",
        "LT",
        "GT",
        "LTE",
        "GTE",
        "NE",
        "KW_IN",
        "KW_ANCHOR",
        "KW_HAS",
        "KW_GLOBAL",
        "KW_PRIVATE",
        "COMMA",
        "KW_CAN",
        "PLUS",
        "MINUS",
        "STAR_MUL",
        "DIV",
        "MOD",
        "POW",
        "LPAREN",
        "RPAREN",
        "LSQUARE",
        "RSQUARE",
        "FLOAT",
        "STRING",
        "BOOL",
        "INT",
        "NULL",
        "NAME",
        "COMMENT",
        "LINE_COMMENT",
        "PY_COMMENT",
        "WS",
        "ErrorChar",
    ]

    RULE_start = 0
    RULE_import_module = 1
    RULE_ver_label = 2
    RULE_import_items = 3
    RULE_import_names = 4
    RULE_element = 5
    RULE_global_var = 6
    RULE_architype = 7
    RULE_walker_block = 8
    RULE_test = 9
    RULE_namespaces = 10
    RULE_walk_entry_block = 11
    RULE_walk_exit_block = 12
    RULE_walk_activity_block = 13
    RULE_attr_block = 14
    RULE_attr_stmt = 15
    RULE_struct_block = 16
    RULE_can_block = 17
    RULE_graph_block = 18
    RULE_graph_block_spawn = 19
    RULE_graph_block_dot = 20
    RULE_has_root = 21
    RULE_has_stmt = 22
    RULE_has_assign = 23
    RULE_can_stmt = 24
    RULE_event_clause = 25
    RULE_preset_in_out = 26
    RULE_dotted_name = 27
    RULE_name_list = 28
    RULE_expr_list = 29
    RULE_code_block = 30
    RULE_node_ctx_block = 31
    RULE_statement = 32
    RULE_if_stmt = 33
    RULE_try_stmt = 34
    RULE_else_from_try = 35
    RULE_elif_stmt = 36
    RULE_else_stmt = 37
    RULE_for_stmt = 38
    RULE_while_stmt = 39
    RULE_ctrl_stmt = 40
    RULE_assert_stmt = 41
    RULE_destroy_action = 42
    RULE_report_action = 43
    RULE_walker_action = 44
    RULE_ignore_action = 45
    RULE_take_action = 46
    RULE_disengage_action = 47
    RULE_yield_action = 48
    RULE_expression = 49
    RULE_assignment = 50
    RULE_copy_assign = 51
    RULE_inc_assign = 52
    RULE_connect = 53
    RULE_logical = 54
    RULE_compare = 55
    RULE_cmp_op = 56
    RULE_nin = 57
    RULE_arithmetic = 58
    RULE_term = 59
    RULE_factor = 60
    RULE_power = 61
    RULE_global_ref = 62
    RULE_atom = 63
    RULE_atom_trailer = 64
    RULE_ability_op = 65
    RULE_ref = 66
    RULE_deref = 67
    RULE_built_in = 68
    RULE_cast_built_in = 69
    RULE_obj_built_in = 70
    RULE_dict_built_in = 71
    RULE_list_built_in = 72
    RULE_string_built_in = 73
    RULE_node_edge_ref = 74
    RULE_node_ref = 75
    RULE_walker_ref = 76
    RULE_graph_ref = 77
    RULE_type_ref = 78
    RULE_edge_ref = 79
    RULE_edge_to = 80
    RULE_edge_from = 81
    RULE_edge_any = 82
    RULE_list_val = 83
    RULE_index_slice = 84
    RULE_dict_val = 85
    RULE_kv_pair = 86
    RULE_spawn = 87
    RULE_spawn_object = 88
    RULE_spawn_edge = 89
    RULE_node_spawn = 90
    RULE_graph_spawn = 91
    RULE_walker_spawn = 92
    RULE_spawn_ctx = 93
    RULE_filter_ctx = 94
    RULE_spawn_assign = 95
    RULE_filter_compare = 96
    RULE_any_type = 97
    RULE_dot_graph = 98
    RULE_dot_stmt_list = 99
    RULE_dot_stmt = 100
    RULE_dot_attr_stmt = 101
    RULE_dot_attr_list = 102
    RULE_dot_a_list = 103
    RULE_dot_edge_stmt = 104
    RULE_dot_edgeRHS = 105
    RULE_dot_edgeop = 106
    RULE_dot_node_stmt = 107
    RULE_dot_node_id = 108
    RULE_dot_port = 109
    RULE_dot_subgraph = 110
    RULE_dot_id = 111

    ruleNames = [
        "start",
        "import_module",
        "ver_label",
        "import_items",
        "import_names",
        "element",
        "global_var",
        "architype",
        "walker_block",
        "test",
        "namespaces",
        "walk_entry_block",
        "walk_exit_block",
        "walk_activity_block",
        "attr_block",
        "attr_stmt",
        "struct_block",
        "can_block",
        "graph_block",
        "graph_block_spawn",
        "graph_block_dot",
        "has_root",
        "has_stmt",
        "has_assign",
        "can_stmt",
        "event_clause",
        "preset_in_out",
        "dotted_name",
        "name_list",
        "expr_list",
        "code_block",
        "node_ctx_block",
        "statement",
        "if_stmt",
        "try_stmt",
        "else_from_try",
        "elif_stmt",
        "else_stmt",
        "for_stmt",
        "while_stmt",
        "ctrl_stmt",
        "assert_stmt",
        "destroy_action",
        "report_action",
        "walker_action",
        "ignore_action",
        "take_action",
        "disengage_action",
        "yield_action",
        "expression",
        "assignment",
        "copy_assign",
        "inc_assign",
        "connect",
        "logical",
        "compare",
        "cmp_op",
        "nin",
        "arithmetic",
        "term",
        "factor",
        "power",
        "global_ref",
        "atom",
        "atom_trailer",
        "ability_op",
        "ref",
        "deref",
        "built_in",
        "cast_built_in",
        "obj_built_in",
        "dict_built_in",
        "list_built_in",
        "string_built_in",
        "node_edge_ref",
        "node_ref",
        "walker_ref",
        "graph_ref",
        "type_ref",
        "edge_ref",
        "edge_to",
        "edge_from",
        "edge_any",
        "list_val",
        "index_slice",
        "dict_val",
        "kv_pair",
        "spawn",
        "spawn_object",
        "spawn_edge",
        "node_spawn",
        "graph_spawn",
        "walker_spawn",
        "spawn_ctx",
        "filter_ctx",
        "spawn_assign",
        "filter_compare",
        "any_type",
        "dot_graph",
        "dot_stmt_list",
        "dot_stmt",
        "dot_attr_stmt",
        "dot_attr_list",
        "dot_a_list",
        "dot_edge_stmt",
        "dot_edgeRHS",
        "dot_edgeop",
        "dot_node_stmt",
        "dot_node_id",
        "dot_port",
        "dot_subgraph",
        "dot_id",
    ]

    EOF = Token.EOF
    T__0 = 1
    T__1 = 2
    T__2 = 3
    T__3 = 4
    T__4 = 5
    T__5 = 6
    T__6 = 7
    T__7 = 8
    TYP_STRING = 9
    TYP_INT = 10
    TYP_FLOAT = 11
    TYP_LIST = 12
    TYP_DICT = 13
    TYP_BOOL = 14
    KW_TYPE = 15
    KW_GRAPH = 16
    KW_STRICT = 17
    KW_DIGRAPH = 18
    KW_SUBGRAPH = 19
    KW_NODE = 20
    KW_IGNORE = 21
    KW_TAKE = 22
    KW_SPAWN = 23
    KW_WITH = 24
    KW_ENTRY = 25
    KW_EXIT = 26
    KW_LENGTH = 27
    KW_KEYS = 28
    KW_CONTEXT = 29
    KW_INFO = 30
    KW_DETAILS = 31
    KW_ACTIVITY = 32
    KW_IMPORT = 33
    COLON = 34
    DBL_COLON = 35
    STR_DBL_COLON = 36
    LIST_DBL_COLON = 37
    DICT_DBL_COLON = 38
    COLON_OUT = 39
    LBRACE = 40
    RBRACE = 41
    KW_EDGE = 42
    KW_WALKER = 43
    KW_ASYNC = 44
    KW_SYNC = 45
    KW_TEST = 46
    KW_ASSERT = 47
    SEMI = 48
    EQ = 49
    PEQ = 50
    MEQ = 51
    TEQ = 52
    DEQ = 53
    CPY_EQ = 54
    KW_AND = 55
    KW_OR = 56
    KW_IF = 57
    KW_ELIF = 58
    KW_ELSE = 59
    KW_FOR = 60
    KW_TO = 61
    KW_BY = 62
    KW_WHILE = 63
    KW_CONTINUE = 64
    KW_BREAK = 65
    KW_DISENGAGE = 66
    KW_YIELD = 67
    KW_SKIP = 68
    KW_REPORT = 69
    KW_DESTROY = 70
    KW_TRY = 71
    DOT = 72
    NOT = 73
    EE = 74
    LT = 75
    GT = 76
    LTE = 77
    GTE = 78
    NE = 79
    KW_IN = 80
    KW_ANCHOR = 81
    KW_HAS = 82
    KW_GLOBAL = 83
    KW_PRIVATE = 84
    COMMA = 85
    KW_CAN = 86
    PLUS = 87
    MINUS = 88
    STAR_MUL = 89
    DIV = 90
    MOD = 91
    POW = 92
    LPAREN = 93
    RPAREN = 94
    LSQUARE = 95
    RSQUARE = 96
    FLOAT = 97
    STRING = 98
    BOOL = 99
    INT = 100
    NULL = 101
    NAME = 102
    COMMENT = 103
    LINE_COMMENT = 104
    PY_COMMENT = 105
    WS = 106
    ErrorChar = 107

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.9.2")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class StartContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(jacParser.EOF, 0)

        def ver_label(self):
            return self.getTypedRuleContext(jacParser.Ver_labelContext, 0)

        def import_module(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Import_moduleContext)
            else:
                return self.getTypedRuleContext(jacParser.Import_moduleContext, i)

        def element(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.ElementContext)
            else:
                return self.getTypedRuleContext(jacParser.ElementContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_start

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStart"):
                listener.enterStart(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStart"):
                listener.exitStart(self)

    def start(self):

        localctx = jacParser.StartContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_start)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 225
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.T__0:
                self.state = 224
                self.ver_label()

            self.state = 230
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.KW_IMPORT:
                self.state = 227
                self.import_module()
                self.state = 232
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 236
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (
                ((_la) & ~0x3F) == 0
                and (
                    (1 << _la)
                    & (
                        (1 << jacParser.KW_TYPE)
                        | (1 << jacParser.KW_GRAPH)
                        | (1 << jacParser.KW_NODE)
                        | (1 << jacParser.KW_EDGE)
                        | (1 << jacParser.KW_WALKER)
                        | (1 << jacParser.KW_ASYNC)
                        | (1 << jacParser.KW_TEST)
                    )
                )
                != 0
            ) or _la == jacParser.KW_GLOBAL:
                self.state = 233
                self.element()
                self.state = 238
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 239
            self.match(jacParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_moduleContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_IMPORT(self):
            return self.getToken(jacParser.KW_IMPORT, 0)

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def KW_WITH(self):
            return self.getToken(jacParser.KW_WITH, 0)

        def STRING(self):
            return self.getToken(jacParser.STRING, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def import_items(self):
            return self.getTypedRuleContext(jacParser.Import_itemsContext, 0)

        def STAR_MUL(self):
            return self.getToken(jacParser.STAR_MUL, 0)

        def getRuleIndex(self):
            return jacParser.RULE_import_module

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_module"):
                listener.enterImport_module(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_module"):
                listener.exitImport_module(self)

    def import_module(self):

        localctx = jacParser.Import_moduleContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_import_module)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 241
            self.match(jacParser.KW_IMPORT)
            self.state = 242
            self.match(jacParser.LBRACE)
            self.state = 245
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [
                jacParser.KW_TYPE,
                jacParser.KW_GRAPH,
                jacParser.KW_NODE,
                jacParser.KW_EDGE,
                jacParser.KW_WALKER,
                jacParser.KW_GLOBAL,
            ]:
                self.state = 243
                self.import_items()
                pass
            elif token in [jacParser.STAR_MUL]:
                self.state = 244
                self.match(jacParser.STAR_MUL)
                pass
            else:
                raise NoViableAltException(self)

            self.state = 247
            self.match(jacParser.RBRACE)
            self.state = 248
            self.match(jacParser.KW_WITH)
            self.state = 249
            self.match(jacParser.STRING)
            self.state = 250
            self.match(jacParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ver_labelContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def STRING(self):
            return self.getToken(jacParser.STRING, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_ver_label

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterVer_label"):
                listener.enterVer_label(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitVer_label"):
                listener.exitVer_label(self)

    def ver_label(self):

        localctx = jacParser.Ver_labelContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_ver_label)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 252
            self.match(jacParser.T__0)
            self.state = 253
            self.match(jacParser.COLON)
            self.state = 254
            self.match(jacParser.STRING)
            self.state = 256
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.SEMI:
                self.state = 255
                self.match(jacParser.SEMI)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_itemsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_WALKER(self):
            return self.getToken(jacParser.KW_WALKER, 0)

        def STAR_MUL(self):
            return self.getToken(jacParser.STAR_MUL, 0)

        def import_names(self):
            return self.getTypedRuleContext(jacParser.Import_namesContext, 0)

        def COMMA(self):
            return self.getToken(jacParser.COMMA, 0)

        def import_items(self):
            return self.getTypedRuleContext(jacParser.Import_itemsContext, 0)

        def KW_NODE(self):
            return self.getToken(jacParser.KW_NODE, 0)

        def KW_EDGE(self):
            return self.getToken(jacParser.KW_EDGE, 0)

        def KW_GRAPH(self):
            return self.getToken(jacParser.KW_GRAPH, 0)

        def KW_GLOBAL(self):
            return self.getToken(jacParser.KW_GLOBAL, 0)

        def KW_TYPE(self):
            return self.getToken(jacParser.KW_TYPE, 0)

        def getRuleIndex(self):
            return jacParser.RULE_import_items

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_items"):
                listener.enterImport_items(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_items"):
                listener.exitImport_items(self)

    def import_items(self):

        localctx = jacParser.Import_itemsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_import_items)
        self._la = 0  # Token type
        try:
            self.state = 312
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_WALKER]:
                self.enterOuterAlt(localctx, 1)
                self.state = 258
                self.match(jacParser.KW_WALKER)
                self.state = 261
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.STAR_MUL]:
                    self.state = 259
                    self.match(jacParser.STAR_MUL)
                    pass
                elif token in [jacParser.DBL_COLON]:
                    self.state = 260
                    self.import_names()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 265
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 263
                    self.match(jacParser.COMMA)
                    self.state = 264
                    self.import_items()

                pass
            elif token in [jacParser.KW_NODE]:
                self.enterOuterAlt(localctx, 2)
                self.state = 267
                self.match(jacParser.KW_NODE)
                self.state = 270
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.STAR_MUL]:
                    self.state = 268
                    self.match(jacParser.STAR_MUL)
                    pass
                elif token in [jacParser.DBL_COLON]:
                    self.state = 269
                    self.import_names()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 274
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 272
                    self.match(jacParser.COMMA)
                    self.state = 273
                    self.import_items()

                pass
            elif token in [jacParser.KW_EDGE]:
                self.enterOuterAlt(localctx, 3)
                self.state = 276
                self.match(jacParser.KW_EDGE)
                self.state = 279
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.STAR_MUL]:
                    self.state = 277
                    self.match(jacParser.STAR_MUL)
                    pass
                elif token in [jacParser.DBL_COLON]:
                    self.state = 278
                    self.import_names()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 283
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 281
                    self.match(jacParser.COMMA)
                    self.state = 282
                    self.import_items()

                pass
            elif token in [jacParser.KW_GRAPH]:
                self.enterOuterAlt(localctx, 4)
                self.state = 285
                self.match(jacParser.KW_GRAPH)
                self.state = 288
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.STAR_MUL]:
                    self.state = 286
                    self.match(jacParser.STAR_MUL)
                    pass
                elif token in [jacParser.DBL_COLON]:
                    self.state = 287
                    self.import_names()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 292
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 290
                    self.match(jacParser.COMMA)
                    self.state = 291
                    self.import_items()

                pass
            elif token in [jacParser.KW_GLOBAL]:
                self.enterOuterAlt(localctx, 5)
                self.state = 294
                self.match(jacParser.KW_GLOBAL)
                self.state = 297
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.STAR_MUL]:
                    self.state = 295
                    self.match(jacParser.STAR_MUL)
                    pass
                elif token in [jacParser.DBL_COLON]:
                    self.state = 296
                    self.import_names()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 301
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 299
                    self.match(jacParser.COMMA)
                    self.state = 300
                    self.import_items()

                pass
            elif token in [jacParser.KW_TYPE]:
                self.enterOuterAlt(localctx, 6)
                self.state = 303
                self.match(jacParser.KW_TYPE)
                self.state = 306
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.STAR_MUL]:
                    self.state = 304
                    self.match(jacParser.STAR_MUL)
                    pass
                elif token in [jacParser.DBL_COLON]:
                    self.state = 305
                    self.import_names()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 310
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 308
                    self.match(jacParser.COMMA)
                    self.state = 309
                    self.import_items()

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Import_namesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def name_list(self):
            return self.getTypedRuleContext(jacParser.Name_listContext, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def getRuleIndex(self):
            return jacParser.RULE_import_names

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterImport_names"):
                listener.enterImport_names(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitImport_names"):
                listener.exitImport_names(self)

    def import_names(self):

        localctx = jacParser.Import_namesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_import_names)
        try:
            self.state = 321
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 18, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 314
                self.match(jacParser.DBL_COLON)
                self.state = 315
                self.match(jacParser.NAME)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 316
                self.match(jacParser.DBL_COLON)
                self.state = 317
                self.match(jacParser.LBRACE)
                self.state = 318
                self.name_list()
                self.state = 319
                self.match(jacParser.RBRACE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ElementContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def global_var(self):
            return self.getTypedRuleContext(jacParser.Global_varContext, 0)

        def architype(self):
            return self.getTypedRuleContext(jacParser.ArchitypeContext, 0)

        def test(self):
            return self.getTypedRuleContext(jacParser.TestContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_element

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterElement"):
                listener.enterElement(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitElement"):
                listener.exitElement(self)

    def element(self):

        localctx = jacParser.ElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_element)
        try:
            self.state = 326
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_GLOBAL]:
                self.enterOuterAlt(localctx, 1)
                self.state = 323
                self.global_var()
                pass
            elif token in [
                jacParser.KW_TYPE,
                jacParser.KW_GRAPH,
                jacParser.KW_NODE,
                jacParser.KW_EDGE,
                jacParser.KW_WALKER,
                jacParser.KW_ASYNC,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 324
                self.architype()
                pass
            elif token in [jacParser.KW_TEST]:
                self.enterOuterAlt(localctx, 3)
                self.state = 325
                self.test()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Global_varContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_GLOBAL(self):
            return self.getToken(jacParser.KW_GLOBAL, 0)

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.NAME)
            else:
                return self.getToken(jacParser.NAME, i)

        def EQ(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.EQ)
            else:
                return self.getToken(jacParser.EQ, i)

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(jacParser.ExpressionContext, i)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_global_var

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGlobal_var"):
                listener.enterGlobal_var(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGlobal_var"):
                listener.exitGlobal_var(self)

    def global_var(self):

        localctx = jacParser.Global_varContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_global_var)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 328
            self.match(jacParser.KW_GLOBAL)
            self.state = 329
            self.match(jacParser.NAME)
            self.state = 330
            self.match(jacParser.EQ)
            self.state = 331
            self.expression()
            self.state = 338
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.COMMA:
                self.state = 332
                self.match(jacParser.COMMA)
                self.state = 333
                self.match(jacParser.NAME)
                self.state = 334
                self.match(jacParser.EQ)
                self.state = 335
                self.expression()
                self.state = 340
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 341
            self.match(jacParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArchitypeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_NODE(self):
            return self.getToken(jacParser.KW_NODE, 0)

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.NAME)
            else:
                return self.getToken(jacParser.NAME, i)

        def attr_block(self):
            return self.getTypedRuleContext(jacParser.Attr_blockContext, 0)

        def COLON(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COLON)
            else:
                return self.getToken(jacParser.COLON, i)

        def INT(self):
            return self.getToken(jacParser.INT, 0)

        def KW_EDGE(self):
            return self.getToken(jacParser.KW_EDGE, 0)

        def KW_TYPE(self):
            return self.getToken(jacParser.KW_TYPE, 0)

        def struct_block(self):
            return self.getTypedRuleContext(jacParser.Struct_blockContext, 0)

        def KW_GRAPH(self):
            return self.getToken(jacParser.KW_GRAPH, 0)

        def graph_block(self):
            return self.getTypedRuleContext(jacParser.Graph_blockContext, 0)

        def KW_WALKER(self):
            return self.getToken(jacParser.KW_WALKER, 0)

        def walker_block(self):
            return self.getTypedRuleContext(jacParser.Walker_blockContext, 0)

        def KW_ASYNC(self):
            return self.getToken(jacParser.KW_ASYNC, 0)

        def namespaces(self):
            return self.getTypedRuleContext(jacParser.NamespacesContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_architype

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArchitype"):
                listener.enterArchitype(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArchitype"):
                listener.exitArchitype(self)

    def architype(self):

        localctx = jacParser.ArchitypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_architype)
        self._la = 0  # Token type
        try:
            self.state = 382
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_NODE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 343
                self.match(jacParser.KW_NODE)
                self.state = 344
                self.match(jacParser.NAME)
                self.state = 349
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 21, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 345
                        self.match(jacParser.COLON)
                        self.state = 346
                        self.match(jacParser.NAME)
                    self.state = 351
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 21, self._ctx)

                self.state = 354
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 22, self._ctx)
                if la_ == 1:
                    self.state = 352
                    self.match(jacParser.COLON)
                    self.state = 353
                    self.match(jacParser.INT)

                self.state = 356
                self.attr_block()
                pass
            elif token in [jacParser.KW_EDGE]:
                self.enterOuterAlt(localctx, 2)
                self.state = 357
                self.match(jacParser.KW_EDGE)
                self.state = 358
                self.match(jacParser.NAME)
                self.state = 363
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 23, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 359
                        self.match(jacParser.COLON)
                        self.state = 360
                        self.match(jacParser.NAME)
                    self.state = 365
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 23, self._ctx)

                self.state = 366
                self.attr_block()
                pass
            elif token in [jacParser.KW_TYPE]:
                self.enterOuterAlt(localctx, 3)
                self.state = 367
                self.match(jacParser.KW_TYPE)
                self.state = 368
                self.match(jacParser.NAME)
                self.state = 369
                self.struct_block()
                pass
            elif token in [jacParser.KW_GRAPH]:
                self.enterOuterAlt(localctx, 4)
                self.state = 370
                self.match(jacParser.KW_GRAPH)
                self.state = 371
                self.match(jacParser.NAME)
                self.state = 372
                self.graph_block()
                pass
            elif token in [jacParser.KW_WALKER, jacParser.KW_ASYNC]:
                self.enterOuterAlt(localctx, 5)
                self.state = 374
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.KW_ASYNC:
                    self.state = 373
                    self.match(jacParser.KW_ASYNC)

                self.state = 376
                self.match(jacParser.KW_WALKER)
                self.state = 377
                self.match(jacParser.NAME)
                self.state = 379
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COLON:
                    self.state = 378
                    self.namespaces()

                self.state = 381
                self.walker_block()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Walker_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def attr_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Attr_stmtContext)
            else:
                return self.getTypedRuleContext(jacParser.Attr_stmtContext, i)

        def walk_entry_block(self):
            return self.getTypedRuleContext(jacParser.Walk_entry_blockContext, 0)

        def statement(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.StatementContext)
            else:
                return self.getTypedRuleContext(jacParser.StatementContext, i)

        def walk_activity_block(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Walk_activity_blockContext)
            else:
                return self.getTypedRuleContext(jacParser.Walk_activity_blockContext, i)

        def walk_exit_block(self):
            return self.getTypedRuleContext(jacParser.Walk_exit_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_walker_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWalker_block"):
                listener.enterWalker_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWalker_block"):
                listener.exitWalker_block(self)

    def walker_block(self):

        localctx = jacParser.Walker_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_walker_block)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 384
            self.match(jacParser.LBRACE)
            self.state = 388
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.KW_HAS or _la == jacParser.KW_CAN:
                self.state = 385
                self.attr_stmt()
                self.state = 390
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 392
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 28, self._ctx)
            if la_ == 1:
                self.state = 391
                self.walk_entry_block()

            self.state = 398
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 30, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 396
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [
                        jacParser.T__1,
                        jacParser.T__2,
                        jacParser.T__4,
                        jacParser.T__5,
                        jacParser.T__6,
                        jacParser.TYP_STRING,
                        jacParser.TYP_INT,
                        jacParser.TYP_FLOAT,
                        jacParser.TYP_LIST,
                        jacParser.TYP_DICT,
                        jacParser.TYP_BOOL,
                        jacParser.KW_TYPE,
                        jacParser.KW_NODE,
                        jacParser.KW_IGNORE,
                        jacParser.KW_TAKE,
                        jacParser.KW_SPAWN,
                        jacParser.COLON,
                        jacParser.DBL_COLON,
                        jacParser.LBRACE,
                        jacParser.KW_EDGE,
                        jacParser.KW_SYNC,
                        jacParser.KW_ASSERT,
                        jacParser.KW_IF,
                        jacParser.KW_FOR,
                        jacParser.KW_WHILE,
                        jacParser.KW_CONTINUE,
                        jacParser.KW_BREAK,
                        jacParser.KW_DISENGAGE,
                        jacParser.KW_YIELD,
                        jacParser.KW_SKIP,
                        jacParser.KW_REPORT,
                        jacParser.KW_DESTROY,
                        jacParser.KW_TRY,
                        jacParser.NOT,
                        jacParser.KW_GLOBAL,
                        jacParser.PLUS,
                        jacParser.MINUS,
                        jacParser.STAR_MUL,
                        jacParser.LPAREN,
                        jacParser.LSQUARE,
                        jacParser.FLOAT,
                        jacParser.STRING,
                        jacParser.BOOL,
                        jacParser.INT,
                        jacParser.NULL,
                        jacParser.NAME,
                    ]:
                        self.state = 394
                        self.statement()
                        pass
                    elif token in [jacParser.KW_WITH]:
                        self.state = 395
                        self.walk_activity_block()
                        pass
                    else:
                        raise NoViableAltException(self)

                self.state = 400
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 30, self._ctx)

            self.state = 402
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.KW_WITH:
                self.state = 401
                self.walk_exit_block()

            self.state = 404
            self.match(jacParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TestContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_TEST(self):
            return self.getToken(jacParser.KW_TEST, 0)

        def STRING(self):
            return self.getToken(jacParser.STRING, 0)

        def KW_WITH(self):
            return self.getToken(jacParser.KW_WITH, 0)

        def KW_BY(self):
            return self.getToken(jacParser.KW_BY, 0)

        def graph_ref(self):
            return self.getTypedRuleContext(jacParser.Graph_refContext, 0)

        def KW_GRAPH(self):
            return self.getToken(jacParser.KW_GRAPH, 0)

        def graph_block(self):
            return self.getTypedRuleContext(jacParser.Graph_blockContext, 0)

        def KW_WALKER(self):
            return self.getToken(jacParser.KW_WALKER, 0)

        def walker_block(self):
            return self.getTypedRuleContext(jacParser.Walker_blockContext, 0)

        def walker_ref(self):
            return self.getTypedRuleContext(jacParser.Walker_refContext, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_test

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTest"):
                listener.enterTest(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTest"):
                listener.exitTest(self)

    def test(self):

        localctx = jacParser.TestContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_test)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 406
            self.match(jacParser.KW_TEST)
            self.state = 407
            self.match(jacParser.STRING)
            self.state = 408
            self.match(jacParser.KW_WITH)
            self.state = 412
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 32, self._ctx)
            if la_ == 1:
                self.state = 409
                self.graph_ref()
                pass

            elif la_ == 2:
                self.state = 410
                self.match(jacParser.KW_GRAPH)
                self.state = 411
                self.graph_block()
                pass

            self.state = 414
            self.match(jacParser.KW_BY)
            self.state = 425
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 35, self._ctx)
            if la_ == 1:
                self.state = 415
                self.walker_ref()
                self.state = 417
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.LPAREN:
                    self.state = 416
                    self.spawn_ctx()

                self.state = 421
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.COLON, jacParser.LBRACE]:
                    self.state = 419
                    self.code_block()
                    pass
                elif token in [jacParser.SEMI]:
                    self.state = 420
                    self.match(jacParser.SEMI)
                    pass
                else:
                    raise NoViableAltException(self)

                pass

            elif la_ == 2:
                self.state = 423
                self.match(jacParser.KW_WALKER)
                self.state = 424
                self.walker_block()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NamespacesContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def name_list(self):
            return self.getTypedRuleContext(jacParser.Name_listContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_namespaces

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNamespaces"):
                listener.enterNamespaces(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNamespaces"):
                listener.exitNamespaces(self)

    def namespaces(self):

        localctx = jacParser.NamespacesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_namespaces)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 427
            self.match(jacParser.COLON)
            self.state = 428
            self.name_list()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Walk_entry_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_WITH(self):
            return self.getToken(jacParser.KW_WITH, 0)

        def KW_ENTRY(self):
            return self.getToken(jacParser.KW_ENTRY, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_walk_entry_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWalk_entry_block"):
                listener.enterWalk_entry_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWalk_entry_block"):
                listener.exitWalk_entry_block(self)

    def walk_entry_block(self):

        localctx = jacParser.Walk_entry_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_walk_entry_block)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 430
            self.match(jacParser.KW_WITH)
            self.state = 431
            self.match(jacParser.KW_ENTRY)
            self.state = 432
            self.code_block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Walk_exit_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_WITH(self):
            return self.getToken(jacParser.KW_WITH, 0)

        def KW_EXIT(self):
            return self.getToken(jacParser.KW_EXIT, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_walk_exit_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWalk_exit_block"):
                listener.enterWalk_exit_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWalk_exit_block"):
                listener.exitWalk_exit_block(self)

    def walk_exit_block(self):

        localctx = jacParser.Walk_exit_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_walk_exit_block)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 434
            self.match(jacParser.KW_WITH)
            self.state = 435
            self.match(jacParser.KW_EXIT)
            self.state = 436
            self.code_block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Walk_activity_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_WITH(self):
            return self.getToken(jacParser.KW_WITH, 0)

        def KW_ACTIVITY(self):
            return self.getToken(jacParser.KW_ACTIVITY, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_walk_activity_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWalk_activity_block"):
                listener.enterWalk_activity_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWalk_activity_block"):
                listener.exitWalk_activity_block(self)

    def walk_activity_block(self):

        localctx = jacParser.Walk_activity_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_walk_activity_block)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 438
            self.match(jacParser.KW_WITH)
            self.state = 439
            self.match(jacParser.KW_ACTIVITY)
            self.state = 440
            self.code_block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Attr_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def attr_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Attr_stmtContext)
            else:
                return self.getTypedRuleContext(jacParser.Attr_stmtContext, i)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_attr_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAttr_block"):
                listener.enterAttr_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAttr_block"):
                listener.exitAttr_block(self)

    def attr_block(self):

        localctx = jacParser.Attr_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_attr_block)
        self._la = 0  # Token type
        try:
            self.state = 453
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.LBRACE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 442
                self.match(jacParser.LBRACE)
                self.state = 446
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == jacParser.KW_HAS or _la == jacParser.KW_CAN:
                    self.state = 443
                    self.attr_stmt()
                    self.state = 448
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 449
                self.match(jacParser.RBRACE)
                pass
            elif token in [jacParser.COLON]:
                self.enterOuterAlt(localctx, 2)
                self.state = 450
                self.match(jacParser.COLON)
                self.state = 451
                self.attr_stmt()
                pass
            elif token in [jacParser.SEMI]:
                self.enterOuterAlt(localctx, 3)
                self.state = 452
                self.match(jacParser.SEMI)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Attr_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def has_stmt(self):
            return self.getTypedRuleContext(jacParser.Has_stmtContext, 0)

        def can_stmt(self):
            return self.getTypedRuleContext(jacParser.Can_stmtContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_attr_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAttr_stmt"):
                listener.enterAttr_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAttr_stmt"):
                listener.exitAttr_stmt(self)

    def attr_stmt(self):

        localctx = jacParser.Attr_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_attr_stmt)
        try:
            self.state = 457
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_HAS]:
                self.enterOuterAlt(localctx, 1)
                self.state = 455
                self.has_stmt()
                pass
            elif token in [jacParser.KW_CAN]:
                self.enterOuterAlt(localctx, 2)
                self.state = 456
                self.can_stmt()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Struct_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def has_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Has_stmtContext)
            else:
                return self.getTypedRuleContext(jacParser.Has_stmtContext, i)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_struct_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStruct_block"):
                listener.enterStruct_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStruct_block"):
                listener.exitStruct_block(self)

    def struct_block(self):

        localctx = jacParser.Struct_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_struct_block)
        self._la = 0  # Token type
        try:
            self.state = 470
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.LBRACE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 459
                self.match(jacParser.LBRACE)
                self.state = 463
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == jacParser.KW_HAS:
                    self.state = 460
                    self.has_stmt()
                    self.state = 465
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 466
                self.match(jacParser.RBRACE)
                pass
            elif token in [jacParser.COLON]:
                self.enterOuterAlt(localctx, 2)
                self.state = 467
                self.match(jacParser.COLON)
                self.state = 468
                self.has_stmt()
                pass
            elif token in [jacParser.SEMI]:
                self.enterOuterAlt(localctx, 3)
                self.state = 469
                self.match(jacParser.SEMI)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Can_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def can_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Can_stmtContext)
            else:
                return self.getTypedRuleContext(jacParser.Can_stmtContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_can_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCan_block"):
                listener.enterCan_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCan_block"):
                listener.exitCan_block(self)

    def can_block(self):

        localctx = jacParser.Can_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_can_block)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 475
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.KW_CAN:
                self.state = 472
                self.can_stmt()
                self.state = 477
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Graph_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def graph_block_spawn(self):
            return self.getTypedRuleContext(jacParser.Graph_block_spawnContext, 0)

        def graph_block_dot(self):
            return self.getTypedRuleContext(jacParser.Graph_block_dotContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_graph_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGraph_block"):
                listener.enterGraph_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGraph_block"):
                listener.exitGraph_block(self)

    def graph_block(self):

        localctx = jacParser.Graph_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_graph_block)
        try:
            self.state = 480
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 42, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 478
                self.graph_block_spawn()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 479
                self.graph_block_dot()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Graph_block_spawnContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def has_root(self):
            return self.getTypedRuleContext(jacParser.Has_rootContext, 0)

        def can_block(self):
            return self.getTypedRuleContext(jacParser.Can_blockContext, 0)

        def KW_SPAWN(self):
            return self.getToken(jacParser.KW_SPAWN, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_graph_block_spawn

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGraph_block_spawn"):
                listener.enterGraph_block_spawn(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGraph_block_spawn"):
                listener.exitGraph_block_spawn(self)

    def graph_block_spawn(self):

        localctx = jacParser.Graph_block_spawnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_graph_block_spawn)
        try:
            self.state = 496
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.LBRACE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 482
                self.match(jacParser.LBRACE)
                self.state = 483
                self.has_root()
                self.state = 484
                self.can_block()
                self.state = 485
                self.match(jacParser.KW_SPAWN)
                self.state = 486
                self.code_block()
                self.state = 487
                self.match(jacParser.RBRACE)
                pass
            elif token in [jacParser.COLON]:
                self.enterOuterAlt(localctx, 2)
                self.state = 489
                self.match(jacParser.COLON)
                self.state = 490
                self.has_root()
                self.state = 491
                self.can_block()
                self.state = 492
                self.match(jacParser.KW_SPAWN)
                self.state = 493
                self.code_block()
                self.state = 494
                self.match(jacParser.SEMI)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Graph_block_dotContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def has_root(self):
            return self.getTypedRuleContext(jacParser.Has_rootContext, 0)

        def dot_graph(self):
            return self.getTypedRuleContext(jacParser.Dot_graphContext, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_graph_block_dot

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGraph_block_dot"):
                listener.enterGraph_block_dot(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGraph_block_dot"):
                listener.exitGraph_block_dot(self)

    def graph_block_dot(self):

        localctx = jacParser.Graph_block_dotContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_graph_block_dot)
        try:
            self.state = 508
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.LBRACE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 498
                self.match(jacParser.LBRACE)
                self.state = 499
                self.has_root()
                self.state = 500
                self.dot_graph()
                self.state = 501
                self.match(jacParser.RBRACE)
                pass
            elif token in [jacParser.COLON]:
                self.enterOuterAlt(localctx, 2)
                self.state = 503
                self.match(jacParser.COLON)
                self.state = 504
                self.has_root()
                self.state = 505
                self.dot_graph()
                self.state = 506
                self.match(jacParser.SEMI)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Has_rootContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_HAS(self):
            return self.getToken(jacParser.KW_HAS, 0)

        def KW_ANCHOR(self):
            return self.getToken(jacParser.KW_ANCHOR, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_has_root

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterHas_root"):
                listener.enterHas_root(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitHas_root"):
                listener.exitHas_root(self)

    def has_root(self):

        localctx = jacParser.Has_rootContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_has_root)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 510
            self.match(jacParser.KW_HAS)
            self.state = 511
            self.match(jacParser.KW_ANCHOR)
            self.state = 512
            self.match(jacParser.NAME)
            self.state = 513
            self.match(jacParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Has_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_HAS(self):
            return self.getToken(jacParser.KW_HAS, 0)

        def has_assign(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Has_assignContext)
            else:
                return self.getTypedRuleContext(jacParser.Has_assignContext, i)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_has_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterHas_stmt"):
                listener.enterHas_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitHas_stmt"):
                listener.exitHas_stmt(self)

    def has_stmt(self):

        localctx = jacParser.Has_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_has_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 515
            self.match(jacParser.KW_HAS)
            self.state = 516
            self.has_assign()
            self.state = 521
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.COMMA:
                self.state = 517
                self.match(jacParser.COMMA)
                self.state = 518
                self.has_assign()
                self.state = 523
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 524
            self.match(jacParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Has_assignContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def EQ(self):
            return self.getToken(jacParser.EQ, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def KW_PRIVATE(self):
            return self.getToken(jacParser.KW_PRIVATE, 0)

        def KW_ANCHOR(self):
            return self.getToken(jacParser.KW_ANCHOR, 0)

        def getRuleIndex(self):
            return jacParser.RULE_has_assign

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterHas_assign"):
                listener.enterHas_assign(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitHas_assign"):
                listener.exitHas_assign(self)

    def has_assign(self):

        localctx = jacParser.Has_assignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 46, self.RULE_has_assign)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 527
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.KW_PRIVATE:
                self.state = 526
                self.match(jacParser.KW_PRIVATE)

            self.state = 530
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.KW_ANCHOR:
                self.state = 529
                self.match(jacParser.KW_ANCHOR)

            self.state = 536
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 48, self._ctx)
            if la_ == 1:
                self.state = 532
                self.match(jacParser.NAME)
                pass

            elif la_ == 2:
                self.state = 533
                self.match(jacParser.NAME)
                self.state = 534
                self.match(jacParser.EQ)
                self.state = 535
                self.expression()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Can_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_CAN(self):
            return self.getToken(jacParser.KW_CAN, 0)

        def dotted_name(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dotted_nameContext)
            else:
                return self.getTypedRuleContext(jacParser.Dotted_nameContext, i)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def preset_in_out(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Preset_in_outContext)
            else:
                return self.getTypedRuleContext(jacParser.Preset_in_outContext, i)

        def event_clause(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Event_clauseContext)
            else:
                return self.getTypedRuleContext(jacParser.Event_clauseContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_can_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCan_stmt"):
                listener.enterCan_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCan_stmt"):
                listener.exitCan_stmt(self)

    def can_stmt(self):

        localctx = jacParser.Can_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 48, self.RULE_can_stmt)
        self._la = 0  # Token type
        try:
            self.state = 565
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 53, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 538
                self.match(jacParser.KW_CAN)
                self.state = 539
                self.dotted_name()
                self.state = 543
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.DBL_COLON:
                    self.state = 540
                    self.preset_in_out()
                    self.state = 541
                    self.event_clause()

                self.state = 554
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == jacParser.COMMA:
                    self.state = 545
                    self.match(jacParser.COMMA)
                    self.state = 546
                    self.dotted_name()
                    self.state = 550
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la == jacParser.DBL_COLON:
                        self.state = 547
                        self.preset_in_out()
                        self.state = 548
                        self.event_clause()

                    self.state = 556
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 557
                self.match(jacParser.SEMI)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 559
                self.match(jacParser.KW_CAN)
                self.state = 560
                self.match(jacParser.NAME)
                self.state = 562
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.KW_WITH:
                    self.state = 561
                    self.event_clause()

                self.state = 564
                self.code_block()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Event_clauseContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_WITH(self):
            return self.getToken(jacParser.KW_WITH, 0)

        def KW_ENTRY(self):
            return self.getToken(jacParser.KW_ENTRY, 0)

        def KW_EXIT(self):
            return self.getToken(jacParser.KW_EXIT, 0)

        def KW_ACTIVITY(self):
            return self.getToken(jacParser.KW_ACTIVITY, 0)

        def name_list(self):
            return self.getTypedRuleContext(jacParser.Name_listContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_event_clause

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEvent_clause"):
                listener.enterEvent_clause(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEvent_clause"):
                listener.exitEvent_clause(self)

    def event_clause(self):

        localctx = jacParser.Event_clauseContext(self, self._ctx, self.state)
        self.enterRule(localctx, 50, self.RULE_event_clause)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 567
            self.match(jacParser.KW_WITH)
            self.state = 569
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.NAME:
                self.state = 568
                self.name_list()

            self.state = 571
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.KW_ENTRY)
                            | (1 << jacParser.KW_EXIT)
                            | (1 << jacParser.KW_ACTIVITY)
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Preset_in_outContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DBL_COLON(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.DBL_COLON)
            else:
                return self.getToken(jacParser.DBL_COLON, i)

        def COLON_OUT(self):
            return self.getToken(jacParser.COLON_OUT, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def expr_list(self):
            return self.getTypedRuleContext(jacParser.Expr_listContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_preset_in_out

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreset_in_out"):
                listener.enterPreset_in_out(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreset_in_out"):
                listener.exitPreset_in_out(self)

    def preset_in_out(self):

        localctx = jacParser.Preset_in_outContext(self, self._ctx, self.state)
        self.enterRule(localctx, 52, self.RULE_preset_in_out)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 573
            self.match(jacParser.DBL_COLON)
            self.state = 575
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 55, self._ctx)
            if la_ == 1:
                self.state = 574
                self.expr_list()

            self.state = 580
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.DBL_COLON]:
                self.state = 577
                self.match(jacParser.DBL_COLON)
                pass
            elif token in [jacParser.COLON_OUT]:
                self.state = 578
                self.match(jacParser.COLON_OUT)
                self.state = 579
                self.expression()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dotted_nameContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.NAME)
            else:
                return self.getToken(jacParser.NAME, i)

        def DOT(self):
            return self.getToken(jacParser.DOT, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dotted_name

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDotted_name"):
                listener.enterDotted_name(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDotted_name"):
                listener.exitDotted_name(self)

    def dotted_name(self):

        localctx = jacParser.Dotted_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 54, self.RULE_dotted_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 582
            self.match(jacParser.NAME)
            self.state = 583
            self.match(jacParser.DOT)
            self.state = 584
            self.match(jacParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Name_listContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.NAME)
            else:
                return self.getToken(jacParser.NAME, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_name_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterName_list"):
                listener.enterName_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitName_list"):
                listener.exitName_list(self)

    def name_list(self):

        localctx = jacParser.Name_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 56, self.RULE_name_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 586
            self.match(jacParser.NAME)
            self.state = 591
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.COMMA:
                self.state = 587
                self.match(jacParser.COMMA)
                self.state = 588
                self.match(jacParser.NAME)
                self.state = 593
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Expr_listContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(jacParser.ExpressionContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_expr_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExpr_list"):
                listener.enterExpr_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExpr_list"):
                listener.exitExpr_list(self)

    def expr_list(self):

        localctx = jacParser.Expr_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 58, self.RULE_expr_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 594
            self.expression()
            self.state = 599
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.COMMA:
                self.state = 595
                self.match(jacParser.COMMA)
                self.state = 596
                self.expression()
                self.state = 601
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Code_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def statement(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.StatementContext)
            else:
                return self.getTypedRuleContext(jacParser.StatementContext, i)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def getRuleIndex(self):
            return jacParser.RULE_code_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCode_block"):
                listener.enterCode_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCode_block"):
                listener.exitCode_block(self)

    def code_block(self):

        localctx = jacParser.Code_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 60, self.RULE_code_block)
        self._la = 0  # Token type
        try:
            self.state = 612
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.LBRACE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 602
                self.match(jacParser.LBRACE)
                self.state = 606
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.T__1)
                            | (1 << jacParser.T__2)
                            | (1 << jacParser.T__4)
                            | (1 << jacParser.T__5)
                            | (1 << jacParser.T__6)
                            | (1 << jacParser.TYP_STRING)
                            | (1 << jacParser.TYP_INT)
                            | (1 << jacParser.TYP_FLOAT)
                            | (1 << jacParser.TYP_LIST)
                            | (1 << jacParser.TYP_DICT)
                            | (1 << jacParser.TYP_BOOL)
                            | (1 << jacParser.KW_TYPE)
                            | (1 << jacParser.KW_NODE)
                            | (1 << jacParser.KW_IGNORE)
                            | (1 << jacParser.KW_TAKE)
                            | (1 << jacParser.KW_SPAWN)
                            | (1 << jacParser.COLON)
                            | (1 << jacParser.DBL_COLON)
                            | (1 << jacParser.LBRACE)
                            | (1 << jacParser.KW_EDGE)
                            | (1 << jacParser.KW_SYNC)
                            | (1 << jacParser.KW_ASSERT)
                            | (1 << jacParser.KW_IF)
                            | (1 << jacParser.KW_FOR)
                            | (1 << jacParser.KW_WHILE)
                        )
                    )
                    != 0
                ) or (
                    (((_la - 64)) & ~0x3F) == 0
                    and (
                        (1 << (_la - 64))
                        & (
                            (1 << (jacParser.KW_CONTINUE - 64))
                            | (1 << (jacParser.KW_BREAK - 64))
                            | (1 << (jacParser.KW_DISENGAGE - 64))
                            | (1 << (jacParser.KW_YIELD - 64))
                            | (1 << (jacParser.KW_SKIP - 64))
                            | (1 << (jacParser.KW_REPORT - 64))
                            | (1 << (jacParser.KW_DESTROY - 64))
                            | (1 << (jacParser.KW_TRY - 64))
                            | (1 << (jacParser.NOT - 64))
                            | (1 << (jacParser.KW_GLOBAL - 64))
                            | (1 << (jacParser.PLUS - 64))
                            | (1 << (jacParser.MINUS - 64))
                            | (1 << (jacParser.STAR_MUL - 64))
                            | (1 << (jacParser.LPAREN - 64))
                            | (1 << (jacParser.LSQUARE - 64))
                            | (1 << (jacParser.FLOAT - 64))
                            | (1 << (jacParser.STRING - 64))
                            | (1 << (jacParser.BOOL - 64))
                            | (1 << (jacParser.INT - 64))
                            | (1 << (jacParser.NULL - 64))
                            | (1 << (jacParser.NAME - 64))
                        )
                    )
                    != 0
                ):
                    self.state = 603
                    self.statement()
                    self.state = 608
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 609
                self.match(jacParser.RBRACE)
                pass
            elif token in [jacParser.COLON]:
                self.enterOuterAlt(localctx, 2)
                self.state = 610
                self.match(jacParser.COLON)
                self.state = 611
                self.statement()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Node_ctx_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def name_list(self):
            return self.getTypedRuleContext(jacParser.Name_listContext, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_node_ctx_block

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNode_ctx_block"):
                listener.enterNode_ctx_block(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNode_ctx_block"):
                listener.exitNode_ctx_block(self)

    def node_ctx_block(self):

        localctx = jacParser.Node_ctx_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 62, self.RULE_node_ctx_block)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 614
            self.name_list()
            self.state = 615
            self.code_block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StatementContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def node_ctx_block(self):
            return self.getTypedRuleContext(jacParser.Node_ctx_blockContext, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def if_stmt(self):
            return self.getTypedRuleContext(jacParser.If_stmtContext, 0)

        def try_stmt(self):
            return self.getTypedRuleContext(jacParser.Try_stmtContext, 0)

        def for_stmt(self):
            return self.getTypedRuleContext(jacParser.For_stmtContext, 0)

        def while_stmt(self):
            return self.getTypedRuleContext(jacParser.While_stmtContext, 0)

        def assert_stmt(self):
            return self.getTypedRuleContext(jacParser.Assert_stmtContext, 0)

        def ctrl_stmt(self):
            return self.getTypedRuleContext(jacParser.Ctrl_stmtContext, 0)

        def destroy_action(self):
            return self.getTypedRuleContext(jacParser.Destroy_actionContext, 0)

        def report_action(self):
            return self.getTypedRuleContext(jacParser.Report_actionContext, 0)

        def walker_action(self):
            return self.getTypedRuleContext(jacParser.Walker_actionContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_statement

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterStatement"):
                listener.enterStatement(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitStatement"):
                listener.exitStatement(self)

    def statement(self):

        localctx = jacParser.StatementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 64, self.RULE_statement)
        try:
            self.state = 635
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 61, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 617
                self.code_block()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 618
                self.node_ctx_block()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 619
                self.expression()
                self.state = 620
                self.match(jacParser.SEMI)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 622
                self.if_stmt()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 623
                self.try_stmt()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 624
                self.for_stmt()
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 625
                self.while_stmt()
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 626
                self.assert_stmt()
                self.state = 627
                self.match(jacParser.SEMI)
                pass

            elif la_ == 9:
                self.enterOuterAlt(localctx, 9)
                self.state = 629
                self.ctrl_stmt()
                self.state = 630
                self.match(jacParser.SEMI)
                pass

            elif la_ == 10:
                self.enterOuterAlt(localctx, 10)
                self.state = 632
                self.destroy_action()
                pass

            elif la_ == 11:
                self.enterOuterAlt(localctx, 11)
                self.state = 633
                self.report_action()
                pass

            elif la_ == 12:
                self.enterOuterAlt(localctx, 12)
                self.state = 634
                self.walker_action()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class If_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_IF(self):
            return self.getToken(jacParser.KW_IF, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def elif_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Elif_stmtContext)
            else:
                return self.getTypedRuleContext(jacParser.Elif_stmtContext, i)

        def else_stmt(self):
            return self.getTypedRuleContext(jacParser.Else_stmtContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_if_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIf_stmt"):
                listener.enterIf_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIf_stmt"):
                listener.exitIf_stmt(self)

    def if_stmt(self):

        localctx = jacParser.If_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 66, self.RULE_if_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 637
            self.match(jacParser.KW_IF)
            self.state = 638
            self.expression()
            self.state = 639
            self.code_block()
            self.state = 643
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 62, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 640
                    self.elif_stmt()
                self.state = 645
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 62, self._ctx)

            self.state = 647
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 63, self._ctx)
            if la_ == 1:
                self.state = 646
                self.else_stmt()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Try_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_TRY(self):
            return self.getToken(jacParser.KW_TRY, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def else_from_try(self):
            return self.getTypedRuleContext(jacParser.Else_from_tryContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_try_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTry_stmt"):
                listener.enterTry_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTry_stmt"):
                listener.exitTry_stmt(self)

    def try_stmt(self):

        localctx = jacParser.Try_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 68, self.RULE_try_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 649
            self.match(jacParser.KW_TRY)
            self.state = 650
            self.code_block()
            self.state = 652
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 64, self._ctx)
            if la_ == 1:
                self.state = 651
                self.else_from_try()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Else_from_tryContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_ELSE(self):
            return self.getToken(jacParser.KW_ELSE, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def KW_WITH(self):
            return self.getToken(jacParser.KW_WITH, 0)

        def getRuleIndex(self):
            return jacParser.RULE_else_from_try

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterElse_from_try"):
                listener.enterElse_from_try(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitElse_from_try"):
                listener.exitElse_from_try(self)

    def else_from_try(self):

        localctx = jacParser.Else_from_tryContext(self, self._ctx, self.state)
        self.enterRule(localctx, 70, self.RULE_else_from_try)
        self._la = 0  # Token type
        try:
            self.state = 667
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 67, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 654
                self.match(jacParser.KW_ELSE)
                self.state = 658
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.LPAREN:
                    self.state = 655
                    self.match(jacParser.LPAREN)
                    self.state = 656
                    self.match(jacParser.NAME)
                    self.state = 657
                    self.match(jacParser.RPAREN)

                self.state = 660
                self.code_block()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 661
                self.match(jacParser.KW_ELSE)
                self.state = 664
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.KW_WITH:
                    self.state = 662
                    self.match(jacParser.KW_WITH)
                    self.state = 663
                    self.match(jacParser.NAME)

                self.state = 666
                self.code_block()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Elif_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_ELIF(self):
            return self.getToken(jacParser.KW_ELIF, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_elif_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterElif_stmt"):
                listener.enterElif_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitElif_stmt"):
                listener.exitElif_stmt(self)

    def elif_stmt(self):

        localctx = jacParser.Elif_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 72, self.RULE_elif_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 669
            self.match(jacParser.KW_ELIF)
            self.state = 670
            self.expression()
            self.state = 671
            self.code_block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Else_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_ELSE(self):
            return self.getToken(jacParser.KW_ELSE, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_else_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterElse_stmt"):
                listener.enterElse_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitElse_stmt"):
                listener.exitElse_stmt(self)

    def else_stmt(self):

        localctx = jacParser.Else_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 74, self.RULE_else_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 673
            self.match(jacParser.KW_ELSE)
            self.state = 674
            self.code_block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class For_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_FOR(self):
            return self.getToken(jacParser.KW_FOR, 0)

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(jacParser.ExpressionContext, i)

        def KW_TO(self):
            return self.getToken(jacParser.KW_TO, 0)

        def KW_BY(self):
            return self.getToken(jacParser.KW_BY, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def NAME(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.NAME)
            else:
                return self.getToken(jacParser.NAME, i)

        def KW_IN(self):
            return self.getToken(jacParser.KW_IN, 0)

        def COMMA(self):
            return self.getToken(jacParser.COMMA, 0)

        def getRuleIndex(self):
            return jacParser.RULE_for_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFor_stmt"):
                listener.enterFor_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFor_stmt"):
                listener.exitFor_stmt(self)

    def for_stmt(self):

        localctx = jacParser.For_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 76, self.RULE_for_stmt)
        self._la = 0  # Token type
        try:
            self.state = 694
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 69, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 676
                self.match(jacParser.KW_FOR)
                self.state = 677
                self.expression()
                self.state = 678
                self.match(jacParser.KW_TO)
                self.state = 679
                self.expression()
                self.state = 680
                self.match(jacParser.KW_BY)
                self.state = 681
                self.expression()
                self.state = 682
                self.code_block()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 684
                self.match(jacParser.KW_FOR)
                self.state = 685
                self.match(jacParser.NAME)
                self.state = 688
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 686
                    self.match(jacParser.COMMA)
                    self.state = 687
                    self.match(jacParser.NAME)

                self.state = 690
                self.match(jacParser.KW_IN)
                self.state = 691
                self.expression()
                self.state = 692
                self.code_block()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class While_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_WHILE(self):
            return self.getToken(jacParser.KW_WHILE, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def code_block(self):
            return self.getTypedRuleContext(jacParser.Code_blockContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_while_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWhile_stmt"):
                listener.enterWhile_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWhile_stmt"):
                listener.exitWhile_stmt(self)

    def while_stmt(self):

        localctx = jacParser.While_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 78, self.RULE_while_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 696
            self.match(jacParser.KW_WHILE)
            self.state = 697
            self.expression()
            self.state = 698
            self.code_block()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ctrl_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_CONTINUE(self):
            return self.getToken(jacParser.KW_CONTINUE, 0)

        def KW_BREAK(self):
            return self.getToken(jacParser.KW_BREAK, 0)

        def KW_SKIP(self):
            return self.getToken(jacParser.KW_SKIP, 0)

        def getRuleIndex(self):
            return jacParser.RULE_ctrl_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCtrl_stmt"):
                listener.enterCtrl_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCtrl_stmt"):
                listener.exitCtrl_stmt(self)

    def ctrl_stmt(self):

        localctx = jacParser.Ctrl_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 80, self.RULE_ctrl_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 700
            _la = self._input.LA(1)
            if not (
                (
                    (((_la - 64)) & ~0x3F) == 0
                    and (
                        (1 << (_la - 64))
                        & (
                            (1 << (jacParser.KW_CONTINUE - 64))
                            | (1 << (jacParser.KW_BREAK - 64))
                            | (1 << (jacParser.KW_SKIP - 64))
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Assert_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_ASSERT(self):
            return self.getToken(jacParser.KW_ASSERT, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_assert_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAssert_stmt"):
                listener.enterAssert_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAssert_stmt"):
                listener.exitAssert_stmt(self)

    def assert_stmt(self):

        localctx = jacParser.Assert_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 82, self.RULE_assert_stmt)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 702
            self.match(jacParser.KW_ASSERT)
            self.state = 703
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Destroy_actionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_DESTROY(self):
            return self.getToken(jacParser.KW_DESTROY, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_destroy_action

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDestroy_action"):
                listener.enterDestroy_action(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDestroy_action"):
                listener.exitDestroy_action(self)

    def destroy_action(self):

        localctx = jacParser.Destroy_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 84, self.RULE_destroy_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 705
            self.match(jacParser.KW_DESTROY)
            self.state = 706
            self.expression()
            self.state = 707
            self.match(jacParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Report_actionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_REPORT(self):
            return self.getToken(jacParser.KW_REPORT, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def EQ(self):
            return self.getToken(jacParser.EQ, 0)

        def getRuleIndex(self):
            return jacParser.RULE_report_action

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterReport_action"):
                listener.enterReport_action(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitReport_action"):
                listener.exitReport_action(self)

    def report_action(self):

        localctx = jacParser.Report_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 86, self.RULE_report_action)
        try:
            self.state = 720
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 70, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 709
                self.match(jacParser.KW_REPORT)
                self.state = 710
                self.expression()
                self.state = 711
                self.match(jacParser.SEMI)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 713
                self.match(jacParser.KW_REPORT)
                self.state = 714
                self.match(jacParser.COLON)
                self.state = 715
                self.match(jacParser.NAME)
                self.state = 716
                self.match(jacParser.EQ)
                self.state = 717
                self.expression()
                self.state = 718
                self.match(jacParser.SEMI)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Walker_actionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ignore_action(self):
            return self.getTypedRuleContext(jacParser.Ignore_actionContext, 0)

        def take_action(self):
            return self.getTypedRuleContext(jacParser.Take_actionContext, 0)

        def disengage_action(self):
            return self.getTypedRuleContext(jacParser.Disengage_actionContext, 0)

        def yield_action(self):
            return self.getTypedRuleContext(jacParser.Yield_actionContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_walker_action

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWalker_action"):
                listener.enterWalker_action(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWalker_action"):
                listener.exitWalker_action(self)

    def walker_action(self):

        localctx = jacParser.Walker_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 88, self.RULE_walker_action)
        try:
            self.state = 726
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_IGNORE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 722
                self.ignore_action()
                pass
            elif token in [jacParser.KW_TAKE]:
                self.enterOuterAlt(localctx, 2)
                self.state = 723
                self.take_action()
                pass
            elif token in [jacParser.KW_DISENGAGE]:
                self.enterOuterAlt(localctx, 3)
                self.state = 724
                self.disengage_action()
                pass
            elif token in [jacParser.KW_YIELD]:
                self.enterOuterAlt(localctx, 4)
                self.state = 725
                self.yield_action()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ignore_actionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_IGNORE(self):
            return self.getToken(jacParser.KW_IGNORE, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_ignore_action

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIgnore_action"):
                listener.enterIgnore_action(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIgnore_action"):
                listener.exitIgnore_action(self)

    def ignore_action(self):

        localctx = jacParser.Ignore_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 90, self.RULE_ignore_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 728
            self.match(jacParser.KW_IGNORE)
            self.state = 729
            self.expression()
            self.state = 730
            self.match(jacParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Take_actionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_TAKE(self):
            return self.getToken(jacParser.KW_TAKE, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def else_stmt(self):
            return self.getTypedRuleContext(jacParser.Else_stmtContext, 0)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def getRuleIndex(self):
            return jacParser.RULE_take_action

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTake_action"):
                listener.enterTake_action(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTake_action"):
                listener.exitTake_action(self)

    def take_action(self):

        localctx = jacParser.Take_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 92, self.RULE_take_action)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 732
            self.match(jacParser.KW_TAKE)
            self.state = 735
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.COLON:
                self.state = 733
                self.match(jacParser.COLON)
                self.state = 734
                self.match(jacParser.NAME)

            self.state = 737
            self.expression()
            self.state = 740
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.SEMI]:
                self.state = 738
                self.match(jacParser.SEMI)
                pass
            elif token in [jacParser.KW_ELSE]:
                self.state = 739
                self.else_stmt()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Disengage_actionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_DISENGAGE(self):
            return self.getToken(jacParser.KW_DISENGAGE, 0)

        def report_action(self):
            return self.getTypedRuleContext(jacParser.Report_actionContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_disengage_action

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDisengage_action"):
                listener.enterDisengage_action(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDisengage_action"):
                listener.exitDisengage_action(self)

    def disengage_action(self):

        localctx = jacParser.Disengage_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 94, self.RULE_disengage_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 742
            self.match(jacParser.KW_DISENGAGE)
            self.state = 745
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_REPORT]:
                self.state = 743
                self.report_action()
                pass
            elif token in [jacParser.SEMI]:
                self.state = 744
                self.match(jacParser.SEMI)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Yield_actionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_YIELD(self):
            return self.getToken(jacParser.KW_YIELD, 0)

        def report_action(self):
            return self.getTypedRuleContext(jacParser.Report_actionContext, 0)

        def disengage_action(self):
            return self.getTypedRuleContext(jacParser.Disengage_actionContext, 0)

        def take_action(self):
            return self.getTypedRuleContext(jacParser.Take_actionContext, 0)

        def SEMI(self):
            return self.getToken(jacParser.SEMI, 0)

        def getRuleIndex(self):
            return jacParser.RULE_yield_action

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterYield_action"):
                listener.enterYield_action(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitYield_action"):
                listener.exitYield_action(self)

    def yield_action(self):

        localctx = jacParser.Yield_actionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 96, self.RULE_yield_action)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 747
            self.match(jacParser.KW_YIELD)
            self.state = 752
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_REPORT]:
                self.state = 748
                self.report_action()
                pass
            elif token in [jacParser.KW_DISENGAGE]:
                self.state = 749
                self.disengage_action()
                pass
            elif token in [jacParser.KW_TAKE]:
                self.state = 750
                self.take_action()
                pass
            elif token in [jacParser.SEMI]:
                self.state = 751
                self.match(jacParser.SEMI)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExpressionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def connect(self):
            return self.getTypedRuleContext(jacParser.ConnectContext, 0)

        def assignment(self):
            return self.getTypedRuleContext(jacParser.AssignmentContext, 0)

        def copy_assign(self):
            return self.getTypedRuleContext(jacParser.Copy_assignContext, 0)

        def inc_assign(self):
            return self.getTypedRuleContext(jacParser.Inc_assignContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterExpression"):
                listener.enterExpression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitExpression"):
                listener.exitExpression(self)

    def expression(self):

        localctx = jacParser.ExpressionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 98, self.RULE_expression)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 754
            self.connect()
            self.state = 758
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 76, self._ctx)
            if la_ == 1:
                self.state = 755
                self.assignment()

            elif la_ == 2:
                self.state = 756
                self.copy_assign()

            elif la_ == 3:
                self.state = 757
                self.inc_assign()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AssignmentContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EQ(self):
            return self.getToken(jacParser.EQ, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_assignment

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAssignment"):
                listener.enterAssignment(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAssignment"):
                listener.exitAssignment(self)

    def assignment(self):

        localctx = jacParser.AssignmentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 100, self.RULE_assignment)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 760
            self.match(jacParser.EQ)
            self.state = 761
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Copy_assignContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CPY_EQ(self):
            return self.getToken(jacParser.CPY_EQ, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_copy_assign

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCopy_assign"):
                listener.enterCopy_assign(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCopy_assign"):
                listener.exitCopy_assign(self)

    def copy_assign(self):

        localctx = jacParser.Copy_assignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 102, self.RULE_copy_assign)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 763
            self.match(jacParser.CPY_EQ)
            self.state = 764
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Inc_assignContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def PEQ(self):
            return self.getToken(jacParser.PEQ, 0)

        def MEQ(self):
            return self.getToken(jacParser.MEQ, 0)

        def TEQ(self):
            return self.getToken(jacParser.TEQ, 0)

        def DEQ(self):
            return self.getToken(jacParser.DEQ, 0)

        def getRuleIndex(self):
            return jacParser.RULE_inc_assign

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterInc_assign"):
                listener.enterInc_assign(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitInc_assign"):
                listener.exitInc_assign(self)

    def inc_assign(self):

        localctx = jacParser.Inc_assignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 104, self.RULE_inc_assign)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 766
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.PEQ)
                            | (1 << jacParser.MEQ)
                            | (1 << jacParser.TEQ)
                            | (1 << jacParser.DEQ)
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 767
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ConnectContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def logical(self):
            return self.getTypedRuleContext(jacParser.LogicalContext, 0)

        def edge_ref(self):
            return self.getTypedRuleContext(jacParser.Edge_refContext, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def NOT(self):
            return self.getToken(jacParser.NOT, 0)

        def getRuleIndex(self):
            return jacParser.RULE_connect

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterConnect"):
                listener.enterConnect(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitConnect"):
                listener.exitConnect(self)

    def connect(self):

        localctx = jacParser.ConnectContext(self, self._ctx, self.state)
        self.enterRule(localctx, 106, self.RULE_connect)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 769
            self.logical()
            self.state = 776
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 78, self._ctx)
            if la_ == 1:
                self.state = 771
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.NOT:
                    self.state = 770
                    self.match(jacParser.NOT)

                self.state = 773
                self.edge_ref()
                self.state = 774
                self.expression()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LogicalContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def compare(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.CompareContext)
            else:
                return self.getTypedRuleContext(jacParser.CompareContext, i)

        def KW_AND(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.KW_AND)
            else:
                return self.getToken(jacParser.KW_AND, i)

        def KW_OR(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.KW_OR)
            else:
                return self.getToken(jacParser.KW_OR, i)

        def getRuleIndex(self):
            return jacParser.RULE_logical

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterLogical"):
                listener.enterLogical(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitLogical"):
                listener.exitLogical(self)

    def logical(self):

        localctx = jacParser.LogicalContext(self, self._ctx, self.state)
        self.enterRule(localctx, 108, self.RULE_logical)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 778
            self.compare()
            self.state = 783
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la == jacParser.KW_AND or _la == jacParser.KW_OR:
                self.state = 779
                _la = self._input.LA(1)
                if not (_la == jacParser.KW_AND or _la == jacParser.KW_OR):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 780
                self.compare()
                self.state = 785
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class CompareContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NOT(self):
            return self.getToken(jacParser.NOT, 0)

        def compare(self):
            return self.getTypedRuleContext(jacParser.CompareContext, 0)

        def arithmetic(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.ArithmeticContext)
            else:
                return self.getTypedRuleContext(jacParser.ArithmeticContext, i)

        def cmp_op(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Cmp_opContext)
            else:
                return self.getTypedRuleContext(jacParser.Cmp_opContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_compare

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCompare"):
                listener.enterCompare(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCompare"):
                listener.exitCompare(self)

    def compare(self):

        localctx = jacParser.CompareContext(self, self._ctx, self.state)
        self.enterRule(localctx, 110, self.RULE_compare)
        try:
            self.state = 797
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.NOT]:
                self.enterOuterAlt(localctx, 1)
                self.state = 786
                self.match(jacParser.NOT)
                self.state = 787
                self.compare()
                pass
            elif token in [
                jacParser.T__1,
                jacParser.T__2,
                jacParser.T__4,
                jacParser.T__5,
                jacParser.T__6,
                jacParser.TYP_STRING,
                jacParser.TYP_INT,
                jacParser.TYP_FLOAT,
                jacParser.TYP_LIST,
                jacParser.TYP_DICT,
                jacParser.TYP_BOOL,
                jacParser.KW_TYPE,
                jacParser.KW_NODE,
                jacParser.KW_SPAWN,
                jacParser.DBL_COLON,
                jacParser.LBRACE,
                jacParser.KW_EDGE,
                jacParser.KW_SYNC,
                jacParser.KW_GLOBAL,
                jacParser.PLUS,
                jacParser.MINUS,
                jacParser.STAR_MUL,
                jacParser.LPAREN,
                jacParser.LSQUARE,
                jacParser.FLOAT,
                jacParser.STRING,
                jacParser.BOOL,
                jacParser.INT,
                jacParser.NULL,
                jacParser.NAME,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 788
                self.arithmetic()
                self.state = 794
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 80, self._ctx)
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 789
                        self.cmp_op()
                        self.state = 790
                        self.arithmetic()
                    self.state = 796
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 80, self._ctx)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Cmp_opContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EE(self):
            return self.getToken(jacParser.EE, 0)

        def LT(self):
            return self.getToken(jacParser.LT, 0)

        def GT(self):
            return self.getToken(jacParser.GT, 0)

        def LTE(self):
            return self.getToken(jacParser.LTE, 0)

        def GTE(self):
            return self.getToken(jacParser.GTE, 0)

        def NE(self):
            return self.getToken(jacParser.NE, 0)

        def KW_IN(self):
            return self.getToken(jacParser.KW_IN, 0)

        def nin(self):
            return self.getTypedRuleContext(jacParser.NinContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_cmp_op

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCmp_op"):
                listener.enterCmp_op(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCmp_op"):
                listener.exitCmp_op(self)

    def cmp_op(self):

        localctx = jacParser.Cmp_opContext(self, self._ctx, self.state)
        self.enterRule(localctx, 112, self.RULE_cmp_op)
        try:
            self.state = 807
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.EE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 799
                self.match(jacParser.EE)
                pass
            elif token in [jacParser.LT]:
                self.enterOuterAlt(localctx, 2)
                self.state = 800
                self.match(jacParser.LT)
                pass
            elif token in [jacParser.GT]:
                self.enterOuterAlt(localctx, 3)
                self.state = 801
                self.match(jacParser.GT)
                pass
            elif token in [jacParser.LTE]:
                self.enterOuterAlt(localctx, 4)
                self.state = 802
                self.match(jacParser.LTE)
                pass
            elif token in [jacParser.GTE]:
                self.enterOuterAlt(localctx, 5)
                self.state = 803
                self.match(jacParser.GTE)
                pass
            elif token in [jacParser.NE]:
                self.enterOuterAlt(localctx, 6)
                self.state = 804
                self.match(jacParser.NE)
                pass
            elif token in [jacParser.KW_IN]:
                self.enterOuterAlt(localctx, 7)
                self.state = 805
                self.match(jacParser.KW_IN)
                pass
            elif token in [jacParser.NOT]:
                self.enterOuterAlt(localctx, 8)
                self.state = 806
                self.nin()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NinContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NOT(self):
            return self.getToken(jacParser.NOT, 0)

        def KW_IN(self):
            return self.getToken(jacParser.KW_IN, 0)

        def getRuleIndex(self):
            return jacParser.RULE_nin

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNin"):
                listener.enterNin(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNin"):
                listener.exitNin(self)

    def nin(self):

        localctx = jacParser.NinContext(self, self._ctx, self.state)
        self.enterRule(localctx, 114, self.RULE_nin)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 809
            self.match(jacParser.NOT)
            self.state = 810
            self.match(jacParser.KW_IN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArithmeticContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def term(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.TermContext)
            else:
                return self.getTypedRuleContext(jacParser.TermContext, i)

        def PLUS(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.PLUS)
            else:
                return self.getToken(jacParser.PLUS, i)

        def MINUS(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.MINUS)
            else:
                return self.getToken(jacParser.MINUS, i)

        def getRuleIndex(self):
            return jacParser.RULE_arithmetic

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterArithmetic"):
                listener.enterArithmetic(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitArithmetic"):
                listener.exitArithmetic(self)

    def arithmetic(self):

        localctx = jacParser.ArithmeticContext(self, self._ctx, self.state)
        self.enterRule(localctx, 116, self.RULE_arithmetic)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 812
            self.term()
            self.state = 817
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 83, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 813
                    _la = self._input.LA(1)
                    if not (_la == jacParser.PLUS or _la == jacParser.MINUS):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()
                    self.state = 814
                    self.term()
                self.state = 819
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 83, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TermContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def factor(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.FactorContext)
            else:
                return self.getTypedRuleContext(jacParser.FactorContext, i)

        def STAR_MUL(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.STAR_MUL)
            else:
                return self.getToken(jacParser.STAR_MUL, i)

        def DIV(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.DIV)
            else:
                return self.getToken(jacParser.DIV, i)

        def MOD(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.MOD)
            else:
                return self.getToken(jacParser.MOD, i)

        def getRuleIndex(self):
            return jacParser.RULE_term

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterTerm"):
                listener.enterTerm(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitTerm"):
                listener.exitTerm(self)

    def term(self):

        localctx = jacParser.TermContext(self, self._ctx, self.state)
        self.enterRule(localctx, 118, self.RULE_term)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 820
            self.factor()
            self.state = 825
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la - 89)) & ~0x3F) == 0 and (
                (1 << (_la - 89))
                & (
                    (1 << (jacParser.STAR_MUL - 89))
                    | (1 << (jacParser.DIV - 89))
                    | (1 << (jacParser.MOD - 89))
                )
            ) != 0:
                self.state = 821
                _la = self._input.LA(1)
                if not (
                    (
                        (((_la - 89)) & ~0x3F) == 0
                        and (
                            (1 << (_la - 89))
                            & (
                                (1 << (jacParser.STAR_MUL - 89))
                                | (1 << (jacParser.DIV - 89))
                                | (1 << (jacParser.MOD - 89))
                            )
                        )
                        != 0
                    )
                ):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 822
                self.factor()
                self.state = 827
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FactorContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def factor(self):
            return self.getTypedRuleContext(jacParser.FactorContext, 0)

        def PLUS(self):
            return self.getToken(jacParser.PLUS, 0)

        def MINUS(self):
            return self.getToken(jacParser.MINUS, 0)

        def power(self):
            return self.getTypedRuleContext(jacParser.PowerContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_factor

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFactor"):
                listener.enterFactor(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFactor"):
                listener.exitFactor(self)

    def factor(self):

        localctx = jacParser.FactorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 120, self.RULE_factor)
        self._la = 0  # Token type
        try:
            self.state = 831
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 85, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 828
                _la = self._input.LA(1)
                if not (_la == jacParser.PLUS or _la == jacParser.MINUS):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 829
                self.factor()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 830
                self.power()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PowerContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def atom(self):
            return self.getTypedRuleContext(jacParser.AtomContext, 0)

        def POW(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.POW)
            else:
                return self.getToken(jacParser.POW, i)

        def factor(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.FactorContext)
            else:
                return self.getTypedRuleContext(jacParser.FactorContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_power

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPower"):
                listener.enterPower(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPower"):
                listener.exitPower(self)

    def power(self):

        localctx = jacParser.PowerContext(self, self._ctx, self.state)
        self.enterRule(localctx, 122, self.RULE_power)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 833
            self.atom(0)
            self.state = 838
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 86, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 834
                    self.match(jacParser.POW)
                    self.state = 835
                    self.factor()
                self.state = 840
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 86, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Global_refContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_GLOBAL(self):
            return self.getToken(jacParser.KW_GLOBAL, 0)

        def DOT(self):
            return self.getToken(jacParser.DOT, 0)

        def obj_built_in(self):
            return self.getTypedRuleContext(jacParser.Obj_built_inContext, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def getRuleIndex(self):
            return jacParser.RULE_global_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGlobal_ref"):
                listener.enterGlobal_ref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGlobal_ref"):
                listener.exitGlobal_ref(self)

    def global_ref(self):

        localctx = jacParser.Global_refContext(self, self._ctx, self.state)
        self.enterRule(localctx, 124, self.RULE_global_ref)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 841
            self.match(jacParser.KW_GLOBAL)
            self.state = 842
            self.match(jacParser.DOT)
            self.state = 845
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_CONTEXT, jacParser.KW_INFO, jacParser.KW_DETAILS]:
                self.state = 843
                self.obj_built_in()
                pass
            elif token in [jacParser.NAME]:
                self.state = 844
                self.match(jacParser.NAME)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AtomContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(jacParser.INT, 0)

        def FLOAT(self):
            return self.getToken(jacParser.FLOAT, 0)

        def STRING(self):
            return self.getToken(jacParser.STRING, 0)

        def BOOL(self):
            return self.getToken(jacParser.BOOL, 0)

        def NULL(self):
            return self.getToken(jacParser.NULL, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def global_ref(self):
            return self.getTypedRuleContext(jacParser.Global_refContext, 0)

        def node_edge_ref(self):
            return self.getTypedRuleContext(jacParser.Node_edge_refContext, 0)

        def type_ref(self):
            return self.getTypedRuleContext(jacParser.Type_refContext, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def list_val(self):
            return self.getTypedRuleContext(jacParser.List_valContext, 0)

        def dict_val(self):
            return self.getTypedRuleContext(jacParser.Dict_valContext, 0)

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def ability_op(self):
            return self.getTypedRuleContext(jacParser.Ability_opContext, 0)

        def KW_SYNC(self):
            return self.getToken(jacParser.KW_SYNC, 0)

        def atom(self):
            return self.getTypedRuleContext(jacParser.AtomContext, 0)

        def spawn(self):
            return self.getTypedRuleContext(jacParser.SpawnContext, 0)

        def ref(self):
            return self.getTypedRuleContext(jacParser.RefContext, 0)

        def deref(self):
            return self.getTypedRuleContext(jacParser.DerefContext, 0)

        def any_type(self):
            return self.getTypedRuleContext(jacParser.Any_typeContext, 0)

        def atom_trailer(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Atom_trailerContext)
            else:
                return self.getTypedRuleContext(jacParser.Atom_trailerContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_atom

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAtom"):
                listener.enterAtom(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAtom"):
                listener.exitAtom(self)

    def atom(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = jacParser.AtomContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 126
        self.enterRecursionRule(localctx, 126, self.RULE_atom, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 877
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 90, self._ctx)
            if la_ == 1:
                self.state = 848
                self.match(jacParser.INT)
                pass

            elif la_ == 2:
                self.state = 849
                self.match(jacParser.FLOAT)
                pass

            elif la_ == 3:
                self.state = 850
                self.match(jacParser.STRING)
                pass

            elif la_ == 4:
                self.state = 851
                self.match(jacParser.BOOL)
                pass

            elif la_ == 5:
                self.state = 852
                self.match(jacParser.NULL)
                pass

            elif la_ == 6:
                self.state = 853
                self.match(jacParser.NAME)
                pass

            elif la_ == 7:
                self.state = 854
                self.global_ref()
                pass

            elif la_ == 8:
                self.state = 855
                self.node_edge_ref()
                pass

            elif la_ == 9:
                self.state = 856
                self.type_ref()
                self.state = 858
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 88, self._ctx)
                if la_ == 1:
                    self.state = 857
                    self.spawn_ctx()

                pass

            elif la_ == 10:
                self.state = 860
                self.list_val()
                pass

            elif la_ == 11:
                self.state = 861
                self.dict_val()
                pass

            elif la_ == 12:
                self.state = 862
                self.match(jacParser.LPAREN)
                self.state = 863
                self.expression()
                self.state = 864
                self.match(jacParser.RPAREN)
                pass

            elif la_ == 13:
                self.state = 866
                self.ability_op()
                self.state = 867
                self.match(jacParser.NAME)
                self.state = 869
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 89, self._ctx)
                if la_ == 1:
                    self.state = 868
                    self.spawn_ctx()

                pass

            elif la_ == 14:
                self.state = 871
                self.match(jacParser.KW_SYNC)
                self.state = 872
                self.atom(5)
                pass

            elif la_ == 15:
                self.state = 873
                self.spawn()
                pass

            elif la_ == 16:
                self.state = 874
                self.ref()
                pass

            elif la_ == 17:
                self.state = 875
                self.deref()
                pass

            elif la_ == 18:
                self.state = 876
                self.any_type()
                pass

            self._ctx.stop = self._input.LT(-1)
            self.state = 887
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 92, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = jacParser.AtomContext(self, _parentctx, _parentState)
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_atom)
                    self.state = 879
                    if not self.precpred(self._ctx, 6):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(
                            self, "self.precpred(self._ctx, 6)"
                        )
                    self.state = 881
                    self._errHandler.sync(self)
                    _alt = 1
                    while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                        if _alt == 1:
                            self.state = 880
                            self.atom_trailer()

                        else:
                            raise NoViableAltException(self)
                        self.state = 883
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input, 91, self._ctx)

                self.state = 889
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 92, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Atom_trailerContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DOT(self):
            return self.getToken(jacParser.DOT, 0)

        def built_in(self):
            return self.getTypedRuleContext(jacParser.Built_inContext, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def index_slice(self):
            return self.getTypedRuleContext(jacParser.Index_sliceContext, 0)

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def expr_list(self):
            return self.getTypedRuleContext(jacParser.Expr_listContext, 0)

        def ability_op(self):
            return self.getTypedRuleContext(jacParser.Ability_opContext, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_atom_trailer

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAtom_trailer"):
                listener.enterAtom_trailer(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAtom_trailer"):
                listener.exitAtom_trailer(self)

    def atom_trailer(self):

        localctx = jacParser.Atom_trailerContext(self, self._ctx, self.state)
        self.enterRule(localctx, 128, self.RULE_atom_trailer)
        self._la = 0  # Token type
        try:
            self.state = 905
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 95, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 890
                self.match(jacParser.DOT)
                self.state = 891
                self.built_in()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 892
                self.match(jacParser.DOT)
                self.state = 893
                self.match(jacParser.NAME)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 894
                self.index_slice()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 895
                self.match(jacParser.LPAREN)
                self.state = 897
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.T__1)
                            | (1 << jacParser.T__2)
                            | (1 << jacParser.T__4)
                            | (1 << jacParser.T__5)
                            | (1 << jacParser.T__6)
                            | (1 << jacParser.TYP_STRING)
                            | (1 << jacParser.TYP_INT)
                            | (1 << jacParser.TYP_FLOAT)
                            | (1 << jacParser.TYP_LIST)
                            | (1 << jacParser.TYP_DICT)
                            | (1 << jacParser.TYP_BOOL)
                            | (1 << jacParser.KW_TYPE)
                            | (1 << jacParser.KW_NODE)
                            | (1 << jacParser.KW_SPAWN)
                            | (1 << jacParser.DBL_COLON)
                            | (1 << jacParser.LBRACE)
                            | (1 << jacParser.KW_EDGE)
                            | (1 << jacParser.KW_SYNC)
                        )
                    )
                    != 0
                ) or (
                    (((_la - 73)) & ~0x3F) == 0
                    and (
                        (1 << (_la - 73))
                        & (
                            (1 << (jacParser.NOT - 73))
                            | (1 << (jacParser.KW_GLOBAL - 73))
                            | (1 << (jacParser.PLUS - 73))
                            | (1 << (jacParser.MINUS - 73))
                            | (1 << (jacParser.STAR_MUL - 73))
                            | (1 << (jacParser.LPAREN - 73))
                            | (1 << (jacParser.LSQUARE - 73))
                            | (1 << (jacParser.FLOAT - 73))
                            | (1 << (jacParser.STRING - 73))
                            | (1 << (jacParser.BOOL - 73))
                            | (1 << (jacParser.INT - 73))
                            | (1 << (jacParser.NULL - 73))
                            | (1 << (jacParser.NAME - 73))
                        )
                    )
                    != 0
                ):
                    self.state = 896
                    self.expr_list()

                self.state = 899
                self.match(jacParser.RPAREN)
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 900
                self.ability_op()
                self.state = 901
                self.match(jacParser.NAME)
                self.state = 903
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 94, self._ctx)
                if la_ == 1:
                    self.state = 902
                    self.spawn_ctx()

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Ability_opContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def getRuleIndex(self):
            return jacParser.RULE_ability_op

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAbility_op"):
                listener.enterAbility_op(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAbility_op"):
                listener.exitAbility_op(self)

    def ability_op(self):

        localctx = jacParser.Ability_opContext(self, self._ctx, self.state)
        self.enterRule(localctx, 130, self.RULE_ability_op)
        try:
            self.state = 911
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 96, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 907
                self.match(jacParser.DBL_COLON)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 908
                self.match(jacParser.DBL_COLON)
                self.state = 909
                self.match(jacParser.NAME)
                self.state = 910
                self.match(jacParser.COLON)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RefContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def atom(self):
            return self.getTypedRuleContext(jacParser.AtomContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterRef"):
                listener.enterRef(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitRef"):
                listener.exitRef(self)

    def ref(self):

        localctx = jacParser.RefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 132, self.RULE_ref)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 913
            self.match(jacParser.T__1)
            self.state = 914
            self.atom(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DerefContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR_MUL(self):
            return self.getToken(jacParser.STAR_MUL, 0)

        def atom(self):
            return self.getTypedRuleContext(jacParser.AtomContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_deref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDeref"):
                listener.enterDeref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDeref"):
                listener.exitDeref(self)

    def deref(self):

        localctx = jacParser.DerefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 134, self.RULE_deref)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 916
            self.match(jacParser.STAR_MUL)
            self.state = 917
            self.atom(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Built_inContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def string_built_in(self):
            return self.getTypedRuleContext(jacParser.String_built_inContext, 0)

        def dict_built_in(self):
            return self.getTypedRuleContext(jacParser.Dict_built_inContext, 0)

        def list_built_in(self):
            return self.getTypedRuleContext(jacParser.List_built_inContext, 0)

        def obj_built_in(self):
            return self.getTypedRuleContext(jacParser.Obj_built_inContext, 0)

        def cast_built_in(self):
            return self.getTypedRuleContext(jacParser.Cast_built_inContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_built_in

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterBuilt_in"):
                listener.enterBuilt_in(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitBuilt_in"):
                listener.exitBuilt_in(self)

    def built_in(self):

        localctx = jacParser.Built_inContext(self, self._ctx, self.state)
        self.enterRule(localctx, 136, self.RULE_built_in)
        try:
            self.state = 925
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 97, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 920
                self.string_built_in()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 921
                self.dict_built_in()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 922
                self.list_built_in()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 923
                self.obj_built_in()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 924
                self.cast_built_in()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Cast_built_inContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def any_type(self):
            return self.getTypedRuleContext(jacParser.Any_typeContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_cast_built_in

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterCast_built_in"):
                listener.enterCast_built_in(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitCast_built_in"):
                listener.exitCast_built_in(self)

    def cast_built_in(self):

        localctx = jacParser.Cast_built_inContext(self, self._ctx, self.state)
        self.enterRule(localctx, 138, self.RULE_cast_built_in)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 927
            self.any_type()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Obj_built_inContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_CONTEXT(self):
            return self.getToken(jacParser.KW_CONTEXT, 0)

        def KW_INFO(self):
            return self.getToken(jacParser.KW_INFO, 0)

        def KW_DETAILS(self):
            return self.getToken(jacParser.KW_DETAILS, 0)

        def getRuleIndex(self):
            return jacParser.RULE_obj_built_in

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterObj_built_in"):
                listener.enterObj_built_in(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitObj_built_in"):
                listener.exitObj_built_in(self)

    def obj_built_in(self):

        localctx = jacParser.Obj_built_inContext(self, self._ctx, self.state)
        self.enterRule(localctx, 140, self.RULE_obj_built_in)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 929
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.KW_CONTEXT)
                            | (1 << jacParser.KW_INFO)
                            | (1 << jacParser.KW_DETAILS)
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dict_built_inContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_KEYS(self):
            return self.getToken(jacParser.KW_KEYS, 0)

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def name_list(self):
            return self.getTypedRuleContext(jacParser.Name_listContext, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def TYP_DICT(self):
            return self.getToken(jacParser.TYP_DICT, 0)

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def DICT_DBL_COLON(self):
            return self.getToken(jacParser.DICT_DBL_COLON, 0)

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def expr_list(self):
            return self.getTypedRuleContext(jacParser.Expr_listContext, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dict_built_in

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDict_built_in"):
                listener.enterDict_built_in(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDict_built_in"):
                listener.exitDict_built_in(self)

    def dict_built_in(self):

        localctx = jacParser.Dict_built_inContext(self, self._ctx, self.state)
        self.enterRule(localctx, 142, self.RULE_dict_built_in)
        self._la = 0  # Token type
        try:
            self.state = 948
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_KEYS]:
                self.enterOuterAlt(localctx, 1)
                self.state = 931
                self.match(jacParser.KW_KEYS)
                pass
            elif token in [jacParser.LBRACE]:
                self.enterOuterAlt(localctx, 2)
                self.state = 932
                self.match(jacParser.LBRACE)
                self.state = 933
                self.name_list()
                self.state = 934
                self.match(jacParser.RBRACE)
                pass
            elif token in [jacParser.TYP_DICT, jacParser.DICT_DBL_COLON]:
                self.enterOuterAlt(localctx, 3)
                self.state = 939
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.TYP_DICT]:
                    self.state = 936
                    self.match(jacParser.TYP_DICT)
                    self.state = 937
                    self.match(jacParser.DBL_COLON)
                    pass
                elif token in [jacParser.DICT_DBL_COLON]:
                    self.state = 938
                    self.match(jacParser.DICT_DBL_COLON)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 941
                _la = self._input.LA(1)
                if not (_la == jacParser.KW_KEYS or _la == jacParser.NAME):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 946
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 99, self._ctx)
                if la_ == 1:
                    self.state = 942
                    self.match(jacParser.LPAREN)
                    self.state = 943
                    self.expr_list()
                    self.state = 944
                    self.match(jacParser.RPAREN)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class List_built_inContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_LENGTH(self):
            return self.getToken(jacParser.KW_LENGTH, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def TYP_LIST(self):
            return self.getToken(jacParser.TYP_LIST, 0)

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def LIST_DBL_COLON(self):
            return self.getToken(jacParser.LIST_DBL_COLON, 0)

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def expr_list(self):
            return self.getTypedRuleContext(jacParser.Expr_listContext, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def getRuleIndex(self):
            return jacParser.RULE_list_built_in

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterList_built_in"):
                listener.enterList_built_in(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitList_built_in"):
                listener.exitList_built_in(self)

    def list_built_in(self):

        localctx = jacParser.List_built_inContext(self, self._ctx, self.state)
        self.enterRule(localctx, 144, self.RULE_list_built_in)
        try:
            self.state = 963
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_LENGTH]:
                self.enterOuterAlt(localctx, 1)
                self.state = 950
                self.match(jacParser.KW_LENGTH)
                pass
            elif token in [jacParser.TYP_LIST, jacParser.LIST_DBL_COLON]:
                self.enterOuterAlt(localctx, 2)
                self.state = 954
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [jacParser.TYP_LIST]:
                    self.state = 951
                    self.match(jacParser.TYP_LIST)
                    self.state = 952
                    self.match(jacParser.DBL_COLON)
                    pass
                elif token in [jacParser.LIST_DBL_COLON]:
                    self.state = 953
                    self.match(jacParser.LIST_DBL_COLON)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 956
                self.match(jacParser.NAME)
                self.state = 961
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 102, self._ctx)
                if la_ == 1:
                    self.state = 957
                    self.match(jacParser.LPAREN)
                    self.state = 958
                    self.expr_list()
                    self.state = 959
                    self.match(jacParser.RPAREN)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class String_built_inContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def TYP_STRING(self):
            return self.getToken(jacParser.TYP_STRING, 0)

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def STR_DBL_COLON(self):
            return self.getToken(jacParser.STR_DBL_COLON, 0)

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def expr_list(self):
            return self.getTypedRuleContext(jacParser.Expr_listContext, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def getRuleIndex(self):
            return jacParser.RULE_string_built_in

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterString_built_in"):
                listener.enterString_built_in(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitString_built_in"):
                listener.exitString_built_in(self)

    def string_built_in(self):

        localctx = jacParser.String_built_inContext(self, self._ctx, self.state)
        self.enterRule(localctx, 146, self.RULE_string_built_in)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 968
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.TYP_STRING]:
                self.state = 965
                self.match(jacParser.TYP_STRING)
                self.state = 966
                self.match(jacParser.DBL_COLON)
                pass
            elif token in [jacParser.STR_DBL_COLON]:
                self.state = 967
                self.match(jacParser.STR_DBL_COLON)
                pass
            else:
                raise NoViableAltException(self)

            self.state = 970
            self.match(jacParser.NAME)
            self.state = 975
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 105, self._ctx)
            if la_ == 1:
                self.state = 971
                self.match(jacParser.LPAREN)
                self.state = 972
                self.expr_list()
                self.state = 973
                self.match(jacParser.RPAREN)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Node_edge_refContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def node_ref(self):
            return self.getTypedRuleContext(jacParser.Node_refContext, 0)

        def filter_ctx(self):
            return self.getTypedRuleContext(jacParser.Filter_ctxContext, 0)

        def edge_ref(self):
            return self.getTypedRuleContext(jacParser.Edge_refContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_node_edge_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNode_edge_ref"):
                listener.enterNode_edge_ref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNode_edge_ref"):
                listener.exitNode_edge_ref(self)

    def node_edge_ref(self):

        localctx = jacParser.Node_edge_refContext(self, self._ctx, self.state)
        self.enterRule(localctx, 148, self.RULE_node_edge_ref)
        try:
            self.state = 988
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.KW_NODE]:
                self.enterOuterAlt(localctx, 1)
                self.state = 977
                self.node_ref()
                self.state = 979
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 106, self._ctx)
                if la_ == 1:
                    self.state = 978
                    self.filter_ctx()

                pass
            elif token in [
                jacParser.T__2,
                jacParser.T__4,
                jacParser.T__5,
                jacParser.T__6,
                jacParser.MINUS,
            ]:
                self.enterOuterAlt(localctx, 2)
                self.state = 981
                self.edge_ref()
                self.state = 986
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 108, self._ctx)
                if la_ == 1:
                    self.state = 982
                    self.node_ref()
                    self.state = 984
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 107, self._ctx)
                    if la_ == 1:
                        self.state = 983
                        self.filter_ctx()

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Node_refContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_NODE(self):
            return self.getToken(jacParser.KW_NODE, 0)

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def getRuleIndex(self):
            return jacParser.RULE_node_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNode_ref"):
                listener.enterNode_ref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNode_ref"):
                listener.exitNode_ref(self)

    def node_ref(self):

        localctx = jacParser.Node_refContext(self, self._ctx, self.state)
        self.enterRule(localctx, 150, self.RULE_node_ref)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 990
            self.match(jacParser.KW_NODE)
            self.state = 991
            self.match(jacParser.DBL_COLON)
            self.state = 992
            self.match(jacParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Walker_refContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_WALKER(self):
            return self.getToken(jacParser.KW_WALKER, 0)

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def getRuleIndex(self):
            return jacParser.RULE_walker_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWalker_ref"):
                listener.enterWalker_ref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWalker_ref"):
                listener.exitWalker_ref(self)

    def walker_ref(self):

        localctx = jacParser.Walker_refContext(self, self._ctx, self.state)
        self.enterRule(localctx, 152, self.RULE_walker_ref)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 994
            self.match(jacParser.KW_WALKER)
            self.state = 995
            self.match(jacParser.DBL_COLON)
            self.state = 996
            self.match(jacParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Graph_refContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_GRAPH(self):
            return self.getToken(jacParser.KW_GRAPH, 0)

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def getRuleIndex(self):
            return jacParser.RULE_graph_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGraph_ref"):
                listener.enterGraph_ref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGraph_ref"):
                listener.exitGraph_ref(self)

    def graph_ref(self):

        localctx = jacParser.Graph_refContext(self, self._ctx, self.state)
        self.enterRule(localctx, 154, self.RULE_graph_ref)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 998
            self.match(jacParser.KW_GRAPH)
            self.state = 999
            self.match(jacParser.DBL_COLON)
            self.state = 1000
            self.match(jacParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Type_refContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_TYPE(self):
            return self.getToken(jacParser.KW_TYPE, 0)

        def DBL_COLON(self):
            return self.getToken(jacParser.DBL_COLON, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def getRuleIndex(self):
            return jacParser.RULE_type_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterType_ref"):
                listener.enterType_ref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitType_ref"):
                listener.exitType_ref(self)

    def type_ref(self):

        localctx = jacParser.Type_refContext(self, self._ctx, self.state)
        self.enterRule(localctx, 156, self.RULE_type_ref)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1002
            self.match(jacParser.KW_TYPE)
            self.state = 1003
            self.match(jacParser.DBL_COLON)
            self.state = 1004
            self.match(jacParser.NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Edge_refContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def edge_to(self):
            return self.getTypedRuleContext(jacParser.Edge_toContext, 0)

        def edge_from(self):
            return self.getTypedRuleContext(jacParser.Edge_fromContext, 0)

        def edge_any(self):
            return self.getTypedRuleContext(jacParser.Edge_anyContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_edge_ref

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEdge_ref"):
                listener.enterEdge_ref(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEdge_ref"):
                listener.exitEdge_ref(self)

    def edge_ref(self):

        localctx = jacParser.Edge_refContext(self, self._ctx, self.state)
        self.enterRule(localctx, 158, self.RULE_edge_ref)
        try:
            self.state = 1009
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 110, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1006
                self.edge_to()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1007
                self.edge_from()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1008
                self.edge_any()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Edge_toContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def MINUS(self):
            return self.getToken(jacParser.MINUS, 0)

        def LSQUARE(self):
            return self.getToken(jacParser.LSQUARE, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def RSQUARE(self):
            return self.getToken(jacParser.RSQUARE, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def filter_ctx(self):
            return self.getTypedRuleContext(jacParser.Filter_ctxContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_edge_to

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEdge_to"):
                listener.enterEdge_to(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEdge_to"):
                listener.exitEdge_to(self)

    def edge_to(self):

        localctx = jacParser.Edge_toContext(self, self._ctx, self.state)
        self.enterRule(localctx, 160, self.RULE_edge_to)
        self._la = 0  # Token type
        try:
            self.state = 1023
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.T__2]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1011
                self.match(jacParser.T__2)
                pass
            elif token in [jacParser.MINUS]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1012
                self.match(jacParser.MINUS)
                self.state = 1020
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.LSQUARE:
                    self.state = 1013
                    self.match(jacParser.LSQUARE)
                    self.state = 1014
                    self.match(jacParser.NAME)
                    self.state = 1017
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 111, self._ctx)
                    if la_ == 1:
                        self.state = 1015
                        self.spawn_ctx()

                    elif la_ == 2:
                        self.state = 1016
                        self.filter_ctx()

                    self.state = 1019
                    self.match(jacParser.RSQUARE)

                self.state = 1022
                self.match(jacParser.T__3)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Edge_fromContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def MINUS(self):
            return self.getToken(jacParser.MINUS, 0)

        def LSQUARE(self):
            return self.getToken(jacParser.LSQUARE, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def RSQUARE(self):
            return self.getToken(jacParser.RSQUARE, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def filter_ctx(self):
            return self.getTypedRuleContext(jacParser.Filter_ctxContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_edge_from

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEdge_from"):
                listener.enterEdge_from(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEdge_from"):
                listener.exitEdge_from(self)

    def edge_from(self):

        localctx = jacParser.Edge_fromContext(self, self._ctx, self.state)
        self.enterRule(localctx, 162, self.RULE_edge_from)
        self._la = 0  # Token type
        try:
            self.state = 1037
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.T__4]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1025
                self.match(jacParser.T__4)
                pass
            elif token in [jacParser.T__5]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1026
                self.match(jacParser.T__5)
                self.state = 1034
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.LSQUARE:
                    self.state = 1027
                    self.match(jacParser.LSQUARE)
                    self.state = 1028
                    self.match(jacParser.NAME)
                    self.state = 1031
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 114, self._ctx)
                    if la_ == 1:
                        self.state = 1029
                        self.spawn_ctx()

                    elif la_ == 2:
                        self.state = 1030
                        self.filter_ctx()

                    self.state = 1033
                    self.match(jacParser.RSQUARE)

                self.state = 1036
                self.match(jacParser.MINUS)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Edge_anyContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LSQUARE(self):
            return self.getToken(jacParser.LSQUARE, 0)

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def RSQUARE(self):
            return self.getToken(jacParser.RSQUARE, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def filter_ctx(self):
            return self.getTypedRuleContext(jacParser.Filter_ctxContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_edge_any

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterEdge_any"):
                listener.enterEdge_any(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitEdge_any"):
                listener.exitEdge_any(self)

    def edge_any(self):

        localctx = jacParser.Edge_anyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 164, self.RULE_edge_any)
        self._la = 0  # Token type
        try:
            self.state = 1051
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [jacParser.T__6]:
                self.enterOuterAlt(localctx, 1)
                self.state = 1039
                self.match(jacParser.T__6)
                pass
            elif token in [jacParser.T__5]:
                self.enterOuterAlt(localctx, 2)
                self.state = 1040
                self.match(jacParser.T__5)
                self.state = 1048
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.LSQUARE:
                    self.state = 1041
                    self.match(jacParser.LSQUARE)
                    self.state = 1042
                    self.match(jacParser.NAME)
                    self.state = 1045
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input, 117, self._ctx)
                    if la_ == 1:
                        self.state = 1043
                        self.spawn_ctx()

                    elif la_ == 2:
                        self.state = 1044
                        self.filter_ctx()

                    self.state = 1047
                    self.match(jacParser.RSQUARE)

                self.state = 1050
                self.match(jacParser.T__3)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class List_valContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LSQUARE(self):
            return self.getToken(jacParser.LSQUARE, 0)

        def RSQUARE(self):
            return self.getToken(jacParser.RSQUARE, 0)

        def expr_list(self):
            return self.getTypedRuleContext(jacParser.Expr_listContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_list_val

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterList_val"):
                listener.enterList_val(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitList_val"):
                listener.exitList_val(self)

    def list_val(self):

        localctx = jacParser.List_valContext(self, self._ctx, self.state)
        self.enterRule(localctx, 166, self.RULE_list_val)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1053
            self.match(jacParser.LSQUARE)
            self.state = 1055
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (
                ((_la) & ~0x3F) == 0
                and (
                    (1 << _la)
                    & (
                        (1 << jacParser.T__1)
                        | (1 << jacParser.T__2)
                        | (1 << jacParser.T__4)
                        | (1 << jacParser.T__5)
                        | (1 << jacParser.T__6)
                        | (1 << jacParser.TYP_STRING)
                        | (1 << jacParser.TYP_INT)
                        | (1 << jacParser.TYP_FLOAT)
                        | (1 << jacParser.TYP_LIST)
                        | (1 << jacParser.TYP_DICT)
                        | (1 << jacParser.TYP_BOOL)
                        | (1 << jacParser.KW_TYPE)
                        | (1 << jacParser.KW_NODE)
                        | (1 << jacParser.KW_SPAWN)
                        | (1 << jacParser.DBL_COLON)
                        | (1 << jacParser.LBRACE)
                        | (1 << jacParser.KW_EDGE)
                        | (1 << jacParser.KW_SYNC)
                    )
                )
                != 0
            ) or (
                (((_la - 73)) & ~0x3F) == 0
                and (
                    (1 << (_la - 73))
                    & (
                        (1 << (jacParser.NOT - 73))
                        | (1 << (jacParser.KW_GLOBAL - 73))
                        | (1 << (jacParser.PLUS - 73))
                        | (1 << (jacParser.MINUS - 73))
                        | (1 << (jacParser.STAR_MUL - 73))
                        | (1 << (jacParser.LPAREN - 73))
                        | (1 << (jacParser.LSQUARE - 73))
                        | (1 << (jacParser.FLOAT - 73))
                        | (1 << (jacParser.STRING - 73))
                        | (1 << (jacParser.BOOL - 73))
                        | (1 << (jacParser.INT - 73))
                        | (1 << (jacParser.NULL - 73))
                        | (1 << (jacParser.NAME - 73))
                    )
                )
                != 0
            ):
                self.state = 1054
                self.expr_list()

            self.state = 1057
            self.match(jacParser.RSQUARE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Index_sliceContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LSQUARE(self):
            return self.getToken(jacParser.LSQUARE, 0)

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(jacParser.ExpressionContext, i)

        def RSQUARE(self):
            return self.getToken(jacParser.RSQUARE, 0)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def getRuleIndex(self):
            return jacParser.RULE_index_slice

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterIndex_slice"):
                listener.enterIndex_slice(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitIndex_slice"):
                listener.exitIndex_slice(self)

    def index_slice(self):

        localctx = jacParser.Index_sliceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 168, self.RULE_index_slice)
        try:
            self.state = 1069
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 121, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1059
                self.match(jacParser.LSQUARE)
                self.state = 1060
                self.expression()
                self.state = 1061
                self.match(jacParser.RSQUARE)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1063
                self.match(jacParser.LSQUARE)
                self.state = 1064
                self.expression()
                self.state = 1065
                self.match(jacParser.COLON)
                self.state = 1066
                self.expression()
                self.state = 1067
                self.match(jacParser.RSQUARE)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dict_valContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def kv_pair(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Kv_pairContext)
            else:
                return self.getTypedRuleContext(jacParser.Kv_pairContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_dict_val

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDict_val"):
                listener.enterDict_val(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDict_val"):
                listener.exitDict_val(self)

    def dict_val(self):

        localctx = jacParser.Dict_valContext(self, self._ctx, self.state)
        self.enterRule(localctx, 170, self.RULE_dict_val)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1071
            self.match(jacParser.LBRACE)
            self.state = 1080
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (
                ((_la) & ~0x3F) == 0
                and (
                    (1 << _la)
                    & (
                        (1 << jacParser.T__1)
                        | (1 << jacParser.T__2)
                        | (1 << jacParser.T__4)
                        | (1 << jacParser.T__5)
                        | (1 << jacParser.T__6)
                        | (1 << jacParser.TYP_STRING)
                        | (1 << jacParser.TYP_INT)
                        | (1 << jacParser.TYP_FLOAT)
                        | (1 << jacParser.TYP_LIST)
                        | (1 << jacParser.TYP_DICT)
                        | (1 << jacParser.TYP_BOOL)
                        | (1 << jacParser.KW_TYPE)
                        | (1 << jacParser.KW_NODE)
                        | (1 << jacParser.KW_SPAWN)
                        | (1 << jacParser.DBL_COLON)
                        | (1 << jacParser.LBRACE)
                        | (1 << jacParser.KW_EDGE)
                        | (1 << jacParser.KW_SYNC)
                    )
                )
                != 0
            ) or (
                (((_la - 73)) & ~0x3F) == 0
                and (
                    (1 << (_la - 73))
                    & (
                        (1 << (jacParser.NOT - 73))
                        | (1 << (jacParser.KW_GLOBAL - 73))
                        | (1 << (jacParser.PLUS - 73))
                        | (1 << (jacParser.MINUS - 73))
                        | (1 << (jacParser.STAR_MUL - 73))
                        | (1 << (jacParser.LPAREN - 73))
                        | (1 << (jacParser.LSQUARE - 73))
                        | (1 << (jacParser.FLOAT - 73))
                        | (1 << (jacParser.STRING - 73))
                        | (1 << (jacParser.BOOL - 73))
                        | (1 << (jacParser.INT - 73))
                        | (1 << (jacParser.NULL - 73))
                        | (1 << (jacParser.NAME - 73))
                    )
                )
                != 0
            ):
                self.state = 1072
                self.kv_pair()
                self.state = 1077
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == jacParser.COMMA:
                    self.state = 1073
                    self.match(jacParser.COMMA)
                    self.state = 1074
                    self.kv_pair()
                    self.state = 1079
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

            self.state = 1082
            self.match(jacParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Kv_pairContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.ExpressionContext)
            else:
                return self.getTypedRuleContext(jacParser.ExpressionContext, i)

        def COLON(self):
            return self.getToken(jacParser.COLON, 0)

        def getRuleIndex(self):
            return jacParser.RULE_kv_pair

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterKv_pair"):
                listener.enterKv_pair(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitKv_pair"):
                listener.exitKv_pair(self)

    def kv_pair(self):

        localctx = jacParser.Kv_pairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 172, self.RULE_kv_pair)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1084
            self.expression()
            self.state = 1085
            self.match(jacParser.COLON)
            self.state = 1086
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SpawnContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def KW_SPAWN(self):
            return self.getToken(jacParser.KW_SPAWN, 0)

        def spawn_object(self):
            return self.getTypedRuleContext(jacParser.Spawn_objectContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_spawn

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSpawn"):
                listener.enterSpawn(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSpawn"):
                listener.exitSpawn(self)

    def spawn(self):

        localctx = jacParser.SpawnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 174, self.RULE_spawn)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1088
            self.match(jacParser.KW_SPAWN)
            self.state = 1089
            self.spawn_object()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Spawn_objectContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def node_spawn(self):
            return self.getTypedRuleContext(jacParser.Node_spawnContext, 0)

        def walker_spawn(self):
            return self.getTypedRuleContext(jacParser.Walker_spawnContext, 0)

        def graph_spawn(self):
            return self.getTypedRuleContext(jacParser.Graph_spawnContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_spawn_object

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSpawn_object"):
                listener.enterSpawn_object(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSpawn_object"):
                listener.exitSpawn_object(self)

    def spawn_object(self):

        localctx = jacParser.Spawn_objectContext(self, self._ctx, self.state)
        self.enterRule(localctx, 176, self.RULE_spawn_object)
        try:
            self.state = 1094
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 124, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1091
                self.node_spawn()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1092
                self.walker_spawn()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1093
                self.graph_spawn()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Spawn_edgeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def edge_ref(self):
            return self.getTypedRuleContext(jacParser.Edge_refContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_spawn_edge

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSpawn_edge"):
                listener.enterSpawn_edge(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSpawn_edge"):
                listener.exitSpawn_edge(self)

    def spawn_edge(self):

        localctx = jacParser.Spawn_edgeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 178, self.RULE_spawn_edge)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1096
            self.expression()
            self.state = 1097
            self.edge_ref()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Node_spawnContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def node_ref(self):
            return self.getTypedRuleContext(jacParser.Node_refContext, 0)

        def spawn_edge(self):
            return self.getTypedRuleContext(jacParser.Spawn_edgeContext, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_node_spawn

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterNode_spawn"):
                listener.enterNode_spawn(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitNode_spawn"):
                listener.exitNode_spawn(self)

    def node_spawn(self):

        localctx = jacParser.Node_spawnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 180, self.RULE_node_spawn)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1100
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 125, self._ctx)
            if la_ == 1:
                self.state = 1099
                self.spawn_edge()

            self.state = 1102
            self.node_ref()
            self.state = 1104
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 126, self._ctx)
            if la_ == 1:
                self.state = 1103
                self.spawn_ctx()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Graph_spawnContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def graph_ref(self):
            return self.getTypedRuleContext(jacParser.Graph_refContext, 0)

        def spawn_edge(self):
            return self.getTypedRuleContext(jacParser.Spawn_edgeContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_graph_spawn

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterGraph_spawn"):
                listener.enterGraph_spawn(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitGraph_spawn"):
                listener.exitGraph_spawn(self)

    def graph_spawn(self):

        localctx = jacParser.Graph_spawnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 182, self.RULE_graph_spawn)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1107
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (
                ((_la) & ~0x3F) == 0
                and (
                    (1 << _la)
                    & (
                        (1 << jacParser.T__1)
                        | (1 << jacParser.T__2)
                        | (1 << jacParser.T__4)
                        | (1 << jacParser.T__5)
                        | (1 << jacParser.T__6)
                        | (1 << jacParser.TYP_STRING)
                        | (1 << jacParser.TYP_INT)
                        | (1 << jacParser.TYP_FLOAT)
                        | (1 << jacParser.TYP_LIST)
                        | (1 << jacParser.TYP_DICT)
                        | (1 << jacParser.TYP_BOOL)
                        | (1 << jacParser.KW_TYPE)
                        | (1 << jacParser.KW_NODE)
                        | (1 << jacParser.KW_SPAWN)
                        | (1 << jacParser.DBL_COLON)
                        | (1 << jacParser.LBRACE)
                        | (1 << jacParser.KW_EDGE)
                        | (1 << jacParser.KW_SYNC)
                    )
                )
                != 0
            ) or (
                (((_la - 73)) & ~0x3F) == 0
                and (
                    (1 << (_la - 73))
                    & (
                        (1 << (jacParser.NOT - 73))
                        | (1 << (jacParser.KW_GLOBAL - 73))
                        | (1 << (jacParser.PLUS - 73))
                        | (1 << (jacParser.MINUS - 73))
                        | (1 << (jacParser.STAR_MUL - 73))
                        | (1 << (jacParser.LPAREN - 73))
                        | (1 << (jacParser.LSQUARE - 73))
                        | (1 << (jacParser.FLOAT - 73))
                        | (1 << (jacParser.STRING - 73))
                        | (1 << (jacParser.BOOL - 73))
                        | (1 << (jacParser.INT - 73))
                        | (1 << (jacParser.NULL - 73))
                        | (1 << (jacParser.NAME - 73))
                    )
                )
                != 0
            ):
                self.state = 1106
                self.spawn_edge()

            self.state = 1109
            self.graph_ref()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Walker_spawnContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def walker_ref(self):
            return self.getTypedRuleContext(jacParser.Walker_refContext, 0)

        def KW_SYNC(self):
            return self.getToken(jacParser.KW_SYNC, 0)

        def spawn_ctx(self):
            return self.getTypedRuleContext(jacParser.Spawn_ctxContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_walker_spawn

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterWalker_spawn"):
                listener.enterWalker_spawn(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitWalker_spawn"):
                listener.exitWalker_spawn(self)

    def walker_spawn(self):

        localctx = jacParser.Walker_spawnContext(self, self._ctx, self.state)
        self.enterRule(localctx, 184, self.RULE_walker_spawn)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1111
            self.expression()
            self.state = 1113
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.KW_SYNC:
                self.state = 1112
                self.match(jacParser.KW_SYNC)

            self.state = 1115
            self.walker_ref()
            self.state = 1117
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 129, self._ctx)
            if la_ == 1:
                self.state = 1116
                self.spawn_ctx()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Spawn_ctxContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def spawn_assign(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Spawn_assignContext)
            else:
                return self.getTypedRuleContext(jacParser.Spawn_assignContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_spawn_ctx

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSpawn_ctx"):
                listener.enterSpawn_ctx(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSpawn_ctx"):
                listener.exitSpawn_ctx(self)

    def spawn_ctx(self):

        localctx = jacParser.Spawn_ctxContext(self, self._ctx, self.state)
        self.enterRule(localctx, 186, self.RULE_spawn_ctx)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1119
            self.match(jacParser.LPAREN)
            self.state = 1128
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.NAME:
                self.state = 1120
                self.spawn_assign()
                self.state = 1125
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == jacParser.COMMA:
                    self.state = 1121
                    self.match(jacParser.COMMA)
                    self.state = 1122
                    self.spawn_assign()
                    self.state = 1127
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

            self.state = 1130
            self.match(jacParser.RPAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Filter_ctxContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LPAREN(self):
            return self.getToken(jacParser.LPAREN, 0)

        def RPAREN(self):
            return self.getToken(jacParser.RPAREN, 0)

        def filter_compare(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Filter_compareContext)
            else:
                return self.getTypedRuleContext(jacParser.Filter_compareContext, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_filter_ctx

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFilter_ctx"):
                listener.enterFilter_ctx(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFilter_ctx"):
                listener.exitFilter_ctx(self)

    def filter_ctx(self):

        localctx = jacParser.Filter_ctxContext(self, self._ctx, self.state)
        self.enterRule(localctx, 188, self.RULE_filter_ctx)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1132
            self.match(jacParser.LPAREN)
            self.state = 1141
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.NAME:
                self.state = 1133
                self.filter_compare()
                self.state = 1138
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == jacParser.COMMA:
                    self.state = 1134
                    self.match(jacParser.COMMA)
                    self.state = 1135
                    self.filter_compare()
                    self.state = 1140
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

            self.state = 1143
            self.match(jacParser.RPAREN)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Spawn_assignContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def EQ(self):
            return self.getToken(jacParser.EQ, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_spawn_assign

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterSpawn_assign"):
                listener.enterSpawn_assign(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitSpawn_assign"):
                listener.exitSpawn_assign(self)

    def spawn_assign(self):

        localctx = jacParser.Spawn_assignContext(self, self._ctx, self.state)
        self.enterRule(localctx, 190, self.RULE_spawn_assign)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1145
            self.match(jacParser.NAME)
            self.state = 1146
            self.match(jacParser.EQ)
            self.state = 1147
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Filter_compareContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def cmp_op(self):
            return self.getTypedRuleContext(jacParser.Cmp_opContext, 0)

        def expression(self):
            return self.getTypedRuleContext(jacParser.ExpressionContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_filter_compare

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterFilter_compare"):
                listener.enterFilter_compare(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitFilter_compare"):
                listener.exitFilter_compare(self)

    def filter_compare(self):

        localctx = jacParser.Filter_compareContext(self, self._ctx, self.state)
        self.enterRule(localctx, 192, self.RULE_filter_compare)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1149
            self.match(jacParser.NAME)
            self.state = 1150
            self.cmp_op()
            self.state = 1151
            self.expression()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Any_typeContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TYP_STRING(self):
            return self.getToken(jacParser.TYP_STRING, 0)

        def TYP_INT(self):
            return self.getToken(jacParser.TYP_INT, 0)

        def TYP_FLOAT(self):
            return self.getToken(jacParser.TYP_FLOAT, 0)

        def TYP_LIST(self):
            return self.getToken(jacParser.TYP_LIST, 0)

        def TYP_DICT(self):
            return self.getToken(jacParser.TYP_DICT, 0)

        def TYP_BOOL(self):
            return self.getToken(jacParser.TYP_BOOL, 0)

        def KW_NODE(self):
            return self.getToken(jacParser.KW_NODE, 0)

        def KW_EDGE(self):
            return self.getToken(jacParser.KW_EDGE, 0)

        def KW_TYPE(self):
            return self.getToken(jacParser.KW_TYPE, 0)

        def getRuleIndex(self):
            return jacParser.RULE_any_type

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterAny_type"):
                listener.enterAny_type(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitAny_type"):
                listener.exitAny_type(self)

    def any_type(self):

        localctx = jacParser.Any_typeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 194, self.RULE_any_type)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1153
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.TYP_STRING)
                            | (1 << jacParser.TYP_INT)
                            | (1 << jacParser.TYP_FLOAT)
                            | (1 << jacParser.TYP_LIST)
                            | (1 << jacParser.TYP_DICT)
                            | (1 << jacParser.TYP_BOOL)
                            | (1 << jacParser.KW_TYPE)
                            | (1 << jacParser.KW_NODE)
                            | (1 << jacParser.KW_EDGE)
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_graphContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def dot_stmt_list(self):
            return self.getTypedRuleContext(jacParser.Dot_stmt_listContext, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def KW_GRAPH(self):
            return self.getToken(jacParser.KW_GRAPH, 0)

        def KW_DIGRAPH(self):
            return self.getToken(jacParser.KW_DIGRAPH, 0)

        def KW_STRICT(self):
            return self.getToken(jacParser.KW_STRICT, 0)

        def dot_id(self):
            return self.getTypedRuleContext(jacParser.Dot_idContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_graph

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_graph"):
                listener.enterDot_graph(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_graph"):
                listener.exitDot_graph(self)

    def dot_graph(self):

        localctx = jacParser.Dot_graphContext(self, self._ctx, self.state)
        self.enterRule(localctx, 196, self.RULE_dot_graph)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1156
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.KW_STRICT:
                self.state = 1155
                self.match(jacParser.KW_STRICT)

            self.state = 1158
            _la = self._input.LA(1)
            if not (_la == jacParser.KW_GRAPH or _la == jacParser.KW_DIGRAPH):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 1160
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (
                ((_la) & ~0x3F) == 0
                and (
                    (1 << _la)
                    & (
                        (1 << jacParser.KW_GRAPH)
                        | (1 << jacParser.KW_NODE)
                        | (1 << jacParser.KW_EDGE)
                    )
                )
                != 0
            ) or (
                (((_la - 97)) & ~0x3F) == 0
                and (
                    (1 << (_la - 97))
                    & (
                        (1 << (jacParser.FLOAT - 97))
                        | (1 << (jacParser.STRING - 97))
                        | (1 << (jacParser.INT - 97))
                        | (1 << (jacParser.NAME - 97))
                    )
                )
                != 0
            ):
                self.state = 1159
                self.dot_id()

            self.state = 1162
            self.match(jacParser.LBRACE)
            self.state = 1163
            self.dot_stmt_list()
            self.state = 1164
            self.match(jacParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_stmt_listContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_stmt(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_stmtContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_stmtContext, i)

        def SEMI(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.SEMI)
            else:
                return self.getToken(jacParser.SEMI, i)

        def getRuleIndex(self):
            return jacParser.RULE_dot_stmt_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_stmt_list"):
                listener.enterDot_stmt_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_stmt_list"):
                listener.exitDot_stmt_list(self)

    def dot_stmt_list(self):

        localctx = jacParser.Dot_stmt_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 198, self.RULE_dot_stmt_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1172
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (
                ((_la) & ~0x3F) == 0
                and (
                    (1 << _la)
                    & (
                        (1 << jacParser.KW_GRAPH)
                        | (1 << jacParser.KW_SUBGRAPH)
                        | (1 << jacParser.KW_NODE)
                        | (1 << jacParser.LBRACE)
                        | (1 << jacParser.KW_EDGE)
                    )
                )
                != 0
            ) or (
                (((_la - 97)) & ~0x3F) == 0
                and (
                    (1 << (_la - 97))
                    & (
                        (1 << (jacParser.FLOAT - 97))
                        | (1 << (jacParser.STRING - 97))
                        | (1 << (jacParser.INT - 97))
                        | (1 << (jacParser.NAME - 97))
                    )
                )
                != 0
            ):
                self.state = 1166
                self.dot_stmt()
                self.state = 1168
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.SEMI:
                    self.state = 1167
                    self.match(jacParser.SEMI)

                self.state = 1174
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_node_stmt(self):
            return self.getTypedRuleContext(jacParser.Dot_node_stmtContext, 0)

        def dot_edge_stmt(self):
            return self.getTypedRuleContext(jacParser.Dot_edge_stmtContext, 0)

        def dot_attr_stmt(self):
            return self.getTypedRuleContext(jacParser.Dot_attr_stmtContext, 0)

        def dot_id(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_idContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_idContext, i)

        def EQ(self):
            return self.getToken(jacParser.EQ, 0)

        def dot_subgraph(self):
            return self.getTypedRuleContext(jacParser.Dot_subgraphContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_stmt"):
                listener.enterDot_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_stmt"):
                listener.exitDot_stmt(self)

    def dot_stmt(self):

        localctx = jacParser.Dot_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 200, self.RULE_dot_stmt)
        try:
            self.state = 1183
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 138, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 1175
                self.dot_node_stmt()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 1176
                self.dot_edge_stmt()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 1177
                self.dot_attr_stmt()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 1178
                self.dot_id()
                self.state = 1179
                self.match(jacParser.EQ)
                self.state = 1180
                self.dot_id()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 1182
                self.dot_subgraph()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_attr_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_attr_list(self):
            return self.getTypedRuleContext(jacParser.Dot_attr_listContext, 0)

        def KW_GRAPH(self):
            return self.getToken(jacParser.KW_GRAPH, 0)

        def KW_NODE(self):
            return self.getToken(jacParser.KW_NODE, 0)

        def KW_EDGE(self):
            return self.getToken(jacParser.KW_EDGE, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_attr_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_attr_stmt"):
                listener.enterDot_attr_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_attr_stmt"):
                listener.exitDot_attr_stmt(self)

    def dot_attr_stmt(self):

        localctx = jacParser.Dot_attr_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 202, self.RULE_dot_attr_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1185
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.KW_GRAPH)
                            | (1 << jacParser.KW_NODE)
                            | (1 << jacParser.KW_EDGE)
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 1186
            self.dot_attr_list()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_attr_listContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LSQUARE(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.LSQUARE)
            else:
                return self.getToken(jacParser.LSQUARE, i)

        def RSQUARE(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.RSQUARE)
            else:
                return self.getToken(jacParser.RSQUARE, i)

        def dot_a_list(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_a_listContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_a_listContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_dot_attr_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_attr_list"):
                listener.enterDot_attr_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_attr_list"):
                listener.exitDot_attr_list(self)

    def dot_attr_list(self):

        localctx = jacParser.Dot_attr_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 204, self.RULE_dot_attr_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1193
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 1188
                self.match(jacParser.LSQUARE)
                self.state = 1190
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.KW_GRAPH)
                            | (1 << jacParser.KW_NODE)
                            | (1 << jacParser.KW_EDGE)
                        )
                    )
                    != 0
                ) or (
                    (((_la - 97)) & ~0x3F) == 0
                    and (
                        (1 << (_la - 97))
                        & (
                            (1 << (jacParser.FLOAT - 97))
                            | (1 << (jacParser.STRING - 97))
                            | (1 << (jacParser.INT - 97))
                            | (1 << (jacParser.NAME - 97))
                        )
                    )
                    != 0
                ):
                    self.state = 1189
                    self.dot_a_list()

                self.state = 1192
                self.match(jacParser.RSQUARE)
                self.state = 1195
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == jacParser.LSQUARE):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_a_listContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_id(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_idContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_idContext, i)

        def EQ(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.EQ)
            else:
                return self.getToken(jacParser.EQ, i)

        def COMMA(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COMMA)
            else:
                return self.getToken(jacParser.COMMA, i)

        def getRuleIndex(self):
            return jacParser.RULE_dot_a_list

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_a_list"):
                listener.enterDot_a_list(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_a_list"):
                listener.exitDot_a_list(self)

    def dot_a_list(self):

        localctx = jacParser.Dot_a_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 206, self.RULE_dot_a_list)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1205
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 1197
                self.dot_id()
                self.state = 1200
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.EQ:
                    self.state = 1198
                    self.match(jacParser.EQ)
                    self.state = 1199
                    self.dot_id()

                self.state = 1203
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == jacParser.COMMA:
                    self.state = 1202
                    self.match(jacParser.COMMA)

                self.state = 1207
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (
                    (
                        ((_la) & ~0x3F) == 0
                        and (
                            (1 << _la)
                            & (
                                (1 << jacParser.KW_GRAPH)
                                | (1 << jacParser.KW_NODE)
                                | (1 << jacParser.KW_EDGE)
                            )
                        )
                        != 0
                    )
                    or (
                        (((_la - 97)) & ~0x3F) == 0
                        and (
                            (1 << (_la - 97))
                            & (
                                (1 << (jacParser.FLOAT - 97))
                                | (1 << (jacParser.STRING - 97))
                                | (1 << (jacParser.INT - 97))
                                | (1 << (jacParser.NAME - 97))
                            )
                        )
                        != 0
                    )
                ):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_edge_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_edgeRHS(self):
            return self.getTypedRuleContext(jacParser.Dot_edgeRHSContext, 0)

        def dot_node_id(self):
            return self.getTypedRuleContext(jacParser.Dot_node_idContext, 0)

        def dot_subgraph(self):
            return self.getTypedRuleContext(jacParser.Dot_subgraphContext, 0)

        def dot_attr_list(self):
            return self.getTypedRuleContext(jacParser.Dot_attr_listContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_edge_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_edge_stmt"):
                listener.enterDot_edge_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_edge_stmt"):
                listener.exitDot_edge_stmt(self)

    def dot_edge_stmt(self):

        localctx = jacParser.Dot_edge_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 208, self.RULE_dot_edge_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1211
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [
                jacParser.KW_GRAPH,
                jacParser.KW_NODE,
                jacParser.KW_EDGE,
                jacParser.FLOAT,
                jacParser.STRING,
                jacParser.INT,
                jacParser.NAME,
            ]:
                self.state = 1209
                self.dot_node_id()
                pass
            elif token in [jacParser.KW_SUBGRAPH, jacParser.LBRACE]:
                self.state = 1210
                self.dot_subgraph()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 1213
            self.dot_edgeRHS()
            self.state = 1215
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.LSQUARE:
                self.state = 1214
                self.dot_attr_list()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_edgeRHSContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_edgeop(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_edgeopContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_edgeopContext, i)

        def dot_node_id(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_node_idContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_node_idContext, i)

        def dot_subgraph(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_subgraphContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_subgraphContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_dot_edgeRHS

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_edgeRHS"):
                listener.enterDot_edgeRHS(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_edgeRHS"):
                listener.exitDot_edgeRHS(self)

    def dot_edgeRHS(self):

        localctx = jacParser.Dot_edgeRHSContext(self, self._ctx, self.state)
        self.enterRule(localctx, 210, self.RULE_dot_edgeRHS)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1222
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 1217
                self.dot_edgeop()
                self.state = 1220
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [
                    jacParser.KW_GRAPH,
                    jacParser.KW_NODE,
                    jacParser.KW_EDGE,
                    jacParser.FLOAT,
                    jacParser.STRING,
                    jacParser.INT,
                    jacParser.NAME,
                ]:
                    self.state = 1218
                    self.dot_node_id()
                    pass
                elif token in [jacParser.KW_SUBGRAPH, jacParser.LBRACE]:
                    self.state = 1219
                    self.dot_subgraph()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 1224
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == jacParser.T__3 or _la == jacParser.T__7):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_edgeopContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return jacParser.RULE_dot_edgeop

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_edgeop"):
                listener.enterDot_edgeop(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_edgeop"):
                listener.exitDot_edgeop(self)

    def dot_edgeop(self):

        localctx = jacParser.Dot_edgeopContext(self, self._ctx, self.state)
        self.enterRule(localctx, 212, self.RULE_dot_edgeop)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1226
            _la = self._input.LA(1)
            if not (_la == jacParser.T__3 or _la == jacParser.T__7):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_node_stmtContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_node_id(self):
            return self.getTypedRuleContext(jacParser.Dot_node_idContext, 0)

        def dot_attr_list(self):
            return self.getTypedRuleContext(jacParser.Dot_attr_listContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_node_stmt

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_node_stmt"):
                listener.enterDot_node_stmt(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_node_stmt"):
                listener.exitDot_node_stmt(self)

    def dot_node_stmt(self):

        localctx = jacParser.Dot_node_stmtContext(self, self._ctx, self.state)
        self.enterRule(localctx, 214, self.RULE_dot_node_stmt)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1228
            self.dot_node_id()
            self.state = 1230
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.LSQUARE:
                self.state = 1229
                self.dot_attr_list()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_node_idContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def dot_id(self):
            return self.getTypedRuleContext(jacParser.Dot_idContext, 0)

        def dot_port(self):
            return self.getTypedRuleContext(jacParser.Dot_portContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_node_id

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_node_id"):
                listener.enterDot_node_id(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_node_id"):
                listener.exitDot_node_id(self)

    def dot_node_id(self):

        localctx = jacParser.Dot_node_idContext(self, self._ctx, self.state)
        self.enterRule(localctx, 216, self.RULE_dot_node_id)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1232
            self.dot_id()
            self.state = 1234
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.COLON:
                self.state = 1233
                self.dot_port()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_portContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COLON(self, i: int = None):
            if i is None:
                return self.getTokens(jacParser.COLON)
            else:
                return self.getToken(jacParser.COLON, i)

        def dot_id(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(jacParser.Dot_idContext)
            else:
                return self.getTypedRuleContext(jacParser.Dot_idContext, i)

        def getRuleIndex(self):
            return jacParser.RULE_dot_port

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_port"):
                listener.enterDot_port(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_port"):
                listener.exitDot_port(self)

    def dot_port(self):

        localctx = jacParser.Dot_portContext(self, self._ctx, self.state)
        self.enterRule(localctx, 218, self.RULE_dot_port)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1236
            self.match(jacParser.COLON)
            self.state = 1237
            self.dot_id()
            self.state = 1240
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.COLON:
                self.state = 1238
                self.match(jacParser.COLON)
                self.state = 1239
                self.dot_id()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_subgraphContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(jacParser.LBRACE, 0)

        def dot_stmt_list(self):
            return self.getTypedRuleContext(jacParser.Dot_stmt_listContext, 0)

        def RBRACE(self):
            return self.getToken(jacParser.RBRACE, 0)

        def KW_SUBGRAPH(self):
            return self.getToken(jacParser.KW_SUBGRAPH, 0)

        def dot_id(self):
            return self.getTypedRuleContext(jacParser.Dot_idContext, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_subgraph

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_subgraph"):
                listener.enterDot_subgraph(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_subgraph"):
                listener.exitDot_subgraph(self)

    def dot_subgraph(self):

        localctx = jacParser.Dot_subgraphContext(self, self._ctx, self.state)
        self.enterRule(localctx, 220, self.RULE_dot_subgraph)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1246
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la == jacParser.KW_SUBGRAPH:
                self.state = 1242
                self.match(jacParser.KW_SUBGRAPH)
                self.state = 1244
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.KW_GRAPH)
                            | (1 << jacParser.KW_NODE)
                            | (1 << jacParser.KW_EDGE)
                        )
                    )
                    != 0
                ) or (
                    (((_la - 97)) & ~0x3F) == 0
                    and (
                        (1 << (_la - 97))
                        & (
                            (1 << (jacParser.FLOAT - 97))
                            | (1 << (jacParser.STRING - 97))
                            | (1 << (jacParser.INT - 97))
                            | (1 << (jacParser.NAME - 97))
                        )
                    )
                    != 0
                ):
                    self.state = 1243
                    self.dot_id()

            self.state = 1248
            self.match(jacParser.LBRACE)
            self.state = 1249
            self.dot_stmt_list()
            self.state = 1250
            self.match(jacParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Dot_idContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NAME(self):
            return self.getToken(jacParser.NAME, 0)

        def STRING(self):
            return self.getToken(jacParser.STRING, 0)

        def INT(self):
            return self.getToken(jacParser.INT, 0)

        def FLOAT(self):
            return self.getToken(jacParser.FLOAT, 0)

        def KW_GRAPH(self):
            return self.getToken(jacParser.KW_GRAPH, 0)

        def KW_NODE(self):
            return self.getToken(jacParser.KW_NODE, 0)

        def KW_EDGE(self):
            return self.getToken(jacParser.KW_EDGE, 0)

        def getRuleIndex(self):
            return jacParser.RULE_dot_id

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDot_id"):
                listener.enterDot_id(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDot_id"):
                listener.exitDot_id(self)

    def dot_id(self):

        localctx = jacParser.Dot_idContext(self, self._ctx, self.state)
        self.enterRule(localctx, 222, self.RULE_dot_id)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 1252
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << jacParser.KW_GRAPH)
                            | (1 << jacParser.KW_NODE)
                            | (1 << jacParser.KW_EDGE)
                        )
                    )
                    != 0
                )
                or (
                    (((_la - 97)) & ~0x3F) == 0
                    and (
                        (1 << (_la - 97))
                        & (
                            (1 << (jacParser.FLOAT - 97))
                            | (1 << (jacParser.STRING - 97))
                            | (1 << (jacParser.INT - 97))
                            | (1 << (jacParser.NAME - 97))
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[63] = self.atom_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def atom_sempred(self, localctx: AtomContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 6)
